// src/index.ts
import { lex } from "coffee-lex";
import { nodes as getCoffee1Nodes, tokens as getCoffee1Tokens } from "decaffeinate-coffeescript";
import { nodes as getCoffee2Nodes, tokens as getCoffee2Tokens } from "decaffeinate-coffeescript2";
import { parse as decaffeinateParse2 } from "decaffeinate-parser";

// src/stages/add-variable-declarations/index.ts
import addVariableDeclarations from "add-variable-declarations";
import MagicString from "magic-string";

// src/utils/debug.ts
function logger(name) {
  if (isLoggingEnabled(name)) {
    return (...args) => console.log(name, ...args);
  } else {
    return () => {
    };
  }
}
function isLoggingEnabled(name) {
  return !!process.env[`DEBUG:${name}`] || !!process.env["DEBUG:*"];
}

// src/stages/add-variable-declarations/index.ts
var AddVariableDeclarationsStage = class {
  static run(content) {
    const log = logger(this.name);
    log(content);
    const editor = new MagicString(content);
    addVariableDeclarations(content, editor);
    return {
      code: editor.toString(),
      suggestions: []
    };
  }
};

// src/stages/resugar/index.ts
import * as t from "@babel/types";
import declarationsBlockScope from "@resugar/codemod-declarations-block-scope";
import functionsArrow from "@resugar/codemod-functions-arrow";
import modulesCommonjs from "@resugar/codemod-modules-commonjs";
import objectsConcise from "@resugar/codemod-objects-concise";
import objectsDestructuring from "@resugar/codemod-objects-destructuring";
import objectsShorthand from "@resugar/codemod-objects-shorthand";
import stringsTemplate from "@resugar/codemod-strings-template";
import { transform } from "@codemod/core";
var ResugarStage = class {
  static run(content, options) {
    const log = logger(this.name);
    log(content);
    const code = transform(content, {
      plugins: Array.from(this.getPluginsForOptions(options)),
      configFile: false,
      babelrc: false
    }).code;
    return {
      code,
      suggestions: []
    };
  }
  static *getPluginsForOptions(options) {
    yield objectsShorthand;
    yield objectsConcise;
    if (options.useJSModules) {
      yield [
        modulesCommonjs,
        {
          forceDefaultExport: !options.looseJSModules,
          safeFunctionIdentifiers: options.safeImportFunctionIdentifiers
        }
      ];
    }
    yield functionsArrow;
    yield [
      declarationsBlockScope,
      {
        disableConst({ node, parent }) {
          if (options.preferLet) {
            return !t.isProgram(parent) || node.declarations.length !== 1 || !t.isIdentifier(node.declarations[0].id) || !/^[$_]?[A-Z]+$/.test(node.declarations[0].id.name);
          } else {
            return false;
          }
        }
      }
    ];
    yield objectsDestructuring;
    yield stringsTemplate;
  }
};

// src/stages/literate/index.ts
var VALID_INDENTATIONS = ["    ", "   	", "  	", " 	", "	"];
var LiterateStage = class {
  static run(content) {
    return {
      code: convertCodeFromLiterate(content),
      suggestions: []
    };
  }
};
function convertCodeFromLiterate(code) {
  const lines = code.split("\n");
  const resultLines = [];
  let commentLines = null;
  for (let i = 0; i < lines.length; i++) {
    const line = lines[i];
    if (commentLines === null) {
      if (lineIsEmpty(line) || lineIsIndented(line)) {
        resultLines.push(removeIndentation(line));
      } else {
        commentLines = [line];
      }
    } else {
      if (lineIsEmpty(line) || !lineIsIndented(line) || i > 0 && !lineIsEmpty(lines[i - 1])) {
        commentLines.push(line);
      } else {
        resultLines.push(...convertCommentLines(commentLines));
        commentLines = null;
        resultLines.push(removeIndentation(line));
      }
    }
  }
  resultLines.push(...convertCommentLines(commentLines));
  return resultLines.join("\n");
}
function convertCommentLines(commentLines) {
  if (commentLines === null) {
    return [];
  }
  commentLines = commentLines.slice();
  while (commentLines.length > 0 && lineIsEmpty(commentLines[commentLines.length - 1])) {
    commentLines.pop();
  }
  return commentLines.map((line) => `# ${line}`);
}
function lineIsEmpty(line) {
  return /^\s*$/.test(line);
}
function lineIsIndented(line) {
  return VALID_INDENTATIONS.some((indent) => line.startsWith(indent));
}
function removeIndentation(line) {
  for (const indent of VALID_INDENTATIONS) {
    if (line.startsWith(indent)) {
      return line.slice(indent.length);
    }
  }
  if (lineIsEmpty(line)) {
    return line;
  }
  throw new Error("Unexpectedly removed indentation from an unindented line.");
}

// src/stages/TransformCoffeeScriptStage.ts
import MagicString2 from "magic-string";

// src/utils/DecaffeinateContext.ts
import { parse as decaffeinateParse, traverse } from "decaffeinate-parser";

// src/utils/Scope.ts
import {
  ArrayInitialiser as ArrayInitialiser2,
  AssignOp,
  BaseFunction,
  Class,
  CompoundAssignOp,
  DefaultParam,
  Expansion,
  For,
  Identifier as Identifier2,
  MemberAccessOp,
  ObjectInitialiser as ObjectInitialiser2,
  PostDecrementOp,
  PostIncrementOp,
  PreDecrementOp,
  PreIncrementOp,
  Rest,
  Try
} from "decaffeinate-parser";

// src/utils/flatMap.ts
function flatMap(list, map) {
  return list.reduce((memo, item) => memo.concat(map(item)), []);
}

// src/utils/isReservedWord.ts
var JS_KEYWORDS = [
  "true",
  "false",
  "null",
  "this",
  "new",
  "delete",
  "typeof",
  "in",
  "instanceof",
  "return",
  "throw",
  "break",
  "continue",
  "debugger",
  "yield",
  "if",
  "else",
  "switch",
  "for",
  "while",
  "do",
  "try",
  "catch",
  "finally",
  "class",
  "extends",
  "super",
  "import",
  "export",
  "default"
];
var COFFEE_KEYWORDS = ["undefined", "Infinity", "NaN", "then", "unless", "until", "loop", "of", "by", "when"];
var COFFEE_ALIASES = ["and", "or", "is", "isnt", "not", "yes", "no", "on", "off"];
var RESERVED = [
  "case",
  "default",
  "function",
  "var",
  "void",
  "with",
  "const",
  "let",
  "enum",
  "export",
  "import",
  "native",
  "implements",
  "interface",
  "package",
  "private",
  "protected",
  "public",
  "static"
];
var STRICT_PROSCRIBED = ["arguments", "eval"];
var JS_FORBIDDEN = /* @__PURE__ */ new Set([...JS_KEYWORDS, ...RESERVED, ...STRICT_PROSCRIBED]);
var RESERVED_WORDS = /* @__PURE__ */ new Set([
  ...JS_KEYWORDS,
  ...COFFEE_KEYWORDS,
  ...COFFEE_ALIASES,
  ...RESERVED,
  ...STRICT_PROSCRIBED,
  "await"
]);
function isReservedWord(name) {
  return RESERVED_WORDS.has(name);
}
function isForbiddenJsName(name) {
  return JS_FORBIDDEN.has(name);
}

// src/utils/leftHandIdentifiers.ts
import {
  ArrayInitialiser,
  Identifier,
  ObjectInitialiser,
  ObjectInitialiserMember,
  Spread
} from "decaffeinate-parser";
function leftHandIdentifiers(node) {
  if (node instanceof Identifier) {
    return [node];
  } else if (node instanceof ArrayInitialiser) {
    return flatMap(node.members, leftHandIdentifiers);
  } else if (node instanceof ObjectInitialiser) {
    return flatMap(node.members, (member) => {
      if (member instanceof ObjectInitialiserMember) {
        return leftHandIdentifiers(member.expression || member.key);
      } else if (member instanceof Spread) {
        return leftHandIdentifiers(member.expression);
      }
      {
        return leftHandIdentifiers(member.assignee);
      }
    });
  } else {
    return [];
  }
}

// src/utils/Scope.ts
var Scope = class {
  constructor(containerNode, parent = null) {
    this.containerNode = containerNode;
    this.parent = parent;
    this.bindings = Object.create(parent ? parent.bindings : {});
    this.modificationsAfterDeclaration = {};
    this.innerClosureModifications = {};
  }
  getBinding(name) {
    return this.bindings[this.key(name)] || null;
  }
  isBindingAvailable(name) {
    return !this.getBinding(name) && !isReservedWord(name);
  }
  hasBinding(name) {
    return this.getBinding(name) !== null;
  }
  hasModificationAfterDeclaration(name) {
    return this.modificationsAfterDeclaration[this.key(name)] || false;
  }
  hasInnerClosureModification(name) {
    return this.innerClosureModifications[this.key(name)] || false;
  }
  getOwnNames() {
    return Object.getOwnPropertyNames(this.bindings).map((key) => this.unkey(key));
  }
  hasOwnBinding(name) {
    return Object.prototype.hasOwnProperty.call(this.bindings, this.key(name));
  }
  declares(name, node) {
    const key = this.key(name);
    this.bindings[key] = node;
  }
  assigns(name, node) {
    if (!this.bindings[this.key(name)]) {
      this.declares(name, node);
    } else {
      this.modifies(name);
    }
  }
  modifies(name) {
    let scope = this;
    while (scope) {
      if (scope.hasOwnBinding(name)) {
        scope.modificationsAfterDeclaration[this.key(name)] = true;
        if (scope !== this) {
          scope.innerClosureModifications[this.key(name)] = true;
        }
        break;
      }
      scope = scope.parent;
    }
  }
  claimFreeBinding(node, name = null) {
    if (!name) {
      name = "ref";
    }
    const names = Array.isArray(name) ? name : [name];
    let binding = names.find((name2) => this.isBindingAvailable(name2));
    if (!binding) {
      let counter = 0;
      while (!binding) {
        if (counter > 1e3) {
          throw new Error(`Unable to find free binding for names ${names.toString()}`);
        }
        counter += 1;
        binding = names.find((name2) => this.isBindingAvailable(`${name2}${counter}`));
      }
      binding = `${binding}${counter}`;
    }
    this.declares(binding, node);
    return binding;
  }
  key(name) {
    return `$${name}`;
  }
  unkey(key) {
    return key.slice(1);
  }
  processNode(node) {
    if (node instanceof AssignOp) {
      leftHandIdentifiers(node.assignee).forEach((identifier) => this.assigns(identifier.data, identifier));
    } else if (node instanceof CompoundAssignOp) {
      if (node.assignee instanceof Identifier2) {
        this.modifies(node.assignee.data);
      }
    } else if (node instanceof PostDecrementOp || node instanceof PostIncrementOp || node instanceof PreDecrementOp || node instanceof PreIncrementOp) {
      if (node.expression instanceof Identifier2) {
        this.modifies(node.expression.data);
      }
    } else if (node instanceof BaseFunction) {
      getBindingsForNode(node).forEach((identifier) => this.declares(identifier.data, identifier));
    } else if (node instanceof For) {
      [node.keyAssignee, node.valAssignee].forEach((assignee) => {
        if (assignee) {
          leftHandIdentifiers(assignee).forEach((identifier) => this.assigns(identifier.data, identifier));
        }
      });
    } else if (node instanceof Try) {
      if (node.catchAssignee) {
        leftHandIdentifiers(node.catchAssignee).forEach((identifier) => this.assigns(identifier.data, identifier));
      }
    } else if (node instanceof Class) {
      if (node.nameAssignee && node.nameAssignee instanceof Identifier2 && this.parent) {
        this.parent.assigns(node.nameAssignee.data, node.nameAssignee);
      }
    }
  }
  toString() {
    const parts = this.getOwnNames();
    if (this.parent) {
      parts.push(`parent = ${this.parent.toString()}`);
    }
    return `${this.constructor.name} {${parts.length > 0 ? ` ${parts.join(", ")} ` : ""}}`;
  }
  inspect() {
    return this.toString();
  }
};
function getBindingsForNode(node) {
  if (node instanceof BaseFunction) {
    return flatMap(node.parameters, getBindingsForNode);
  } else if (node instanceof Identifier2 || node instanceof ArrayInitialiser2 || node instanceof ObjectInitialiser2) {
    return leftHandIdentifiers(node);
  } else if (node instanceof DefaultParam) {
    return getBindingsForNode(node.param);
  } else if (node instanceof Rest) {
    return getBindingsForNode(node.expression);
  } else if (node instanceof Expansion || node instanceof MemberAccessOp) {
    return [];
  } else {
    throw new Error(`unexpected parameter type: ${node.type}`);
  }
}

// src/utils/DecaffeinateContext.ts
var DecaffeinateContext = class {
  constructor(programNode, source, sourceTokens, coffeeAST, linesAndColumns, parentMap, scopeMap) {
    this.programNode = programNode;
    this.source = source;
    this.sourceTokens = sourceTokens;
    this.coffeeAST = coffeeAST;
    this.linesAndColumns = linesAndColumns;
    this.parentMap = parentMap;
    this.scopeMap = scopeMap;
  }
  static create(source, useCS2) {
    const program = decaffeinateParse(source, { useCS2 });
    return new DecaffeinateContext(program, source, program.context.sourceTokens, program.context.ast, program.context.linesAndColumns, computeParentMap(program), computeScopeMap(program));
  }
  getParent(node) {
    const result = this.parentMap.get(node);
    if (result === void 0) {
      throw new Error("Unexpected parent lookup; node was not in the map.");
    }
    return result;
  }
  getScope(node) {
    const result = this.scopeMap.get(node);
    if (result === void 0) {
      throw new Error("Unexpected scope lookup; node was not in the map.");
    }
    return result;
  }
};
function computeParentMap(program) {
  const resultMap = /* @__PURE__ */ new Map();
  traverse(program, (node, parent) => {
    resultMap.set(node, parent);
  });
  return resultMap;
}
function computeScopeMap(program) {
  const scopeMap = /* @__PURE__ */ new Map();
  traverse(program, (node, parent) => {
    let scope;
    switch (node.type) {
      case "Program":
        scope = new Scope(node);
        break;
      case "Function":
      case "BoundFunction":
      case "GeneratorFunction":
      case "BoundGeneratorFunction":
      case "AsyncFunction":
      case "Class": {
        const parentScope = parent && scopeMap.get(parent);
        if (!parentScope) {
          throw new Error("Expected to find parent scope.");
        }
        scope = new Scope(node, parentScope);
        break;
      }
      default: {
        const parentScope = parent && scopeMap.get(parent);
        if (!parentScope) {
          throw new Error("Expected to find parent scope.");
        }
        scope = parentScope;
        break;
      }
    }
    scope.processNode(node);
    scopeMap.set(node, scope);
  });
  return scopeMap;
}

// src/utils/notNull.ts
function notNull(t2) {
  if (t2 === null || t2 === void 0) {
    throw new Error("Unexpected null value.");
  }
  return t2;
}

// src/utils/PatchError.ts
import { LinesAndColumns } from "lines-and-columns";

// src/utils/printTable.ts
function printTable(table, buffer = " ") {
  const widths = [];
  table.rows.forEach((row) => {
    row.forEach((cell, i) => {
      if (widths.length <= i) {
        widths[i] = cell.length;
      } else if (widths[i] < cell.length) {
        widths[i] = cell.length;
      }
    });
  });
  let output = "";
  table.rows.forEach((row) => {
    row.forEach((cell, i) => {
      const column = table.columns[i];
      if (column.align === "left") {
        output += cell;
      } else if (column.align === "right") {
        output += " ".repeat(widths[i] - cell.length) + cell;
      }
      if (i < row.length - 1) {
        output += buffer;
      }
    });
    output += "\n";
  });
  return output;
}

// src/utils/PatchError.ts
var PatchError = class extends Error {
  constructor(message, source, start, end) {
    super(message);
    this.message = message;
    this.source = source;
    this.start = start;
    this.end = end;
  }
  toString() {
    return this.message;
  }
  static detect(error) {
    return error instanceof Error && "source" in error && "start" in error && "end" in error;
  }
  static prettyPrint(error) {
    const { source, message } = error;
    let { start, end } = error;
    start = Math.min(Math.max(start, 0), source.length);
    end = Math.min(Math.max(end, start), source.length);
    const lineMap = new LinesAndColumns(source);
    const startLoc = lineMap.locationForIndex(start);
    const endLoc = lineMap.locationForIndex(end);
    if (!startLoc || !endLoc) {
      throw new Error(`unable to find locations for range: [${start}, ${end})`);
    }
    const displayStartLine = Math.max(0, startLoc.line - 2);
    const displayEndLine = endLoc.line + 2;
    const rows = [];
    for (let line = displayStartLine; line <= displayEndLine; line++) {
      const startOfLine = lineMap.indexForLocation({ line, column: 0 });
      let endOfLine = lineMap.indexForLocation({ line: line + 1, column: 0 });
      if (startOfLine === null) {
        break;
      }
      if (endOfLine === null) {
        endOfLine = source.length;
      }
      const lineSource = trimRight(source.slice(startOfLine, endOfLine));
      if (startLoc.line !== endLoc.line) {
        if (line >= startLoc.line && line <= endLoc.line) {
          rows.push([`>`, `${line + 1} |`, lineSource]);
        } else {
          rows.push([``, `${line + 1} |`, lineSource]);
        }
      } else if (line === startLoc.line) {
        const highlightLength = Math.max(endLoc.column - startLoc.column, 1);
        rows.push([`>`, `${line + 1} |`, lineSource], [``, `|`, " ".repeat(startLoc.column) + "^".repeat(highlightLength)]);
      } else {
        rows.push([``, `${line + 1} |`, lineSource]);
      }
    }
    const columns = [
      { id: "marker", align: "right" },
      { id: "line", align: "right" },
      { id: "source", align: "left" }
    ];
    return `${message}
${printTable({ rows, columns })}`;
  }
};
function trimRight(string) {
  return string.replace(/\s+$/, "");
}

// src/stages/TransformCoffeeScriptStage.ts
var TransformCoffeeScriptStage = class {
  constructor(ast, context, editor, options) {
    this.ast = ast;
    this.context = context;
    this.editor = editor;
    this.options = options;
    this.root = null;
    this.patchers = [];
    this.suggestions = [];
  }
  static run(content, options) {
    const log = logger(this.name);
    log(content);
    const context = DecaffeinateContext.create(content, Boolean(options.useCS2));
    const editor = new MagicString2(content);
    const stage = new this(context.programNode, context, editor, options);
    const patcher = stage.build();
    patcher.patch();
    return {
      code: editor.toString(),
      suggestions: stage.suggestions
    };
  }
  patcherConstructorForNode(_node) {
    return null;
  }
  build() {
    this.root = this.patcherForNode(this.ast);
    this.patchers.forEach((patcher) => patcher.initialize());
    return this.root;
  }
  patcherForNode(node, parent = null, property = null) {
    let constructor = this._patcherConstructorForNode(node);
    if (parent) {
      const override = parent.patcherClassForChildNode(node, notNull(property));
      if (override) {
        constructor = override;
      }
    }
    const children = node.getChildNames().map((name) => {
      const child = node[name];
      if (!child) {
        return null;
      } else if (Array.isArray(child)) {
        return child.map((item) => item ? this.patcherForNode(item, constructor, name) : null);
      } else {
        return this.patcherForNode(child, constructor, name);
      }
    });
    const patcherContext = {
      node,
      context: this.context,
      editor: this.editor,
      options: this.options,
      addSuggestion: (suggestion) => {
        this.suggestions.push(suggestion);
      }
    };
    const patcher = new constructor(patcherContext, ...children);
    this.patchers.push(patcher);
    this.associateParent(patcher, children);
    return patcher;
  }
  associateParent(parent, child) {
    if (Array.isArray(child)) {
      child.forEach((item) => this.associateParent(parent, item));
    } else if (child) {
      child.parent = parent;
    }
  }
  _patcherConstructorForNode(node) {
    const constructor = this.patcherConstructorForNode(node);
    if (constructor === null) {
      const props = node.getChildNames();
      throw new PatchError(`no patcher available for node type: ${node.type}${props.length ? ` (props: ${props.join(", ")})` : ""}`, this.context.source, node.start, node.end);
    }
    return constructor.patcherClassOverrideForNode(node) || constructor;
  }
};

// src/patchers/NodePatcher.ts
import assert from "assert";
import { SourceType as SourceType2 } from "coffee-lex";
import { FunctionApplication, NewOp, SoakedFunctionApplication } from "decaffeinate-parser";

// src/suggestions.ts
var FIX_INVALID_CONSTRUCTOR = {
  suggestionCode: "DS002",
  message: "Fix invalid constructor"
};
var REMOVE_ARRAY_FROM = {
  suggestionCode: "DS101",
  message: "Remove unnecessary use of Array.from"
};
var CLEAN_UP_IMPLICIT_RETURNS = {
  suggestionCode: "DS102",
  message: "Remove unnecessary code created because of implicit returns"
};
var REMOVE_GUARD = {
  suggestionCode: "DS103",
  message: "Rewrite code to no longer use __guard__, or convert again using --optional-chaining"
};
var AVOID_INLINE_ASSIGNMENTS = {
  suggestionCode: "DS104",
  message: "Avoid inline assignments"
};
var SIMPLIFY_COMPLEX_ASSIGNMENTS = {
  suggestionCode: "DS201",
  message: "Simplify complex destructure assignments"
};
var SIMPLIFY_DYNAMIC_RANGE_LOOPS = {
  suggestionCode: "DS202",
  message: "Simplify dynamic range loops"
};
var CLEAN_UP_FOR_OWN_LOOPS = {
  suggestionCode: "DS203",
  message: "Remove `|| {}` from converted for-own loops"
};
var FIX_INCLUDES_EVALUATION_ORDER = {
  suggestionCode: "DS204",
  message: "Change includes calls to have a more natural evaluation order"
};
var AVOID_IIFES = {
  suggestionCode: "DS205",
  message: "Consider reworking code to avoid use of IIFEs"
};
var AVOID_INITCLASS = {
  suggestionCode: "DS206",
  message: "Consider reworking classes to avoid initClass"
};
var SHORTEN_NULL_CHECKS = {
  suggestionCode: "DS207",
  message: "Consider shorter variations of null checks"
};
var AVOID_TOP_LEVEL_THIS = {
  suggestionCode: "DS208",
  message: "Avoid top-level this"
};
var AVOID_TOP_LEVEL_RETURN = {
  suggestionCode: "DS209",
  message: "Avoid top-level return"
};
function mergeSuggestions(suggestions) {
  const suggestionsByCode = {};
  for (const suggestion of suggestions) {
    suggestionsByCode[suggestion.suggestionCode] = suggestion;
  }
  return Object.keys(suggestionsByCode).sort().map((code) => suggestionsByCode[code]);
}
function prependSuggestionComment(code, suggestions) {
  if (suggestions.length === 0) {
    return code;
  }
  const commentLines = [
    "/*",
    " * decaffeinate suggestions:",
    ...suggestions.map(({ suggestionCode, message }) => ` * ${suggestionCode}: ${message}`),
    " * Full docs: https://github.com/decaffeinate/decaffeinate/blob/main/docs/suggestions.md",
    " */"
  ];
  const codeLines = code.split("\n");
  if (codeLines[0].startsWith("#!")) {
    return [codeLines[0], ...commentLines, ...codeLines.slice(1)].join("\n");
  } else {
    return [...commentLines, ...codeLines].join("\n");
  }
}

// src/utils/determineIndent.ts
import detectIndent from "detect-indent";
var DEFAULT_INDENT = "  ";
function determineIndent(source) {
  const indent = detectIndent(source);
  if (indent.type === "space" && indent.amount % 2 === 1) {
    return DEFAULT_INDENT;
  }
  return indent.indent || DEFAULT_INDENT;
}

// src/utils/getStartOfLine.ts
function getStartOfLine(source, offset) {
  const lfIndex = source.lastIndexOf("\n", offset - 1);
  if (lfIndex < 0) {
    return 0;
  }
  return lfIndex + 1;
}

// src/utils/getIndent.ts
function getIndent(source, offset) {
  const startOfLine = getStartOfLine(source, offset);
  let indentOffset = startOfLine;
  let indentCharacter;
  switch (source[indentOffset]) {
    case " ":
    case "	":
      indentCharacter = source[indentOffset];
      break;
    default:
      return "";
  }
  while (source[indentOffset] === indentCharacter) {
    indentOffset++;
  }
  return source.slice(startOfLine, indentOffset);
}

// src/utils/adjustIndent.ts
function adjustIndent(source, offset, adjustment) {
  let currentIndent = getIndent(source, offset);
  const determinedIndent = determineIndent(source);
  if (adjustment > 0) {
    while (adjustment--) {
      currentIndent += determinedIndent;
    }
  } else if (adjustment < 0) {
    currentIndent = currentIndent.slice(determinedIndent.length * -adjustment);
  }
  return currentIndent;
}

// src/utils/referencesArguments.ts
import { Identifier as Identifier3 } from "decaffeinate-parser";

// src/utils/containsDescendant.ts
import { traverse as traverse2 } from "decaffeinate-parser";
function containsDescendant(node, predicate, { shouldStopTraversal = () => false } = {}) {
  let found = false;
  traverse2(node, (childNode) => {
    if (found) {
      return false;
    }
    if (predicate(childNode)) {
      found = true;
      return false;
    }
    if (shouldStopTraversal(childNode)) {
      return false;
    }
    return true;
  });
  return found;
}

// src/utils/types.ts
import { SourceType } from "coffee-lex";
function isFunction(node, allowBound = true) {
  return node.type === "Function" || node.type === "GeneratorFunction" || allowBound && (node.type === "BoundFunction" || node.type === "BoundGeneratorFunction");
}
var NON_SEMANTIC_SOURCE_TOKEN_TYPES = [SourceType.COMMENT, SourceType.HERECOMMENT, SourceType.NEWLINE];
function isSemanticToken(token) {
  return NON_SEMANTIC_SOURCE_TOKEN_TYPES.indexOf(token.type) < 0;
}

// src/utils/referencesArguments.ts
function referencesArguments(node) {
  return containsDescendant(node, (child) => child instanceof Identifier3 && child.data === "arguments", {
    shouldStopTraversal: (child) => child !== node && isFunction(child)
  });
}

// src/patchers/NodePatcher.ts
var NodePatcher = class {
  constructor({ node, context, editor, options, addSuggestion }) {
    this.adjustedIndentLevel = 0;
    this._assignee = false;
    this._containsYield = false;
    this._containsAwait = false;
    this._deferredSuffix = "";
    this._expression = false;
    this._hadUnparenthesizedNegation = false;
    this._implicitlyReturns = false;
    this._repeatableOptions = null;
    this._repeatCode = null;
    this._returns = false;
    this.log = logger(this.constructor.name);
    this.node = node;
    this.context = context;
    this.editor = editor;
    this.options = options;
    this.addSuggestion = addSuggestion;
    this.withPrettyErrors(() => this.setupLocationInformation());
  }
  static patcherClassForChildNode(_node, _property) {
    return null;
  }
  static patcherClassOverrideForNode(_node) {
    return null;
  }
  setupLocationInformation() {
    const { node, context } = this;
    this.contentStart = node.start;
    this.contentEnd = node.end;
    if (this.shouldTrimContentRange()) {
      this.trimContentRange();
    }
    const tokens = context.sourceTokens;
    const firstSourceTokenIndex = tokens.indexOfTokenStartingAtSourceIndex(this.contentStart);
    const lastSourceTokenIndex = tokens.indexOfTokenEndingAtSourceIndex(this.contentEnd);
    if (!firstSourceTokenIndex || !lastSourceTokenIndex) {
      if (node.type === "Program") {
        return;
      }
      throw this.error(`cannot find first or last token in ${node.type} node`);
    }
    this.contentStartTokenIndex = firstSourceTokenIndex;
    this.contentEndTokenIndex = lastSourceTokenIndex;
    let outerStartTokenIndex = firstSourceTokenIndex;
    let outerEndTokenIndex = lastSourceTokenIndex;
    let innerStartTokenIndex = firstSourceTokenIndex;
    let innerEndTokenIndex = lastSourceTokenIndex;
    for (; ; ) {
      const previousSurroundingTokenIndex = tokens.lastIndexOfTokenMatchingPredicate(isSemanticToken, outerStartTokenIndex.previous());
      const nextSurroundingTokenIndex = tokens.indexOfTokenMatchingPredicate(isSemanticToken, outerEndTokenIndex.next());
      if (!previousSurroundingTokenIndex || !nextSurroundingTokenIndex) {
        break;
      }
      const previousSurroundingToken = tokens.tokenAtIndex(previousSurroundingTokenIndex);
      const nextSurroundingToken = tokens.tokenAtIndex(nextSurroundingTokenIndex);
      if (!previousSurroundingToken || previousSurroundingToken.type !== SourceType2.LPAREN && previousSurroundingToken.type !== SourceType2.CALL_START) {
        break;
      }
      if (!nextSurroundingToken || nextSurroundingToken.type !== SourceType2.RPAREN && nextSurroundingToken.type !== SourceType2.CALL_END) {
        break;
      }
      if (innerStartTokenIndex === firstSourceTokenIndex) {
        innerStartTokenIndex = previousSurroundingTokenIndex;
      }
      if (innerEndTokenIndex === lastSourceTokenIndex) {
        innerEndTokenIndex = nextSurroundingTokenIndex;
      }
      outerStartTokenIndex = previousSurroundingTokenIndex;
      outerEndTokenIndex = nextSurroundingTokenIndex;
    }
    this.innerStartTokenIndex = innerStartTokenIndex;
    this.innerEndTokenIndex = innerEndTokenIndex;
    this.outerStartTokenIndex = outerStartTokenIndex;
    this.outerEndTokenIndex = outerEndTokenIndex;
    if (innerStartTokenIndex === firstSourceTokenIndex) {
      this.innerStart = this.contentStart;
    } else {
      this.innerStart = notNull(tokens.tokenAtIndex(innerStartTokenIndex)).end;
    }
    if (innerEndTokenIndex === lastSourceTokenIndex) {
      this.innerEnd = this.contentEnd;
    } else {
      this.innerEnd = notNull(tokens.tokenAtIndex(innerEndTokenIndex)).start;
    }
    this.outerStart = notNull(tokens.tokenAtIndex(outerStartTokenIndex)).start;
    this.outerEnd = notNull(tokens.tokenAtIndex(outerEndTokenIndex)).end;
  }
  trimContentRange() {
    const context = this.context;
    for (; ; ) {
      const startChar = context.source[this.contentStart];
      if (startChar === " " || startChar === "	") {
        this.contentStart++;
      } else {
        break;
      }
    }
    for (; ; ) {
      const lastChar = context.source[this.contentEnd - 1];
      if (lastChar === " " || lastChar === "	") {
        this.contentEnd--;
      } else {
        break;
      }
    }
  }
  shouldTrimContentRange() {
    return false;
  }
  initialize() {
  }
  patch(options = {}) {
    this.withPrettyErrors(() => {
      if (this._repeatableOptions !== null) {
        this._repeatCode = this.patchAsRepeatableExpression(this._repeatableOptions, options);
      } else if (this.forcedToPatchAsExpression()) {
        this.patchAsForcedExpression(options);
        this.commitDeferredSuffix();
      } else if (this.willPatchAsExpression()) {
        this.patchAsExpression(options);
        this.commitDeferredSuffix();
      } else {
        this.patchAsStatement(options);
        this.commitDeferredSuffix();
      }
    });
  }
  patchRepeatable(repeatableOptions = {}) {
    this.setRequiresRepeatableExpression(repeatableOptions);
    this.patch();
    return this.getRepeatCode();
  }
  patchAndGetCode(options = {}) {
    return this.captureCodeForPatchOperation(() => this.patch(options));
  }
  captureCodeForPatchOperation(patchFn) {
    let sliceStart = this.contentStart > 0 ? this.contentStart - 1 : 0;
    let beforeCode = null;
    while (beforeCode === null) {
      try {
        beforeCode = this.slice(sliceStart, this.contentStart);
      } catch (e) {
        sliceStart -= 1;
        if (sliceStart < 0) {
          throw this.error("Could not find a valid index to slice for patch operation.");
        }
      }
    }
    patchFn();
    const code = this.slice(sliceStart, this.contentEnd);
    let startIndex = 0;
    while (startIndex < beforeCode.length && startIndex < code.length && beforeCode[startIndex] === code[startIndex]) {
      startIndex++;
    }
    return code.substr(startIndex);
  }
  withPrettyErrors(body) {
    try {
      body();
    } catch (err) {
      assert(err instanceof Error);
      if (!PatchError.detect(err)) {
        throw this.error(err.message, this.contentStart, this.contentEnd, err);
      } else {
        throw err;
      }
    }
  }
  patchAsRepeatableExpression(repeatableOptions = {}, patchOptions = {}) {
    if (this.isRepeatable() && !repeatableOptions.forceRepeat) {
      return this.captureCodeForPatchOperation(() => {
        this.patchAsForcedExpression(patchOptions);
        this.commitDeferredSuffix();
      });
    } else {
      this.addSuggestion(AVOID_INLINE_ASSIGNMENTS);
      if (repeatableOptions.parens) {
        this.insert(this.innerStart, "(");
      }
      const ref = this.claimFreeBinding(repeatableOptions.ref);
      this.insert(this.innerStart, `${ref} = `);
      this.patchAsForcedExpression(patchOptions);
      this.commitDeferredSuffix();
      if (repeatableOptions.parens) {
        this.insert(this.innerEnd, ")");
      }
      return ref;
    }
  }
  patchAsExpression(_options = {}) {
    throw this.error(`'patchAsExpression' must be overridden in subclasses`);
  }
  patchAsStatement(options = {}) {
    const addParens = this.statementShouldAddParens();
    if (addParens) {
      this.insert(this.outerStart, "(");
    }
    this.patchAsExpression(options);
    if (addParens) {
      this.insert(this.outerEnd, ")");
    }
  }
  patchAsForcedExpression(options = {}) {
    this.patchAsExpression(options);
  }
  insert(index, content) {
    if (typeof index !== "number") {
      throw new Error(`cannot insert ${JSON.stringify(content)} at non-numeric index ${index}`);
    }
    this.log("INSERT", index, JSON.stringify(content), "BEFORE", JSON.stringify(this.context.source.slice(index, index + 8)));
    this.adjustBoundsToInclude(index);
    this.editor.appendLeft(index, content);
  }
  prependLeft(index, content) {
    if (typeof index !== "number") {
      throw new Error(`cannot insert ${JSON.stringify(content)} at non-numeric index ${index}`);
    }
    this.log("PREPEND LEFT", index, JSON.stringify(content), "BEFORE", JSON.stringify(this.context.source.slice(index, index + 8)));
    this.adjustBoundsToInclude(index);
    this.editor.prependLeft(index, content);
  }
  allowPatchingOuterBounds() {
    return false;
  }
  getEditingBounds() {
    let boundingPatcher = this.getBoundingPatcher();
    if (boundingPatcher.parent && (this.isNodeFunctionApplication(boundingPatcher.parent.node) || boundingPatcher.parent.node.type === "ArrayInitialiser")) {
      boundingPatcher = boundingPatcher.parent;
    }
    if (this.allowPatchingOuterBounds()) {
      return [boundingPatcher.outerStart, boundingPatcher.outerEnd];
    } else {
      return [boundingPatcher.innerStart, boundingPatcher.innerEnd];
    }
  }
  isIndexEditable(index) {
    const [start, end] = this.getEditingBounds();
    return index >= start && index <= end;
  }
  assertEditableIndex(index) {
    if (!this.isIndexEditable(index)) {
      const [start, end] = this.getEditingBounds();
      throw this.error(`cannot edit index ${index} because it is not editable (i.e. outside [${start}, ${end}))`, start, end);
    }
  }
  adjustBoundsToInclude(index) {
    this.assertEditableIndex(index);
    if (index < this.innerStart) {
      this.log("Moving `innerStart` from", this.innerStart, "to", index);
      this.innerStart = index;
    }
    if (index > this.innerEnd) {
      this.log("Moving `innerEnd` from", this.innerEnd, "to", index);
      this.innerEnd = index;
    }
    if (index < this.outerStart) {
      this.log("Moving `outerStart` from", this.outerStart, "to", index);
      this.outerStart = index;
    }
    if (index > this.outerEnd) {
      this.log("Moving `outerEnd` from", this.outerEnd, "to", index);
      this.outerEnd = index;
    }
    if (this.parent) {
      this.parent.adjustBoundsToInclude(index);
    }
  }
  overwrite(start, end, content) {
    if (typeof start !== "number" || typeof end !== "number") {
      throw new Error(`cannot overwrite non-numeric range [${start}, ${end}) with ${JSON.stringify(content)}`);
    }
    this.log("OVERWRITE", `[${start}, ${end})`, JSON.stringify(this.context.source.slice(start, end)), "\u2192", JSON.stringify(content));
    this.editor.overwrite(start, end, content);
  }
  remove(start, end) {
    if (typeof start !== "number" || typeof end !== "number") {
      throw new Error(`cannot remove non-numeric range [${start}, ${end})`);
    }
    this.log("REMOVE", `[${start}, ${end})`, JSON.stringify(this.context.source.slice(start, end)));
    this.editor.remove(start, end);
  }
  move(start, end, index) {
    if (typeof start !== "number" || typeof end !== "number") {
      throw this.error(`cannot remove non-numeric range [${start}, ${end})`);
    }
    if (typeof index !== "number") {
      throw this.error(`cannot move to non-numeric index: ${index}`);
    }
    this.log("MOVE", `[${start}, ${end}) \u2192 ${index}`, JSON.stringify(this.context.source.slice(start, end)), "BEFORE", JSON.stringify(this.context.source.slice(index, index + 8)));
    this.editor.move(start, end, index);
  }
  slice(start, end) {
    if (end === 0) {
      return "";
    }
    return this.editor.slice(start, end);
  }
  startsWith(string) {
    return this.context.source.slice(this.contentStart, this.contentStart + string.length) === string;
  }
  endsWith(string) {
    return this.context.source.slice(this.contentEnd - string.length, this.contentEnd) === string;
  }
  setRequiresExpression() {
    this.setExpression(true);
  }
  setExpression(force = false) {
    if (force) {
      if (!this.canPatchAsExpression()) {
        throw this.error(`cannot represent ${this.node.type} as an expression`);
      }
    } else if (!this.prefersToPatchAsExpression()) {
      return false;
    }
    this._expression = true;
    return true;
  }
  prefersToPatchAsExpression() {
    return this.canPatchAsExpression();
  }
  canPatchAsExpression() {
    return true;
  }
  willPatchAsExpression() {
    return this._expression;
  }
  forcedToPatchAsExpression() {
    return this.willPatchAsExpression() && !this.prefersToPatchAsExpression();
  }
  setAssignee() {
    this._assignee = true;
  }
  isAssignee() {
    return this._assignee;
  }
  implicitlyReturns() {
    return this._implicitlyReturns || false;
  }
  setImplicitlyReturns() {
    this._implicitlyReturns = true;
  }
  implicitReturnPatcher() {
    if (this.canHandleImplicitReturn()) {
      return this;
    } else {
      return notNull(this.parent).implicitReturnPatcher();
    }
  }
  canHandleImplicitReturn() {
    return false;
  }
  implicitReturnWillBreak() {
    return true;
  }
  patchImplicitReturnStart(patcher) {
    if (patcher.node.type === "Break" || patcher.node.type === "Continue") {
      if (patcher.isSurroundedByParentheses()) {
        this.remove(patcher.outerStart, patcher.innerStart);
        this.remove(patcher.innerEnd, patcher.outerEnd);
      }
      return;
    }
    if (isFunction(this.node) && this.isMultiline()) {
      this.addSuggestion(CLEAN_UP_IMPLICIT_RETURNS);
    }
    patcher.setRequiresExpression();
    this.insert(patcher.outerStart, "return ");
  }
  getEmptyImplicitReturnCode() {
    return null;
  }
  patchImplicitReturnEnd(_patcher) {
  }
  explicitlyReturns() {
    return this._returns || false;
  }
  setExplicitlyReturns() {
    this._returns = true;
    if (this.parent) {
      this.parent.setExplicitlyReturns();
    }
  }
  appendDeferredSuffix(suffix) {
    this._deferredSuffix += suffix;
  }
  commitDeferredSuffix() {
    if (this._deferredSuffix) {
      this.insert(this.innerEnd, this._deferredSuffix);
    }
  }
  statementNeedsSemicolon() {
    return true;
  }
  statementNeedsParens() {
    return false;
  }
  statementShouldAddParens() {
    return this.statementNeedsParens() && !this.isSurroundedByParentheses();
  }
  getProgramSourceTokens() {
    return this.context.sourceTokens;
  }
  indexOfSourceTokenStartingAtSourceIndex(index) {
    return this.getProgramSourceTokens().indexOfTokenStartingAtSourceIndex(index);
  }
  indexOfSourceTokenBetweenPatchersMatching(left, right, predicate) {
    return this.indexOfSourceTokenBetweenSourceIndicesMatching(left.outerEnd, right.outerStart, predicate);
  }
  indexOfSourceTokenBetweenSourceIndicesMatching(left, right, predicate) {
    const tokenList = this.getProgramSourceTokens();
    return tokenList.indexOfTokenMatchingPredicate((token) => {
      return token.start >= left && token.start <= right && predicate(token);
    }, tokenList.indexOfTokenNearSourceIndex(left), tokenList.indexOfTokenNearSourceIndex(right).next());
  }
  sourceTokenAtIndex(index) {
    return this.getProgramSourceTokens().tokenAtIndex(index);
  }
  sourceOfToken(token) {
    return this.context.source.slice(token.start, token.end);
  }
  firstToken() {
    const token = this.sourceTokenAtIndex(this.contentStartTokenIndex);
    if (!token) {
      throw this.error("Expected to find a first token for node.");
    }
    return token;
  }
  lastToken() {
    const token = this.sourceTokenAtIndex(this.contentEndTokenIndex);
    if (!token) {
      throw this.error("Expected to find a last token for node.");
    }
    return token;
  }
  nextSemanticToken() {
    return this.getFirstSemanticToken(this.contentEnd, this.context.source.length);
  }
  getOriginalSource() {
    return this.context.source.slice(this.contentStart, this.contentEnd);
  }
  isMultiline() {
    return /\n/.test(this.getOriginalSource());
  }
  getPatchedSource() {
    return this.slice(this.contentStart, this.contentEnd);
  }
  indexOfSourceTokenAfterSourceTokenIndex(start, type, predicate = isSemanticToken) {
    const index = this.getProgramSourceTokens().indexOfTokenMatchingPredicate(predicate, start.next());
    if (!index) {
      return null;
    }
    const token = this.sourceTokenAtIndex(index);
    if (!token || token.type !== type) {
      return null;
    }
    return index;
  }
  hasSourceTokenAfter(type, predicate = isSemanticToken) {
    return this.indexOfSourceTokenAfterSourceTokenIndex(this.outerEndTokenIndex, type, predicate) !== null;
  }
  isSurroundedByParentheses() {
    if (this.contentStart === this.outerStart && this.contentEnd === this.outerEnd) {
      return false;
    }
    const beforeToken = this.sourceTokenAtIndex(this.outerStartTokenIndex);
    const afterToken = this.sourceTokenAtIndex(this.outerEndTokenIndex);
    if (!beforeToken || !afterToken) {
      return false;
    }
    let leftTokenType = SourceType2.LPAREN;
    let rightTokenType = SourceType2.RPAREN;
    if (beforeToken.type === SourceType2.LPAREN && afterToken.type === SourceType2.RPAREN) {
    } else if (beforeToken.type === SourceType2.CALL_START && afterToken.type === SourceType2.CALL_END) {
      leftTokenType = SourceType2.CALL_START;
      rightTokenType = SourceType2.CALL_END;
    } else {
      return false;
    }
    const parenRange = this.getProgramSourceTokens().rangeOfMatchingTokensContainingTokenIndex(leftTokenType, rightTokenType, this.outerStartTokenIndex);
    if (!parenRange) {
      return false;
    }
    const rparenIndex = parenRange[1].previous();
    const rparen = this.sourceTokenAtIndex(notNull(rparenIndex));
    return rparen === afterToken;
  }
  surroundInParens() {
    if (!this.isSurroundedByParentheses()) {
      this.insert(this.outerStart, "(");
      this.insert(this.outerEnd, ")");
    }
  }
  getBoundingPatcher() {
    if (this.isSurroundedByParentheses()) {
      return this;
    } else if (this.parent) {
      if (this.isNodeFunctionApplication(this.parent.node) && this.parent.node.arguments.some((arg) => arg === this.node)) {
        return this;
      } else if (this.parent.node.type === "ArrayInitialiser") {
        return this;
      } else if (this.parent.node.type === "ObjectInitialiser") {
        return this;
      }
      return this.parent.getBoundingPatcher();
    } else {
      return this;
    }
  }
  isNodeFunctionApplication(node) {
    return node instanceof FunctionApplication || node instanceof SoakedFunctionApplication || node instanceof NewOp;
  }
  canHandleNegationInternally() {
    return false;
  }
  negate() {
    this.insert(this.contentStart, "!");
    this._hadUnparenthesizedNegation = true;
  }
  hadUnparenthesizedNegation() {
    return this._hadUnparenthesizedNegation;
  }
  getScope() {
    return this.context.getScope(this.node);
  }
  getIndent(offset = 0) {
    return adjustIndent(this.context.source, this.contentStart, this.getAdjustedIndentLevel() + offset);
  }
  setIndent(indentStr) {
    const currentIndent = this.getIndent();
    const indentLength = this.getProgramIndentString().length;
    const currentIndentLevel = currentIndent.length / indentLength;
    const desiredIndentLevel = indentStr.length / indentLength;
    this.indent(desiredIndentLevel - currentIndentLevel);
  }
  getAdjustedIndentLevel() {
    return this.adjustedIndentLevel + (this.parent ? this.parent.getAdjustedIndentLevel() : 0);
  }
  getProgramIndentString() {
    return notNull(this.parent).getProgramIndentString();
  }
  indent(offset = 1, { skipFirstLine = false } = {}) {
    if (offset === 0) {
      return;
    }
    this.adjustedIndentLevel += offset;
    const indentString = this.getProgramIndentString();
    const indentToChange = indentString.repeat(Math.abs(offset));
    let start = this.outerStart;
    const end = this.outerEnd;
    const { source } = this.context;
    if (skipFirstLine || !this.isFirstNodeInLine()) {
      while (start < end && source[start] !== "\n") {
        start++;
      }
    }
    let hasIndentedThisLine = false;
    for (let i = start; i < end; i++) {
      switch (source[i]) {
        case "\n":
          hasIndentedThisLine = false;
          break;
        case " ":
        case "	":
          break;
        default:
          if (!hasIndentedThisLine) {
            if (offset > 0) {
              this.insert(i, indentToChange);
            } else if (source.slice(i - indentToChange.length, i) === indentToChange) {
              this.remove(i - indentToChange.length, i);
            } else {
              this.log("Warning: Ignoring an unindent operation because the line did not start with the proper indentation.");
            }
            hasIndentedThisLine = true;
          }
          break;
      }
    }
  }
  isFirstNodeInLine(startingPoint = this.outerStart) {
    const { source } = this.context;
    for (let i = startingPoint - 1; i >= 0 && source[i] !== "\n"; i--) {
      if (source[i] !== "	" && source[i] !== " ") {
        return false;
      }
    }
    return true;
  }
  getEndOfLine() {
    const { source } = this.context;
    for (let i = this.outerEnd - "\n".length; i < source.length; i++) {
      if (source[i] === "\n") {
        return i;
      }
    }
    return source.length;
  }
  appendLineAfter(content, indentOffset = 0) {
    const boundingPatcher = this.getBoundingPatcher();
    const endOfLine = this.getEndOfLine();
    const nextToken = this.nextSemanticToken();
    let insertPoint = Math.min(Math.min(endOfLine, boundingPatcher.innerEnd));
    if (nextToken) {
      insertPoint = Math.min(insertPoint, nextToken.start);
    }
    this.insert(insertPoint, `
${this.getIndent(indentOffset)}${content}`);
  }
  error(message, start = this.contentStart, end = this.contentEnd, error = null) {
    const patcherError = new PatchError(message, this.context.source, start, end);
    if (error) {
      patcherError.stack = error.stack;
    }
    return patcherError;
  }
  registerHelper(name, code) {
    return notNull(this.parent).registerHelper(name, code);
  }
  isPure() {
    return this.isRepeatable();
  }
  isRepeatable() {
    return false;
  }
  setRequiresRepeatableExpression(repeatableOptions = {}) {
    this._repeatableOptions = repeatableOptions;
  }
  getRepeatableOptions() {
    return this._repeatableOptions;
  }
  getRepeatCode() {
    if (this._repeatCode === null) {
      throw new Error("Must patch as a repeatable expression to access repeat code.");
    }
    return this._repeatCode;
  }
  overrideRepeatCode(repeatCode) {
    this._repeatCode = repeatCode;
  }
  claimFreeBinding(ref = null) {
    return this.getScope().claimFreeBinding(this.node, ref);
  }
  allCodePathsPresent() {
    return true;
  }
  getFirstSemanticToken(from = this.contentStart, to = notNull(this.parent).contentEnd) {
    const nextSemanticIdx = this.indexOfSourceTokenBetweenSourceIndicesMatching(from, to, isSemanticToken);
    return nextSemanticIdx && this.sourceTokenAtIndex(nextSemanticIdx);
  }
  mayBeUnboundReference() {
    return false;
  }
  patchInIIFE(innerPatchFn) {
    this.addSuggestion(AVOID_IIFES);
    if (this.containsYield()) {
      this.insert(this.innerStart, "yield* (function*() {");
    } else if (this.containsAwait()) {
      this.insert(this.innerStart, "await (async () => {");
    } else {
      this.insert(this.innerStart, "(() => {");
    }
    innerPatchFn();
    if (this.containsYield()) {
      if (referencesArguments(this.node)) {
        this.insert(this.innerEnd, "}).apply(this, arguments)");
      } else {
        this.insert(this.innerEnd, "}).call(this)");
      }
    } else if (this.containsAwait()) {
      this.insert(this.innerEnd, "})()");
    } else {
      this.insert(this.innerEnd, "})()");
    }
  }
  yields() {
    this._containsYield = true;
    if (this.parent && !isFunction(this.parent.node)) {
      this.parent.yields();
    }
  }
  containsYield() {
    return this._containsYield;
  }
  awaits() {
    this._containsAwait = true;
    if (this.parent && !isFunction(this.parent.node)) {
      this.parent.awaits();
    }
  }
  containsAwait() {
    return this._containsAwait;
  }
};

// src/patchers/PassthroughPatcher.ts
var PassthroughPatcher = class extends NodePatcher {
  constructor(patcherContext, ...children) {
    super(patcherContext);
    this.children = children;
  }
  patchAsExpression() {
    this.children.forEach((child) => {
      if (Array.isArray(child)) {
        child.forEach((child2) => child2 && child2.patch());
      } else if (child) {
        child.patch();
      }
    });
  }
  isRepeatable() {
    return true;
  }
};

// src/stages/main/patchers/ArrayInitialiserPatcher.ts
import { SourceType as SourceType3 } from "coffee-lex";

// src/stages/main/patchers/ElisionPatcher.ts
var ElisionPatcher = class extends NodePatcher {
  patchAsExpression() {
  }
};

// src/stages/main/patchers/ExpansionPatcher.ts
var ExpansionPatcher = class extends NodePatcher {
  patchAsExpression() {
    throw this.error("Unexpected traversal of expansion node.");
  }
};

// src/stages/main/patchers/ArrayInitialiserPatcher.ts
var ArrayInitialiserPatcher = class extends NodePatcher {
  constructor(patcherContext, members) {
    super(patcherContext);
    this.members = members;
  }
  initialize() {
    this.members.forEach((member) => member.setRequiresExpression());
  }
  setAssignee() {
    this.members.forEach((member) => member.setAssignee());
    super.setAssignee();
  }
  patchAsExpression() {
    this.members.forEach((member, i, members) => {
      const isLast = i === members.length - 1;
      if (isLast && member instanceof ExpansionPatcher) {
        this.remove(members[i - 1].outerEnd, member.outerEnd);
        return;
      }
      const needsComma = !isLast && !member.hasSourceTokenAfter(SourceType3.COMMA) && !(member instanceof ElisionPatcher);
      member.patch();
      if (needsComma) {
        this.insert(member.outerEnd, ",");
      }
    });
  }
  isPure() {
    return this.members.every((member) => member.isPure());
  }
};

// src/utils/canPatchAssigneeToJavaScript.ts
import {
  ArrayInitialiser as ArrayInitialiser3,
  AssignOp as AssignOp2,
  DynamicMemberAccessOp,
  Elision,
  Expansion as Expansion2,
  Identifier as Identifier4,
  MemberAccessOp as MemberAccessOp2,
  ObjectInitialiser as ObjectInitialiser3,
  ObjectInitialiserMember as ObjectInitialiserMember2,
  ProtoMemberAccessOp,
  Rest as Rest2,
  SoakedDynamicMemberAccessOp,
  SoakedMemberAccessOp,
  SoakedProtoMemberAccessOp,
  Spread as Spread2
} from "decaffeinate-parser";
function canPatchAssigneeToJavaScript(node, options, isTopLevel = true) {
  if (node instanceof Identifier4 || node instanceof MemberAccessOp2 || node instanceof SoakedMemberAccessOp || node instanceof ProtoMemberAccessOp || node instanceof DynamicMemberAccessOp || node instanceof SoakedDynamicMemberAccessOp || node instanceof SoakedProtoMemberAccessOp || node instanceof Elision) {
    return true;
  }
  if (node instanceof ArrayInitialiser3) {
    if (!options.useCS2 && !isTopLevel) {
      return false;
    }
    if (node.members.length === 0) {
      return false;
    }
    return node.members.every((member, i) => {
      const isInFinalPosition = i === node.members.length - 1;
      if (isInFinalPosition && member instanceof Expansion2) {
        return true;
      }
      if (isInFinalPosition && (member instanceof Spread2 || member instanceof Rest2) && !(member.expression instanceof ObjectInitialiser3) && canPatchAssigneeToJavaScript(member.expression, options, false)) {
        return true;
      }
      return canPatchAssigneeToJavaScript(member, options, false);
    });
  }
  if (node instanceof ObjectInitialiser3) {
    if (!options.useCS2 && node.members.length === 0) {
      return false;
    }
    return node.members.every((member, i) => {
      const isInFinalPosition = i === node.members.length - 1;
      if (isInFinalPosition && (member instanceof Spread2 || member instanceof Rest2) && member.expression instanceof Identifier4) {
        return true;
      }
      return canPatchAssigneeToJavaScript(member, options, false);
    });
  }
  if (node instanceof ObjectInitialiserMember2) {
    return canPatchAssigneeToJavaScript(node.expression || node.key, options, false);
  }
  if (options.useCS2 && node instanceof AssignOp2 && !isTopLevel) {
    return canPatchAssigneeToJavaScript(node.assignee, options, false);
  }
  return false;
}

// src/utils/containsSuperCall.ts
import { BareSuperFunctionApplication, Class as Class2, Super } from "decaffeinate-parser";
function containsSuperCall(node) {
  return containsDescendant(node, (child) => child instanceof Super || child instanceof BareSuperFunctionApplication, {
    shouldStopTraversal: (child) => child instanceof Class2
  });
}

// src/stages/main/patchers/DynamicMemberAccessOpPatcher.ts
var DynamicMemberAccessOpPatcher = class extends NodePatcher {
  constructor(patcherContext, expression, indexingExpr) {
    super(patcherContext);
    this.expression = expression;
    this.indexingExpr = indexingExpr;
  }
  initialize() {
    this.expression.setRequiresExpression();
    this.indexingExpr.setRequiresExpression();
  }
  patchAsExpression() {
    this.expression.patch();
    this.indexingExpr.patch();
  }
  patchAsRepeatableExpression(repeatableOptions = {}, patchOptions = {}) {
    if (repeatableOptions.isForAssignment) {
      this.expression.setRequiresRepeatableExpression({ isForAssignment: true, parens: true, ref: "base" });
      this.indexingExpr.setRequiresRepeatableExpression({ ref: "name" });
      this.patchAsExpression();
      this.commitDeferredSuffix();
      return `${this.expression.getRepeatCode()}[${this.indexingExpr.getRepeatCode()}]`;
    } else {
      return super.patchAsRepeatableExpression(repeatableOptions, patchOptions);
    }
  }
  isRepeatable() {
    return this.expression.isRepeatable() && this.indexingExpr.isRepeatable();
  }
  statementNeedsParens() {
    return this.expression.statementShouldAddParens();
  }
};

// src/stages/main/patchers/FunctionPatcher.ts
import { SourceType as SourceType5 } from "coffee-lex";

// src/stages/main/patchers/FunctionApplicationPatcher.ts
import { SourceType as SourceType4 } from "coffee-lex";
var FunctionApplicationPatcher = class extends NodePatcher {
  constructor(patcherContext, fn, args) {
    super(patcherContext);
    this.fn = fn;
    this.args = args;
  }
  initialize() {
    this.fn.setRequiresExpression();
    this.args.forEach((arg) => arg.setRequiresExpression());
  }
  patchAsExpression({ fnNeedsParens = false } = {}) {
    const { args, outerEndTokenIndex } = this;
    if (fnNeedsParens) {
      this.insert(this.fn.outerStart, "(");
    }
    this.fn.patch({ skipParens: fnNeedsParens });
    if (fnNeedsParens) {
      this.insert(this.fn.outerEnd, ")");
    }
    args.forEach((arg, i) => {
      arg.patch();
      const isLast = i === args.length - 1;
      let commaTokenIndex = this.indexOfSourceTokenAfterSourceTokenIndex(arg.outerEndTokenIndex, SourceType4.COMMA, isSemanticToken);
      if (commaTokenIndex && commaTokenIndex.compare(outerEndTokenIndex) <= 0) {
        commaTokenIndex = null;
      }
      const commaToken = commaTokenIndex && this.sourceTokenAtIndex(commaTokenIndex);
      if (isLast && commaToken) {
        this.remove(arg.outerEnd, commaToken.end);
      } else if (!isLast && !commaToken) {
        this.insert(arg.outerEnd, ",");
      }
    });
  }
  statementNeedsParens() {
    return this.fn.statementShouldAddParens();
  }
};

// src/stages/main/patchers/FunctionPatcher.ts
var FunctionPatcher = class extends NodePatcher {
  constructor(patcherContext, parameters, body) {
    super(patcherContext);
    this._implicitReturnsDisabled = false;
    this.parameters = parameters;
    this.body = body;
  }
  initialize() {
    if (this.body && !this.implicitReturnsDisabled()) {
      this.body.setImplicitlyReturns();
    }
    this.parameters.forEach((param) => {
      param.setAssignee();
      param.setRequiresExpression();
    });
  }
  patchAsExpression({ method = false } = {}) {
    this.patchFunctionStart({ method });
    this.parameters.forEach((parameter, i) => {
      const isLast = i === this.parameters.length - 1;
      const needsComma = !isLast && !parameter.hasSourceTokenAfter(SourceType5.COMMA);
      parameter.patch();
      if (needsComma) {
        this.insert(parameter.outerEnd, ",");
      }
    });
    this.patchFunctionBody();
  }
  patchFunctionStart({ method = false }) {
    const arrow = this.getArrowToken();
    if (!method) {
      this.insert(this.contentStart, "function");
    }
    if (!this.hasParamStart()) {
      this.insert(this.contentStart, "() ");
    }
    this.overwrite(arrow.start, arrow.end, "{");
  }
  patchFunctionBody() {
    if (this.body) {
      if (this.isSurroundedByParentheses()) {
        this.body.patch({ leftBrace: false, rightBrace: false });
        this.insert(this.innerEnd, this.body.inline() ? " }" : "}");
      } else if (this.isEndOfFunctionCall()) {
        this.body.patch({ leftBrace: false, rightBrace: false });
        this.placeCloseBraceBeforeFunctionCallEnd();
      } else {
        this.body.patch({ leftBrace: false });
      }
    } else {
      this.insert(this.innerEnd, "}");
    }
  }
  isEndOfFunctionCall() {
    return this.parent instanceof FunctionApplicationPatcher && this.parent.args[this.parent.args.length - 1] === this;
  }
  placeCloseBraceBeforeFunctionCallEnd() {
    if (!this.body) {
      throw this.error("Expected non-null body.");
    }
    const closeParenIndex = notNull(this.parent).indexOfSourceTokenBetweenSourceIndicesMatching(this.contentEnd, notNull(this.parent).contentEnd, (token) => token.type === SourceType5.CALL_END || token.type === SourceType5.RPAREN);
    if (!closeParenIndex) {
      throw this.error("Expected to find close paren index after function call.");
    }
    const closeParen = this.sourceTokenAtIndex(closeParenIndex);
    if (!closeParen) {
      throw this.error("Expected to find close paren after function call.");
    }
    const shouldMoveCloseParen = !this.body.inline() && !this.slice(this.contentEnd, closeParen.start).includes("\n");
    if (shouldMoveCloseParen) {
      this.appendLineAfter("}", -1);
    } else {
      this.insert(closeParen.start, this.body.inline() ? " }" : "}");
    }
  }
  getArrowToken() {
    let arrowIndex = this.contentStartTokenIndex;
    if (this.hasParamStart()) {
      const parenRange = this.getProgramSourceTokens().rangeOfMatchingTokensContainingTokenIndex(SourceType5.LPAREN, SourceType5.RPAREN, this.contentStartTokenIndex);
      if (!parenRange) {
        throw this.error("Expected to find function paren range in function.");
      }
      const rparenIndex = parenRange[1].previous();
      if (!rparenIndex) {
        throw this.error("Expected to find rparen index in function.");
      }
      arrowIndex = notNull(this.indexOfSourceTokenAfterSourceTokenIndex(rparenIndex, SourceType5.FUNCTION));
    }
    const arrow = this.sourceTokenAtIndex(arrowIndex);
    if (!arrow) {
      throw this.error("Expected to find arrow token in function.");
    }
    const expectedArrowType = this.expectedArrowType();
    const actualArrowType = this.sourceOfToken(arrow);
    if (actualArrowType !== expectedArrowType) {
      throw this.error(`expected '${expectedArrowType}' but found ${actualArrowType}`, arrow.start, arrow.end);
    }
    return arrow;
  }
  expectedArrowType() {
    return "->";
  }
  hasParamStart() {
    return notNull(this.sourceTokenAtIndex(this.contentStartTokenIndex)).type === SourceType5.LPAREN;
  }
  canHandleImplicitReturn() {
    return true;
  }
  setExplicitlyReturns() {
  }
  disableImplicitReturns() {
    this._implicitReturnsDisabled = true;
  }
  implicitReturnsDisabled() {
    return this._implicitReturnsDisabled;
  }
  statementNeedsParens() {
    return true;
  }
};

// src/stages/main/patchers/MemberAccessOpPatcher.ts
import { SourceType as SourceType6 } from "coffee-lex";
var MemberAccessOpPatcher = class extends NodePatcher {
  constructor(patcherContext, expression, member) {
    super(patcherContext);
    this.expression = expression;
    this.member = member;
    this._skipImplicitDotCreation = false;
  }
  initialize() {
    this.expression.setRequiresExpression();
  }
  setSkipImplicitDotCreation() {
    this._skipImplicitDotCreation = true;
  }
  patchAsExpression() {
    if (this.lhsNeedsParens()) {
      this.insert(this.expression.outerStart, "(");
    }
    this.expression.patch();
    if (this.lhsNeedsParens()) {
      this.insert(this.expression.outerEnd, ")");
    }
    if (this.hasImplicitOperator() && !this._skipImplicitDotCreation) {
      this.insert(this.expression.outerEnd, ".");
    }
  }
  patchAsRepeatableExpression(repeatableOptions = {}, patchOptions = {}) {
    if (repeatableOptions.isForAssignment) {
      this.expression.setRequiresRepeatableExpression({ isForAssignment: true, parens: true, ref: "base" });
      this.patchAsExpression();
      this.commitDeferredSuffix();
      return `${this.expression.getRepeatCode()}.${this.getFullMemberName()}`;
    } else {
      return super.patchAsRepeatableExpression(repeatableOptions, patchOptions);
    }
  }
  hasImplicitOperator() {
    return !this.getMemberOperatorSourceToken();
  }
  getMemberOperatorSourceToken() {
    const dotIndex = this.indexOfSourceTokenBetweenSourceIndicesMatching(this.expression.outerEnd, this.outerEnd - 1, (token) => token.type === SourceType6.DOT);
    if (!dotIndex) {
      const firstIndex = this.contentStartTokenIndex;
      const firstToken = notNull(this.sourceTokenAtIndex(firstIndex));
      if (firstToken.type === SourceType6.AT) {
        return null;
      }
      throw this.error(`cannot find '.' in member access`);
    }
    return this.sourceTokenAtIndex(dotIndex);
  }
  getMemberName() {
    return this.node.member.data;
  }
  getFullMemberName() {
    return this.getMemberName();
  }
  getMemberNameSourceToken() {
    const tokens = this.context.sourceTokens;
    const index = tokens.lastIndexOfTokenMatchingPredicate((token) => token.type === SourceType6.IDENTIFIER, this.contentEndTokenIndex);
    if (!index || index.isBefore(this.contentStartTokenIndex)) {
      throw this.error(`unable to find member name token in access`);
    }
    return notNull(tokens.tokenAtIndex(index));
  }
  isRepeatable() {
    return this.expression.isRepeatable();
  }
  statementNeedsParens() {
    return this.expression.statementShouldAddParens();
  }
  lhsNeedsParens() {
    return this.expression.node.type === "Int";
  }
};

// src/utils/extractPrototypeAssignPatchers.ts
function extractPrototypeAssignPatchers(patcher) {
  if (!(patcher instanceof AssignOpPatcher) || !(patcher.expression instanceof FunctionPatcher) || !(patcher.assignee instanceof MemberAccessOpPatcher || patcher.assignee instanceof DynamicMemberAccessOpPatcher) || !(patcher.assignee.expression instanceof MemberAccessOpPatcher) || patcher.assignee.expression.member.node.data !== "prototype") {
    return null;
  }
  return {
    classRefPatcher: patcher.assignee.expression.expression,
    methodAccessPatcher: patcher.assignee
  };
}

// src/stages/main/patchers/IdentifierPatcher.ts
var IdentifierPatcher = class extends PassthroughPatcher {
  isRepeatable() {
    return true;
  }
  mayBeUnboundReference() {
    return !this.getScope().hasBinding(this.node.data);
  }
};

// src/stages/main/patchers/ObjectInitialiserMemberPatcher.ts
import { SourceType as SourceType15 } from "coffee-lex";
import { Identifier as Identifier6 } from "decaffeinate-parser";

// src/stages/main/patchers/ObjectBodyMemberPatcher.ts
import { ObjectInitialiserMember as ObjectInitialiserMember4 } from "decaffeinate-parser";

// src/stages/main/patchers/AsyncFunctionPatcher.ts
var AsyncFunctionPatcher = class extends FunctionPatcher {
  patchFunctionStart({ method = false }) {
    const arrow = this.getArrowToken();
    if (!method) {
      this.insert(this.contentStart, "async function");
    }
    if (!this.hasParamStart()) {
      this.insert(this.contentStart, "() ");
    }
    this.overwrite(arrow.start, arrow.end, "{");
  }
};

// src/stages/main/patchers/BoundFunctionPatcher.ts
import { AssignOp as AssignOp3 } from "decaffeinate-parser";

// src/utils/blockStartsWithObjectInitialiser.ts
import { ObjectInitialiser as ObjectInitialiser4 } from "decaffeinate-parser";

// src/stages/main/patchers/BlockPatcher.ts
import { SourceType as SourceType10 } from "coffee-lex";
import { traverse as traverse4 } from "decaffeinate-parser";

// src/patchers/SharedBlockPatcher.ts
import { SourceType as SourceType7 } from "coffee-lex";
var SharedBlockPatcher = class extends NodePatcher {
  constructor(patcherContext, statements) {
    super(patcherContext);
    this.shouldPatchInline = null;
    this.statements = statements;
  }
  insertStatementsAtIndex(statements, index) {
    const separator = this.inline() ? "; " : ";\n";
    if (index === this.statements.length) {
      const lastStatement = this.statements[this.statements.length - 1];
      const terminatorTokenIndex = this.context.sourceTokens.indexOfTokenMatchingPredicate((token) => token.type === SourceType7.NEWLINE || token.type === SourceType7.SEMICOLON, lastStatement.outerEndTokenIndex);
      let insertionPoint = terminatorTokenIndex ? notNull(this.sourceTokenAtIndex(terminatorTokenIndex)).start : lastStatement.outerEnd;
      insertionPoint = Math.min(insertionPoint, this.getBoundingPatcher().innerEnd);
      const indent = lastStatement.getIndent();
      statements.forEach((line) => {
        const sep = line.trim().startsWith("//") ? "\n" : separator;
        this.insert(insertionPoint, `${sep}${indent}${line}`);
      });
    } else {
      const statementToInsertBefore = this.statements[index];
      const insertionPoint = statementToInsertBefore.outerStart;
      const indent = statementToInsertBefore.getIndent();
      statements.forEach((line) => {
        const sep = line.trim().startsWith("//") ? "\n" : separator;
        this.insert(insertionPoint, `${line}${sep}${indent}`);
      });
    }
  }
  insertLineBefore(statement, indent = this.getIndent()) {
    if (this.inline()) {
      this.insert(this.outerStart, `${statement}; `);
    } else if (this.node.inline) {
      this.insert(this.outerStart, `${indent}${statement};
`);
    } else {
      let insertIndex = this.outerStart;
      while (insertIndex > 0 && this.context.source[insertIndex] !== "\n") {
        insertIndex--;
      }
      this.insert(insertIndex, `
${indent}${statement};`);
    }
  }
  insertLineAfter(statement, indent) {
    if (this.inline()) {
      this.insert(this.outerEnd, `; ${statement}`);
    } else {
      this.insert(this.outerEnd, `
${indent}${statement};`);
    }
  }
  inline() {
    if (this.shouldPatchInline !== null) {
      return this.shouldPatchInline;
    }
    return this.node.inline;
  }
};

// src/utils/countVariableUsages.ts
import { traverse as traverse3, Identifier as Identifier5, MemberAccessOp as MemberAccessOp3, ObjectInitialiserMember as ObjectInitialiserMember3 } from "decaffeinate-parser";
function countVariableUsages(node, name) {
  let numUsages = 0;
  traverse3(node, (child) => {
    if (!(child instanceof Identifier5) || child.data !== name) {
      return;
    }
    if (child.parentNode instanceof MemberAccessOp3 && child.parentNode.member === child) {
      return;
    }
    if (child.parentNode instanceof ObjectInitialiserMember3 && child.parentNode.key === child && child.parentNode.expression !== null && !child.parentNode.isComputed) {
      return;
    }
    numUsages += 1;
  });
  return numUsages;
}

// src/stages/main/patchers/ConditionalPatcher.ts
import { SourceType as SourceType8 } from "coffee-lex";

// src/utils/getEnclosingScopeBlock.ts
function getEnclosingScopeBlock(patcher) {
  let currentPatcher = patcher;
  while (currentPatcher) {
    if (currentPatcher instanceof BlockPatcher && notNull(currentPatcher.parent).node === patcher.getScope().containerNode) {
      return currentPatcher;
    }
    currentPatcher = currentPatcher.parent;
  }
  throw patcher.error("Expected to find enclosing scope block.");
}

// src/stages/main/patchers/ConditionalPatcher.ts
var ConditionalPatcher = class extends NodePatcher {
  constructor(patcherContext, condition, consequent, alternate) {
    super(patcherContext);
    this.negated = false;
    this.condition = condition;
    this.consequent = consequent;
    this.alternate = alternate;
  }
  initialize() {
    this.condition.setRequiresExpression();
    getEnclosingScopeBlock(this).markIIFEPatcherDescendant(this);
  }
  canPatchAsExpression() {
    if (this.consequent && !this.consequent.canPatchAsExpression()) {
      return false;
    }
    if (this.alternate && !this.alternate.canPatchAsExpression()) {
      return false;
    }
    return true;
  }
  prefersToPatchAsExpression() {
    const { consequent, alternate } = this;
    if (!consequent || !alternate) {
      return false;
    }
    return consequent.prefersToPatchAsExpression() && alternate.prefersToPatchAsExpression();
  }
  setExpression(force = false) {
    const willPatchAsExpression = super.setExpression(force);
    if (willPatchAsExpression && this.willPatchAsTernary()) {
      if (this.consequent) {
        this.consequent.setRequiresExpression();
      }
      if (this.alternate) {
        this.alternate.setRequiresExpression();
      }
      return true;
    }
    return false;
  }
  negate() {
    this.negated = !this.negated;
  }
  willPatchAsTernary() {
    return this.prefersToPatchAsExpression() || this.forcedToPatchAsExpression() && (!this.consequent || this.consequent.prefersToPatchAsExpression()) && (!this.alternate || this.alternate.prefersToPatchAsExpression());
  }
  willPatchAsIIFE() {
    return !this.willPatchAsTernary() && this.forcedToPatchAsExpression();
  }
  patchAsExpression({ needsParens } = {}) {
    const addParens = this.negated || needsParens && !this.isSurroundedByParentheses();
    this.overwrite(this.contentStart, this.condition.outerStart, `${this.negated ? "!" : ""}${addParens ? "(" : ""}`);
    if (this.node.isUnless) {
      this.condition.negate();
    }
    this.condition.patch();
    const thenTokenIndex = this.getThenTokenIndex();
    if (thenTokenIndex) {
      const thenToken = notNull(this.sourceTokenAtIndex(thenTokenIndex));
      this.overwrite(thenToken.start, thenToken.end, "?");
    } else {
      this.insert(this.condition.outerEnd, " ?");
    }
    const elseTokenIndex = this.getElseSourceTokenIndex();
    const elseToken = elseTokenIndex && this.sourceTokenAtIndex(elseTokenIndex);
    const { consequent, alternate } = this;
    if (consequent && alternate) {
      if (!elseToken) {
        throw this.error("Expected else token in conditional.");
      }
      consequent.patch();
      this.overwrite(elseToken.start, elseToken.end, ":");
      alternate.patch();
    } else if (consequent && !alternate) {
      consequent.patch();
      if (elseToken !== null) {
        this.overwrite(elseToken.start, elseToken.end, " : undefined");
      } else {
        this.insert(consequent.outerEnd, " : undefined");
      }
    } else if (alternate) {
      if (!elseToken) {
        throw this.error("Expected else token in conditional.");
      }
      const semicolonTokenIndex = this.indexOfSourceTokenBetweenSourceIndicesMatching(this.condition.outerEnd, elseToken.start, (token) => token.type === SourceType8.SEMICOLON);
      if (semicolonTokenIndex) {
        const semicolonToken = this.sourceTokenAtIndex(semicolonTokenIndex);
        if (semicolonToken) {
          this.remove(semicolonToken.start, semicolonToken.end);
        }
      }
      this.overwrite(elseToken.start, elseToken.end, "undefined :");
      alternate.patch();
    } else {
      if (elseToken !== null) {
        this.overwrite(elseToken.start, elseToken.end, "undefined : undefined");
      } else {
        this.insert(this.condition.outerEnd, " undefined : undefined");
      }
    }
    if (addParens) {
      this.insert(this.contentEnd, ")");
    }
  }
  patchAsForcedExpression() {
    if (this.willPatchAsTernary()) {
      this.patchAsExpression();
    } else if (this.willPatchAsIIFE()) {
      this.patchAsIIFE();
    }
  }
  patchAsIIFE() {
    if (this.negated) {
      this.insert(this.innerStart, "!");
    }
    const baseIndent = notNull(this.parent).getIndent(0);
    const conditionIndent = notNull(this.parent).getIndent(1);
    if (this.consequent) {
      this.consequent.setShouldPatchInline(false);
      this.consequent.setImplicitlyReturns();
    }
    if (this.alternate) {
      this.alternate.setShouldPatchInline(false);
      this.alternate.setImplicitlyReturns();
    }
    this.patchInIIFE(() => {
      this.insert(this.innerStart, `
${conditionIndent}`);
      this.patchAsStatement();
      this.insert(this.innerEnd, `
${baseIndent}`);
    });
  }
  canHandleImplicitReturn() {
    return this.willPatchAsIIFE();
  }
  patchAsStatement() {
    this.patchConditionForStatement();
    this.patchConsequentForStatement();
    this.patchAlternateForStatement();
  }
  patchConditionForStatement() {
    const ifToken = notNull(this.sourceTokenAtIndex(this.getIfSourceTokenIndex()));
    this.overwrite(ifToken.start, ifToken.end, "if");
    const conditionHasParentheses = this.condition.isSurroundedByParentheses();
    if (!conditionHasParentheses) {
      this.insert(this.condition.outerStart, "(");
    }
    if (this.node.isUnless) {
      this.condition.negate();
    }
    this.condition.patch({ needsParens: false });
    if (!conditionHasParentheses) {
      this.insert(this.condition.outerEnd, ")");
    }
    const thenTokenIndex = this.getThenTokenIndex();
    if (thenTokenIndex) {
      const thenToken = notNull(this.sourceTokenAtIndex(thenTokenIndex));
      if (this.consequent) {
        this.remove(thenToken.start, this.consequent.outerStart);
      } else {
        this.remove(thenToken.start, thenToken.end);
      }
    }
  }
  patchConsequentForStatement() {
    this.insert(this.condition.outerEnd, " {");
    if (this.alternate) {
      const elseTokenIndex = notNull(this.getElseSourceTokenIndex());
      const elseToken = notNull(this.sourceTokenAtIndex(elseTokenIndex));
      const rightBracePosition = elseToken.start;
      if (this.consequent !== null) {
        this.consequent.patch({ leftBrace: false, rightBrace: false });
      }
      this.insert(rightBracePosition, "} ");
    } else {
      if (this.consequent !== null) {
        this.consequent.patch({ leftBrace: false });
      } else {
        this.insert(this.condition.outerEnd, "} ");
      }
    }
  }
  patchAlternateForStatement() {
    const elseTokenIndex = this.getElseSourceTokenIndex();
    if (this.alternate && elseTokenIndex) {
      const ifToken = this.sourceTokenAtIndex(notNull(elseTokenIndex.next()));
      const isElseIf = ifToken ? ifToken.type === SourceType8.IF : false;
      if (isElseIf) {
        this.alternate.patch({ leftBrace: false, rightBrace: false });
      } else {
        const elseToken = notNull(this.sourceTokenAtIndex(elseTokenIndex));
        const leftBracePosition = elseToken.end;
        this.insert(leftBracePosition, " {");
        this.alternate.patch({ leftBrace: false });
      }
    } else if (elseTokenIndex !== null) {
      const elseToken = notNull(this.sourceTokenAtIndex(elseTokenIndex));
      this.insert(elseToken.end, " {}");
    } else if (super.implicitlyReturns()) {
      const emptyImplicitReturnCode = this.implicitReturnPatcher().getEmptyImplicitReturnCode();
      if (emptyImplicitReturnCode) {
        this.insert(this.innerEnd, " else {\n");
        this.insert(this.innerEnd, `${this.getIndent(1)}${emptyImplicitReturnCode}
`);
        this.insert(this.innerEnd, `${this.getIndent()}}`);
      }
    }
  }
  implicitlyReturns() {
    return super.implicitlyReturns() && this.willPatchAsExpression();
  }
  setImplicitlyReturns() {
    super.setImplicitlyReturns();
    if (this.consequent) {
      this.consequent.setImplicitlyReturns();
    }
    if (this.alternate) {
      this.alternate.setImplicitlyReturns();
    }
  }
  statementNeedsSemicolon() {
    return false;
  }
  getIfSourceTokenIndex() {
    const ifTokenIndex = this.indexOfSourceTokenStartingAtSourceIndex(this.contentStart);
    if (!ifTokenIndex) {
      throw this.error("expected IF token at start of conditional");
    }
    const ifToken = notNull(this.sourceTokenAtIndex(ifTokenIndex));
    if (ifToken.type !== SourceType8.IF) {
      throw this.error(`expected IF token at start of conditional, but got ${SourceType8[ifToken.type]}`);
    }
    return ifTokenIndex;
  }
  getElseSourceTokenIndex() {
    const elseTokenIndex = this.indexOfSourceTokenBetweenSourceIndicesMatching(this.consequent !== null ? this.consequent.outerEnd : this.condition.outerEnd, this.alternate !== null ? this.alternate.outerStart : this.outerEnd, (token) => token.type === SourceType8.ELSE);
    if (this.alternate !== null && !elseTokenIndex) {
      throw this.error("expected ELSE token between consequent and alternate", this.consequent !== null ? this.consequent.outerEnd : this.condition.outerEnd, this.alternate.outerStart);
    }
    return elseTokenIndex;
  }
  getThenTokenIndex() {
    let searchEnd;
    if (this.consequent) {
      searchEnd = this.consequent.outerStart;
    } else if (this.alternate) {
      searchEnd = this.alternate.outerStart;
    } else {
      const nextToken = this.nextSemanticToken();
      if (nextToken) {
        searchEnd = nextToken.end;
      } else {
        searchEnd = this.contentEnd;
      }
    }
    return this.indexOfSourceTokenBetweenSourceIndicesMatching(this.condition.outerEnd, searchEnd, (token) => token.type === SourceType8.THEN);
  }
  allCodePathsPresent() {
    if (!this.consequent || !this.alternate) {
      return false;
    }
    return this.consequent.allCodePathsPresent() && this.alternate.allCodePathsPresent();
  }
};

// src/stages/main/patchers/SwitchPatcher.ts
import { SourceType as SourceType9 } from "coffee-lex";
var SwitchPatcher = class extends NodePatcher {
  constructor(patcherContext, expression, cases, alternate) {
    super(patcherContext);
    this.expression = expression;
    this.cases = cases;
    this.alternate = alternate;
  }
  initialize() {
    if (this.expression) {
      this.expression.setRequiresExpression();
    }
    getEnclosingScopeBlock(this).markIIFEPatcherDescendant(this);
  }
  prefersToPatchAsExpression() {
    return false;
  }
  patchAsStatement() {
    if (this.expression) {
      if (!this.expression.isSurroundedByParentheses()) {
        this.insert(this.expression.contentStart, "(");
      }
      this.expression.patch();
      if (!this.expression.isSurroundedByParentheses()) {
        this.insert(this.expression.contentEnd, ")");
      }
      this.insert(this.expression.outerEnd, " {");
    } else {
      this.cases.forEach((casePatcher) => casePatcher.negate());
      const switchToken = this.getSwitchToken();
      this.insert(switchToken.end, " (false) {");
    }
    this.cases.forEach((casePatcher) => casePatcher.patch());
    this.overwriteElse();
    if (this.alternate) {
      this.alternate.patch({ leftBrace: false, rightBrace: false });
    } else if (this.getElseToken() === null && super.implicitlyReturns()) {
      const emptyImplicitReturnCode = this.implicitReturnPatcher().getEmptyImplicitReturnCode();
      if (emptyImplicitReturnCode) {
        this.insert(this.contentEnd, `
`);
        this.insert(this.contentEnd, `${this.getIndent(1)}default:
`);
        this.insert(this.contentEnd, `${this.getIndent(2)}${emptyImplicitReturnCode}`);
      }
    }
    this.appendLineAfter("}");
  }
  implicitlyReturns() {
    return super.implicitlyReturns() && this.willPatchAsExpression();
  }
  setImplicitlyReturns() {
    super.setImplicitlyReturns();
    this.cases.forEach((casePatcher) => casePatcher.setImplicitlyReturns());
    if (this.alternate) {
      this.alternate.setImplicitlyReturns();
    }
  }
  patchAsExpression() {
    this.setImplicitlyReturns();
    this.patchInIIFE(() => {
      this.insert(this.innerStart, " ");
      this.patchAsStatement();
      this.insert(this.innerEnd, " ");
    });
  }
  willPatchAsIIFE() {
    return this.willPatchAsExpression();
  }
  canHandleImplicitReturn() {
    return this.willPatchAsExpression();
  }
  overwriteElse() {
    const elseToken = this.getElseToken();
    if (elseToken) {
      this.overwrite(elseToken.start, elseToken.end, "default:");
    }
  }
  getElseToken() {
    let searchStart;
    if (this.cases.length > 0) {
      searchStart = this.cases[this.cases.length - 1].outerEnd;
    } else {
      searchStart = this.expression.outerEnd;
    }
    let searchEnd;
    if (this.alternate) {
      searchEnd = this.alternate.outerStart;
    } else {
      searchEnd = this.contentEnd;
    }
    const elseTokenIndex = this.indexOfSourceTokenBetweenSourceIndicesMatching(searchStart, searchEnd, (token) => token.type === SourceType9.ELSE);
    if (!elseTokenIndex || elseTokenIndex.isBefore(this.contentStartTokenIndex)) {
      if (this.alternate) {
        throw this.alternate.error(`no ELSE token found before 'switch' alternate`);
      } else {
        return null;
      }
    }
    return this.sourceTokenAtIndex(elseTokenIndex);
  }
  getSwitchToken() {
    const switchToken = this.sourceTokenAtIndex(this.contentStartTokenIndex);
    if (!switchToken) {
      throw this.error(`bad token index for start of 'switch'`);
    }
    if (switchToken.type !== SourceType9.SWITCH) {
      throw this.error(`unexpected ${SourceType9[switchToken.type]} token at start of 'switch'`);
    }
    return switchToken;
  }
  allCodePathsPresent() {
    if (!this.alternate) {
      return false;
    }
    return this.cases.every((switchCase) => switchCase.allCodePathsPresent()) && this.alternate.allCodePathsPresent();
  }
};

// src/stages/main/patchers/ReturnPatcher.ts
var ReturnPatcher = class extends NodePatcher {
  constructor(patcherContext, expression) {
    super(patcherContext);
    this.expression = expression;
  }
  initialize() {
    this.setExplicitlyReturns();
    if (this.expression !== null) {
      if (this.willConvertToImplicitReturn()) {
        this.expression.setImplicitlyReturns();
      } else {
        this.expression.setRequiresExpression();
      }
    }
  }
  canPatchAsExpression() {
    return false;
  }
  patchAsStatement() {
    if (this.getScope().containerNode.type === "Program") {
      this.addSuggestion(AVOID_TOP_LEVEL_RETURN);
    }
    if (this.expression) {
      if (this.willConvertToImplicitReturn()) {
        this.remove(this.contentStart, this.expression.outerStart);
      }
      this.expression.patch();
    }
  }
  willConvertToImplicitReturn() {
    if (!this.expression) {
      throw this.error("Expected non-null expression.");
    }
    return !this.expression.isSurroundedByParentheses() && (this.expression instanceof ConditionalPatcher || this.expression instanceof SwitchPatcher);
  }
};

// src/stages/main/patchers/BlockPatcher.ts
var BlockPatcher = class extends SharedBlockPatcher {
  constructor() {
    super(...arguments);
    this._iifePatcherDescendants = [];
    this._explicitDeclarationsToAdd = [];
  }
  markIIFEPatcherDescendant(iifePatcher) {
    this._iifePatcherDescendants.push(iifePatcher);
  }
  initialize() {
    for (const iifePatcher of this._iifePatcherDescendants) {
      if (!iifePatcher.willPatchAsIIFE()) {
        continue;
      }
      const fakeScope = new Scope(iifePatcher.node);
      traverse4(iifePatcher.node, (child) => {
        fakeScope.processNode(child);
      });
      for (const name of fakeScope.getOwnNames()) {
        if (countVariableUsages(this.node, name) > countVariableUsages(iifePatcher.node, name) && this._explicitDeclarationsToAdd.indexOf(name) === -1) {
          this._explicitDeclarationsToAdd.push(name);
        }
      }
    }
  }
  canPatchAsExpression() {
    return this._explicitDeclarationsToAdd.length === 0 && this.statements.every((statement) => statement.canPatchAsExpression());
  }
  prefersToPatchAsExpression() {
    return this.statements.length === 0 || this.statements.length === 1 && this.statements[0].prefersToPatchAsExpression();
  }
  setExpression(force = false) {
    const willPatchAsExpression = super.setExpression(force);
    if (willPatchAsExpression && this.prefersToPatchAsExpression()) {
      this.statements.forEach((statement) => statement.setExpression());
      return true;
    }
    return false;
  }
  setImplicitlyReturns() {
    if (this.statements.length > 0) {
      this.statements[this.statements.length - 1].setImplicitlyReturns();
    }
  }
  setShouldPatchInline(shouldPatchInline) {
    this.shouldPatchInline = shouldPatchInline;
  }
  patchAsStatement({ leftBrace = true, rightBrace = true } = {}) {
    if (this.statements.length > 0) {
      this.removeInitialAndFinalParens();
    }
    if (leftBrace) {
      this.insert(this.innerStart, "{");
    }
    if (this._explicitDeclarationsToAdd.length > 0) {
      this.insertStatementsAtIndex([`var ${this._explicitDeclarationsToAdd.join(", ")};`], 0);
    }
    let constructor = null;
    this.statements.forEach((statement, i, statements) => {
      if (i === statements.length - 1 && this.parent instanceof FunctionPatcher) {
        if (statement instanceof ReturnPatcher && !statement.expression) {
          this.removeFinalEmptyReturn(statement);
          return;
        }
      }
      if (statement.node.type === "Constructor") {
        if (constructor) {
          throw this.error("Unexpectedly found two constructors in the same block.");
        }
        constructor = statement;
      } else {
        this.patchInnerStatement(statement);
      }
    });
    if (constructor) {
      this.patchInnerStatement(constructor);
    }
    if (rightBrace) {
      if (this.inline()) {
        this.insert(this.innerEnd, " }");
      } else {
        this.appendLineAfter("}", -1);
      }
    }
  }
  removeInitialAndFinalParens() {
    const firstChild = this.statements[0];
    for (let tokenIndex = this.contentStartTokenIndex; tokenIndex !== null && tokenIndex.isBefore(firstChild.outerStartTokenIndex); tokenIndex = tokenIndex.next()) {
      const token = this.sourceTokenAtIndex(tokenIndex);
      if (token && token.type === SourceType10.LPAREN) {
        this.remove(token.start, token.end);
      }
    }
    const lastChild = this.statements[this.statements.length - 1];
    for (let tokenIndex = this.contentEndTokenIndex; tokenIndex !== null && tokenIndex.isAfter(lastChild.outerEndTokenIndex); tokenIndex = tokenIndex.previous()) {
      const token = this.sourceTokenAtIndex(tokenIndex);
      if (token && token.type === SourceType10.RPAREN) {
        this.remove(token.start, token.end);
      }
    }
  }
  patchInnerStatement(statement) {
    const hasImplicitReturn = statement.implicitlyReturns() && !statement.explicitlyReturns();
    if (statement.isSurroundedByParentheses() && !statement.statementNeedsParens() && !hasImplicitReturn) {
      this.remove(statement.outerStart, statement.innerStart);
      this.remove(statement.innerEnd, statement.outerEnd);
    }
    const implicitReturnPatcher = hasImplicitReturn ? this.implicitReturnPatcher() : null;
    if (implicitReturnPatcher) {
      implicitReturnPatcher.patchImplicitReturnStart(statement);
    }
    statement.patch();
    if (implicitReturnPatcher) {
      implicitReturnPatcher.patchImplicitReturnEnd(statement);
    }
    if (statement.statementNeedsSemicolon()) {
      this.insert(statement.outerEnd, ";");
    }
  }
  removeFinalEmptyReturn(statement) {
    const previousTokenIndex = statement.contentStartTokenIndex.previous();
    const nextTokenIndex = statement.contentEndTokenIndex.next();
    const previousToken = previousTokenIndex && this.sourceTokenAtIndex(previousTokenIndex);
    const nextToken = nextTokenIndex && this.sourceTokenAtIndex(nextTokenIndex);
    if (previousToken && previousToken.type === SourceType10.NEWLINE && (!nextToken || nextToken.type === SourceType10.NEWLINE)) {
      this.remove(previousToken.start, statement.outerEnd);
    } else if (previousToken && previousToken.type === SourceType10.SEMICOLON) {
      this.remove(previousToken.start, statement.outerEnd);
    } else {
      this.remove(statement.outerStart, statement.outerEnd);
    }
  }
  patchAsExpression({
    leftBrace = this.statements.length > 1,
    rightBrace = this.statements.length > 1
  } = {}) {
    if (leftBrace) {
      this.insert(this.innerStart, "(");
    }
    if (this.statements.length === 0) {
      this.insert(this.contentStart, "undefined");
    } else {
      this.statements.forEach((statement, i, statements) => {
        statement.setRequiresExpression();
        statement.patch();
        if (i !== statements.length - 1) {
          const semicolonTokenIndex = this.getSemicolonSourceTokenIndexBetween(statement, statements[i + 1]);
          if (semicolonTokenIndex) {
            const semicolonToken = notNull(this.sourceTokenAtIndex(semicolonTokenIndex));
            this.overwrite(semicolonToken.start, semicolonToken.end, ",");
          } else {
            this.insert(statement.outerEnd, ",");
          }
        }
      });
    }
    const lastToken = this.lastToken();
    if (lastToken.type === SourceType10.SEMICOLON) {
      this.remove(lastToken.start, lastToken.end);
    }
    if (rightBrace) {
      this.insert(this.innerEnd, ")");
    }
  }
  getSemicolonSourceTokenIndexBetween(left, right) {
    return this.indexOfSourceTokenBetweenPatchersMatching(left, right, (token) => token.type === SourceType10.SEMICOLON);
  }
  allCodePathsPresent() {
    return this.statements[this.statements.length - 1].allCodePathsPresent();
  }
  allowPatchingOuterBounds() {
    return true;
  }
};

// src/utils/blockStartsWithObjectInitialiser.ts
function blockStartsWithObjectInitialiser(patcher) {
  if (!(patcher instanceof BlockPatcher) || patcher.statements.length !== 1) {
    return false;
  }
  const statement = patcher.statements[0];
  return containsDescendant(statement.node, (child) => child instanceof ObjectInitialiser4 && child.start === statement.contentStart);
}

// src/stages/main/patchers/ManuallyBoundFunctionPatcher.ts
var ManuallyBoundFunctionPatcher = class extends FunctionPatcher {
  patchAsStatement(options = {}) {
    this.insert(this.innerStart, "(");
    super.patchAsExpression(options);
    this.insert(this.innerEnd, ".bind(this))");
  }
  patchAsExpression(options = {}) {
    super.patchAsExpression(options);
    if (!options.method) {
      this.insert(this.innerEnd, ".bind(this)");
    }
  }
  expectedArrowType() {
    return "=>";
  }
};

// src/stages/main/patchers/BoundFunctionPatcher.ts
var BoundFunctionPatcher = class extends FunctionPatcher {
  initialize() {
    super.initialize();
    if (this.shouldPatchAsBlocklessArrowFunction()) {
      notNull(this.body).setExpression();
    }
  }
  static patcherClassOverrideForNode(node) {
    if (referencesArguments(node)) {
      return ManuallyBoundFunctionPatcher;
    } else {
      return null;
    }
  }
  patchAsStatement(options = {}) {
    this.patchAsExpression(options);
  }
  patchFunctionStart() {
    const arrow = this.getArrowToken();
    if (!this.hasParamStart()) {
      this.insert(this.contentStart, "() ");
    } else if (!this.parameterListNeedsParentheses()) {
      const [param] = this.parameters;
      if (param.isSurroundedByParentheses()) {
        this.remove(param.outerStart, param.contentStart);
        this.remove(param.contentEnd, param.outerEnd);
      }
    }
    if (!this.willPatchBodyInline()) {
      this.insert(arrow.end, " {");
    }
  }
  parameterListNeedsParentheses() {
    const parameters = this.parameters;
    if (parameters.length !== 1) {
      return true;
    }
    const [param] = parameters;
    return !(param instanceof IdentifierPatcher);
  }
  patchFunctionBody() {
    if (this.body) {
      if (!this.willPatchBodyInline()) {
        if (this.isEndOfFunctionCall()) {
          this.body.patch({ leftBrace: false, rightBrace: false });
          this.placeCloseBraceBeforeFunctionCallEnd();
        } else {
          this.body.patch({ leftBrace: false });
        }
      } else {
        const needsParens = blockStartsWithObjectInitialiser(this.body) && !this.body.isSurroundedByParentheses();
        if (needsParens) {
          this.insert(this.body.innerStart, "(");
        }
        this.body.patch();
        if (needsParens) {
          this.insert(this.body.innerEnd, ")");
        }
      }
    } else {
      this.insert(this.innerEnd, "}");
    }
  }
  expectedArrowType() {
    return "=>";
  }
  willPatchBodyInline() {
    return this.body !== null && this.body.willPatchAsExpression();
  }
  shouldPatchAsBlocklessArrowFunction() {
    if (!this.body) {
      return false;
    }
    if (containsDescendant(this.node, (child) => child instanceof AssignOp3)) {
      return false;
    }
    return this.body.inline();
  }
  statementNeedsParens() {
    return false;
  }
};

// src/stages/main/patchers/BoundAsyncFunctionPatcher.ts
var BoundAsyncFunctionPatcher = class extends BoundFunctionPatcher {
  patchFunctionStart() {
    this.insert(this.contentStart, "async ");
    super.patchFunctionStart();
  }
};

// src/stages/main/patchers/BoundGeneratorFunctionPatcher.ts
var BoundGeneratorFunctionPatcher = class extends ManuallyBoundFunctionPatcher {
  patchFunctionStart({ method = false } = {}) {
    const arrow = this.getArrowToken();
    if (!method) {
      this.insert(this.contentStart, "function*");
    }
    if (!this.hasParamStart()) {
      this.insert(this.contentStart, "() ");
    }
    this.overwrite(arrow.start, arrow.end, "{");
  }
};

// src/stages/main/patchers/GeneratorFunctionPatcher.ts
var GeneratorFunctionPatcher = class extends FunctionPatcher {
  patchFunctionStart({ method = false }) {
    const arrow = this.getArrowToken();
    if (!method) {
      this.insert(this.contentStart, "function*");
    }
    if (!this.hasParamStart()) {
      this.insert(this.contentStart, "() ");
    }
    this.overwrite(arrow.start, arrow.end, "{");
  }
};

// src/stages/main/patchers/StringPatcher.ts
import { SourceType as SourceType14 } from "coffee-lex";

// src/stages/main/patchers/ExportAllDeclarationPatcher.ts
var ExportAllDeclarationPatcher = class extends NodePatcher {
  constructor(patcherContext, source) {
    super(patcherContext);
    this.source = source;
  }
  patchAsStatement() {
    this.source.patch();
  }
};

// src/stages/main/patchers/ExportBindingsDeclarationPatcher.ts
import { SourceType as SourceType11 } from "coffee-lex";
var ExportBindingsDeclarationPatcher = class extends NodePatcher {
  constructor(patcherContext, namedExports, source) {
    super(patcherContext);
    this.namedExports = namedExports;
    this.source = source;
  }
  patchAsStatement() {
    this.namedExports.forEach((namedExport, i) => {
      namedExport.patch();
      const isLast = i === this.namedExports.length - 1;
      const commaTokenIndex = this.indexOfSourceTokenAfterSourceTokenIndex(namedExport.outerEndTokenIndex, SourceType11.COMMA, isSemanticToken);
      const commaToken = commaTokenIndex && this.sourceTokenAtIndex(commaTokenIndex);
      if (!isLast && !commaToken) {
        this.insert(namedExport.outerEnd, ",");
      }
    });
    if (this.source) {
      this.source.patch();
    }
  }
};

// src/stages/main/patchers/ImportDeclarationPatcher.ts
import { SourceType as SourceType12 } from "coffee-lex";
var ImportDeclarationPatcher = class extends NodePatcher {
  constructor(patcherContext, defaultBinding, namespaceImport, namedImports, source) {
    super(patcherContext);
    this.defaultBinding = defaultBinding;
    this.namespaceImport = namespaceImport;
    this.namedImports = namedImports;
    this.source = source;
  }
  patchAsStatement() {
    if (this.defaultBinding) {
      this.defaultBinding.patch();
    }
    if (this.namespaceImport) {
      this.namespaceImport.patch();
    }
    const { namedImports } = this;
    if (namedImports) {
      namedImports.forEach((namedImport, i) => {
        namedImport.patch();
        const isLast = i === namedImports.length - 1;
        const commaTokenIndex = this.indexOfSourceTokenAfterSourceTokenIndex(namedImport.outerEndTokenIndex, SourceType12.COMMA, isSemanticToken);
        const commaToken = commaTokenIndex && this.sourceTokenAtIndex(commaTokenIndex);
        if (!isLast && !commaToken) {
          this.insert(namedImport.outerEnd, ",");
        }
      });
    }
    this.source.patch();
  }
};

// src/stages/main/patchers/InterpolatedPatcher.ts
import { SourceType as SourceType13 } from "coffee-lex";

// src/utils/downgradeUnicodeCodePointEscapesInRange.ts
function downgradeUnicodeCodePointEscapesInRange(start, end, patcher, { needsExtraEscape } = {}) {
  const source = patcher.context.source;
  let numBackslashes = 0;
  for (let i = start; i < end - 2; i++) {
    if (source[i] === "\\") {
      numBackslashes++;
    } else {
      numBackslashes = 0;
    }
    if (numBackslashes % 2 === 1 && source.slice(i, i + 3) === "\\u{") {
      let codePointEnd = i + 3;
      while (codePointEnd < source.length && source[codePointEnd] !== "}") {
        codePointEnd++;
      }
      const codePointText = source.slice(i + 3, codePointEnd);
      const codePoint = parseInt(codePointText, 16);
      if (codePoint < 65536) {
        patcher.overwrite(i, codePointEnd + 1, toUnicodeEscape(codePoint));
      } else {
        const high = Math.floor((codePoint - 65536) / 1024) + 55296;
        const low = (codePoint - 65536) % 1024 + 56320;
        patcher.overwrite(i, codePointEnd + 1, `${toUnicodeEscape(high)}${needsExtraEscape ? "\\" : ""}${toUnicodeEscape(low)}`);
      }
      i = codePointEnd;
    }
  }
}
function toUnicodeEscape(val) {
  const str = val.toString(16);
  return `\\u${"0".repeat(4 - str.length)}${str}`;
}

// src/utils/escape.ts
function escape(source, patcher, skipPattern, escapeStrings, start, end) {
  for (let i = start; i < end; i++) {
    if (skipPattern.test(source.slice(i, end))) {
      i++;
    } else if (escapeStrings.some((str) => source.slice(i, i + str.length) === str)) {
      patcher.appendRight(i, "\\");
    }
  }
}

// src/utils/escapeSpecialWhitespaceInRange.ts
function isAlreadyEscaped(i, start, patcher) {
  let numLeadingBackslashes = 0;
  while (i - numLeadingBackslashes - 1 >= start && patcher.context.source[i - numLeadingBackslashes - 1] === "\\") {
    numLeadingBackslashes++;
  }
  return numLeadingBackslashes % 2 === 1;
}
function escapeSpecialWhitespaceInRange(start, end, patcher) {
  for (let i = start; i < end; i++) {
    let unicodeSequence = null;
    if (patcher.context.source[i] === "\u2028") {
      unicodeSequence = "u2028";
    } else if (patcher.context.source[i] === "\u2029") {
      unicodeSequence = "u2029";
    } else {
      continue;
    }
    if (isAlreadyEscaped(i, start, patcher)) {
      patcher.overwrite(i, i + 1, unicodeSequence);
    } else {
      patcher.overwrite(i, i + 1, `\\${unicodeSequence}`);
    }
  }
}

// src/utils/escapeZeroCharsInRange.ts
function escapeZeroCharsInRange(start, end, patcher) {
  let numBackslashes = 0;
  for (let i = start; i < end; i++) {
    if (patcher.context.source[i] === "\\") {
      numBackslashes++;
      continue;
    } else if (patcher.context.source[i] === "0" && numBackslashes % 2 === 1) {
      patcher.overwrite(i, i + 1, "x00");
    }
    numBackslashes = 0;
  }
}

// src/stages/main/patchers/InterpolatedPatcher.ts
var InterpolatedPatcher = class extends NodePatcher {
  constructor(patcherContext, quasis, expressions) {
    super(patcherContext);
    this.quasis = quasis;
    this.expressions = expressions;
  }
  initialize() {
    for (const expression of this.expressions) {
      if (expression) {
        expression.setRequiresExpression();
      }
    }
  }
  patchInterpolations() {
    for (let i = 0; i < this.expressions.length; i++) {
      const interpolationStart = this.getInterpolationStartTokenAtIndex(i);
      const expression = this.expressions[i];
      if (expression) {
        this.overwrite(interpolationStart.start, interpolationStart.start + 1, "$");
        expression.patch();
      } else {
        const interpolationEndIndex = this.quasis[i + 1].contentStart;
        this.remove(interpolationStart.start, interpolationEndIndex);
      }
    }
  }
  getInterpolationStartTokenAtIndex(index) {
    const interpolationStartIndex = this.indexOfSourceTokenBetweenSourceIndicesMatching(this.quasis[index].contentEnd, this.contentEnd, (token) => token.type === SourceType13.INTERPOLATION_START);
    if (!interpolationStartIndex) {
      throw this.error("Cannot find interpolation start for string interpolation.");
    }
    const interpolationStart = this.sourceTokenAtIndex(interpolationStartIndex);
    if (!interpolationStart || this.slice(interpolationStart.start, interpolationStart.start + 1) !== "#") {
      throw this.error("Cannot find '#' in interpolation start.");
    }
    return interpolationStart;
  }
  processContents() {
    for (const quasi of this.quasis) {
      const tokens = this.getProgramSourceTokens().slice(quasi.contentStartTokenIndex, notNull(quasi.contentEndTokenIndex.next())).toArray();
      for (const token of tokens) {
        if (token.type === SourceType13.STRING_PADDING || token.type === SourceType13.HEREGEXP_COMMENT) {
          const paddingCode = this.slice(token.start, token.end);
          const numNewlines = (paddingCode.match(/\n/g) || []).length;
          this.overwrite(token.start, token.end, "\\\n".repeat(numNewlines));
        } else if (token.type === SourceType13.STRING_LINE_SEPARATOR) {
          this.insert(token.start, " \\");
        } else if (token.type === SourceType13.STRING_CONTENT) {
          escapeSpecialWhitespaceInRange(token.start, token.end, this);
          if (this.shouldExcapeZeroChars()) {
            escapeZeroCharsInRange(token.start, token.end, this);
          }
          if (this.shouldDowngradeUnicodeCodePointEscapes()) {
            downgradeUnicodeCodePointEscapesInRange(token.start, token.end, this, { needsExtraEscape: true });
          }
        }
      }
    }
  }
  shouldExcapeZeroChars() {
    return false;
  }
  shouldDowngradeUnicodeCodePointEscapes() {
    return false;
  }
  escapeQuasis(skipPattern, escapeStrings) {
    for (const quasi of this.quasis) {
      const tokens = this.getProgramSourceTokens().slice(quasi.contentStartTokenIndex, notNull(quasi.contentEndTokenIndex.next())).toArray();
      for (const token of tokens) {
        if (token.type === SourceType13.STRING_CONTENT) {
          escape(this.context.source, this.editor, skipPattern, escapeStrings, token.start, token.end);
        }
      }
    }
  }
  isRepeatable() {
    return this.expressions.every((patcher) => patcher !== null && patcher.isRepeatable());
  }
};

// src/stages/main/patchers/StringPatcher.ts
var StringPatcher = class extends InterpolatedPatcher {
  patchAsExpression({ forceTemplateLiteral } = {}) {
    const shouldBecomeTemplateLiteral = forceTemplateLiteral || this.shouldBecomeTemplateLiteral();
    const escapeStrings = [];
    const openQuoteToken = this.firstToken();
    const closeQuoteToken = this.lastToken();
    if (shouldBecomeTemplateLiteral) {
      escapeStrings.push("`");
      escapeStrings.push("${");
      this.overwrite(openQuoteToken.start, openQuoteToken.end, "`");
      this.overwrite(closeQuoteToken.start, closeQuoteToken.end, "`");
    } else if (openQuoteToken.type === SourceType14.TSSTRING_START) {
      escapeStrings.push("'");
      this.overwrite(openQuoteToken.start, openQuoteToken.end, "'");
      this.overwrite(closeQuoteToken.start, closeQuoteToken.end, "'");
    } else if (openQuoteToken.type === SourceType14.TDSTRING_START) {
      escapeStrings.push('"');
      this.overwrite(openQuoteToken.start, openQuoteToken.end, '"');
      this.overwrite(closeQuoteToken.start, closeQuoteToken.end, '"');
    }
    this.patchInterpolations();
    this.processContents();
    if (escapeStrings.length > 0) {
      this.escapeQuasis(/^\\/, escapeStrings);
    }
  }
  shouldBecomeTemplateLiteral() {
    if ((this.parent instanceof ImportDeclarationPatcher || this.parent instanceof ExportAllDeclarationPatcher || this.parent instanceof ExportBindingsDeclarationPatcher) && this.parent.source === this) {
      return false;
    }
    return this.expressions.length > 0 || this.node.raw.indexOf("\n") > -1;
  }
};

// src/stages/main/patchers/ObjectBodyMemberPatcher.ts
var ObjectBodyMemberPatcher = class extends NodePatcher {
  constructor(patcherContext, key, expression) {
    super(patcherContext);
    this.key = key;
    this.expression = expression;
  }
  initialize() {
    this.key.setRequiresExpression();
    if (this.expression) {
      this.expression.setRequiresExpression();
    }
  }
  patchAsExpression() {
    if (this.isMethod()) {
      this.patchAsMethod();
    } else {
      this.patchAsProperty();
    }
  }
  patchAsMethod() {
    if (!this.expression) {
      throw this.error("Expected expression to be non-null in method case.");
    }
    if (this.isAsyncMethod()) {
      this.insert(this.key.outerStart, "async ");
    }
    if (this.isGeneratorMethod()) {
      this.insert(this.key.outerStart, "*");
    }
    this.patchKey();
    this.remove(this.key.outerEnd, this.expression.outerStart);
    this.remove(this.expression.outerStart, this.expression.contentStart);
    this.remove(this.expression.contentEnd, this.expression.outerEnd);
    this.patchExpression();
  }
  patchAsProperty() {
    this.patchKey();
    this.patchExpression();
  }
  patchKey() {
    if (this.node instanceof ObjectInitialiserMember4 && this.node.isComputed) {
      this.key.patch();
      return;
    }
    const computedKeyPatcher = this.getComputedKeyPatcher();
    if (computedKeyPatcher !== null) {
      const repeatOptions = this.key.getRepeatableOptions();
      if (repeatOptions) {
        computedKeyPatcher.setRequiresRepeatableExpression(repeatOptions);
      }
      this.overwrite(this.key.outerStart, computedKeyPatcher.outerStart, "[");
      computedKeyPatcher.patch();
      this.overwrite(computedKeyPatcher.outerEnd, this.key.outerEnd, "]");
      if (repeatOptions) {
        this.key.overrideRepeatCode(computedKeyPatcher.getRepeatCode());
      }
    } else {
      const needsBrackets = !(this.key instanceof StringPatcher && !this.key.shouldBecomeTemplateLiteral()) && !(this.key instanceof IdentifierPatcher) && this.key.node.type !== "Int" && this.key.node.type !== "Float";
      if (needsBrackets) {
        this.insert(this.key.outerStart, "[");
      }
      this.key.patch();
      if (needsBrackets) {
        this.insert(this.key.outerEnd, "]");
      }
    }
  }
  getComputedKeyPatcher() {
    if (this.key instanceof StringPatcher && this.key.quasis.length === 2 && this.key.expressions.length === 1 && this.key.quasis[0].node.data === "" && this.key.quasis[1].node.data === "") {
      return this.key.expressions[0];
    }
    return null;
  }
  patchExpression() {
    if (this.expression) {
      this.expression.patch({ method: this.isMethod() });
    }
  }
  isMethod() {
    return this.expression instanceof FunctionPatcher && !(this.expression instanceof ManuallyBoundFunctionPatcher) && !(this.expression instanceof BoundFunctionPatcher);
  }
  isGeneratorMethod() {
    return this.expression instanceof GeneratorFunctionPatcher || this.expression instanceof BoundGeneratorFunctionPatcher;
  }
  isAsyncMethod() {
    return this.expression instanceof AsyncFunctionPatcher || this.expression instanceof BoundAsyncFunctionPatcher;
  }
};

// src/stages/main/patchers/ThisPatcher.ts
var ThisPatcher = class extends NodePatcher {
  patchAsExpression() {
    this.reportTopLevelThisIfNecessary();
    if (this.isShorthandThis()) {
      this.overwrite(this.contentStart, this.contentEnd, "this");
    }
  }
  isShorthandThis() {
    return this.getOriginalSource() === "@";
  }
  isRepeatable() {
    return true;
  }
  reportTopLevelThisIfNecessary() {
    let scope = this.getScope();
    while (scope.parent && ["Program", "Function", "GeneratorFunction", "Class"].indexOf(scope.containerNode.type) === -1) {
      scope = scope.parent;
    }
    if (scope.containerNode.type === "Program") {
      this.addSuggestion(AVOID_TOP_LEVEL_THIS);
    }
  }
};

// src/stages/main/patchers/ObjectInitialiserMemberPatcher.ts
var ObjectInitialiserMemberPatcher = class extends ObjectBodyMemberPatcher {
  setAssignee() {
    if (this.expression === null) {
      this.key.setAssignee();
    } else {
      this.expression.setAssignee();
    }
    super.setAssignee();
  }
  patchAsProperty() {
    if (this.expression === null) {
      const shouldExpand = !(this.key.node instanceof Identifier6) || this.node.isComputed;
      this.patchAsShorthand({
        expand: shouldExpand
      });
    } else {
      super.patchAsProperty();
    }
  }
  patchAsShorthand({ expand = false }) {
    const { key } = this;
    if (key instanceof MemberAccessOpPatcher) {
      key.patch();
      if (!(key.expression instanceof ThisPatcher)) {
        throw this.error(`expected property key member access on 'this', e.g. '@name'`);
      }
      this.insert(key.outerStart, `${key.getMemberName()}: `);
    } else if (expand) {
      const needsBrackets = key instanceof StringPatcher && key.shouldBecomeTemplateLiteral();
      if (needsBrackets) {
        this.insert(key.outerStart, "[");
      }
      const valueCode = key.patchRepeatable();
      if (needsBrackets) {
        this.insert(key.outerEnd, "]");
      }
      let keyEnd;
      if (this.node.isComputed) {
        const closeBracketToken = key.outerEndTokenIndex.next();
        const tokenAfterLast = closeBracketToken ? this.sourceTokenAtIndex(closeBracketToken) : null;
        if (!tokenAfterLast || tokenAfterLast.type !== SourceType15.RBRACKET) {
          throw this.error("Expected close-bracket after computed property.");
        }
        keyEnd = tokenAfterLast.end;
      } else {
        keyEnd = key.outerEnd;
      }
      this.insert(keyEnd, `: ${valueCode}`);
    }
  }
};

// src/utils/getObjectAssigneeKeys.ts
function getObjectAssigneeKeys(node) {
  const results = [];
  for (const member of node.members) {
    if (member instanceof ObjectInitialiserMemberPatcher && member.key instanceof IdentifierPatcher) {
      results.push(member.key.node.data);
    } else if (member instanceof AssignOpPatcher && member.assignee instanceof IdentifierPatcher) {
      results.push(member.assignee.node.data);
    }
  }
  return results;
}

// src/stages/main/patchers/DoOpPatcher.ts
import { SourceType as SourceType16 } from "coffee-lex";

// src/stages/main/patchers/DefaultParamPatcher.ts
var DefaultParamPatcher = class extends NodePatcher {
  constructor(patcherContext, param, value) {
    super(patcherContext);
    this.param = param;
    this.value = value;
  }
  initialize() {
    this.param.setRequiresExpression();
    this.value.setRequiresExpression();
  }
  patchAsExpression() {
    this.param.patch();
    this.value.patch();
  }
};

// src/stages/main/patchers/DoOpPatcher.ts
var DoOpPatcher = class extends NodePatcher {
  constructor(patcherContext, expression) {
    super(patcherContext);
    this.expression = expression;
  }
  initialize() {
    this.expression.setRequiresExpression();
  }
  patchAsExpression() {
    const doTokenIndex = this.getDoTokenIndex();
    const doToken = notNull(this.sourceTokenAtIndex(doTokenIndex));
    const nextToken = notNull(this.sourceTokenAtIndex(notNull(doTokenIndex.next())));
    this.remove(doToken.start, nextToken.start);
    const addParens = !(this.expression instanceof IdentifierPatcher);
    if (addParens) {
      this.insert(this.contentStart, "(");
    }
    this.expression.patch();
    if (addParens) {
      this.insert(this.innerEnd, ")");
    }
    const args = [];
    if (this.hasDoFunction()) {
      const func = this.getDoFunction();
      func.parameters.forEach((param) => {
        if (param instanceof DefaultParamPatcher) {
          const valueSource = param.value.getPatchedSource();
          this.remove(param.param.outerEnd, param.value.outerEnd);
          args.push(valueSource);
        } else {
          args.push(param.getPatchedSource());
        }
      });
    }
    this.insert(this.innerEnd, `(${args.join(", ")})`);
  }
  hasDoFunction() {
    return this.expression instanceof FunctionPatcher || this.expression instanceof AssignOpPatcher && this.expression.expression instanceof FunctionPatcher;
  }
  getDoFunction() {
    if (this.expression instanceof FunctionPatcher) {
      return this.expression;
    } else if (this.expression instanceof AssignOpPatcher && this.expression.expression instanceof FunctionPatcher) {
      return this.expression.expression;
    } else {
      throw this.error("Should only call getDoFunction if hasDoFunction is true.");
    }
  }
  getDoTokenIndex() {
    const index = this.contentStartTokenIndex;
    const token = this.sourceTokenAtIndex(index);
    if (!token || token.type !== SourceType16.DO) {
      throw this.error(`expected 'do' at start of expression`);
    }
    return index;
  }
};

// src/stages/main/patchers/ObjectInitialiserPatcher.ts
import { SourceType as SourceType17 } from "coffee-lex";
var ObjectInitialiserPatcher = class extends NodePatcher {
  constructor(patcherContext, members) {
    super(patcherContext);
    this.members = members;
  }
  initialize() {
    this.members.forEach((member) => member.setRequiresExpression());
  }
  setAssignee() {
    this.members.forEach((member) => member.setAssignee());
    super.setAssignee();
  }
  setExpression(force) {
    if (this.isImplicitObject()) {
      const { curlyBraceInsertionPosition } = this.getOpenCurlyInfo();
      this.adjustBoundsToInclude(curlyBraceInsertionPosition);
    }
    return super.setExpression(force);
  }
  patchAsExpression() {
    const implicitObject = this.isImplicitObject();
    if (implicitObject) {
      const { curlyBraceInsertionPosition, textToInsert, shouldIndent } = this.getOpenCurlyInfo();
      this.insert(curlyBraceInsertionPosition, textToInsert);
      if (shouldIndent) {
        this.indent();
      }
    }
    this.patchMembers();
    if (implicitObject) {
      if (this.shouldExpandCurlyBraces() && !this.isSurroundedByParentheses()) {
        this.appendLineAfter("}", -1);
      } else {
        this.insert(this.innerEnd, "}");
      }
    }
  }
  getOpenCurlyInfo() {
    let curlyBraceInsertionPosition = this.innerStart;
    let textToInsert = "{";
    let shouldIndent = false;
    if (this.shouldExpandCurlyBraces()) {
      if (this.implicitlyReturns() && !this.isSurroundedByParentheses()) {
        textToInsert = `{
${this.getIndent()}`;
        shouldIndent = true;
      } else {
        let tokenIndexBeforeOuterStartTokenIndex = this.outerStartTokenIndex;
        if (!this.isSurroundedByParentheses()) {
          tokenIndexBeforeOuterStartTokenIndex = tokenIndexBeforeOuterStartTokenIndex.previous();
        }
        if (tokenIndexBeforeOuterStartTokenIndex) {
          const precedingTokenIndex = this.context.sourceTokens.lastIndexOfTokenMatchingPredicate(isSemanticToken, tokenIndexBeforeOuterStartTokenIndex);
          if (precedingTokenIndex) {
            const precedingToken = notNull(this.sourceTokenAtIndex(precedingTokenIndex));
            curlyBraceInsertionPosition = precedingToken.end;
            const precedingTokenText = this.sourceOfToken(precedingToken);
            const lastCharOfToken = precedingTokenText[precedingTokenText.length - 1];
            const needsSpace = lastCharOfToken === ":" || lastCharOfToken === "=" || lastCharOfToken === ",";
            if (needsSpace) {
              textToInsert = " {";
            }
          }
        }
      }
    }
    return { curlyBraceInsertionPosition, textToInsert, shouldIndent };
  }
  patchAsStatement() {
    const needsParentheses = !this.isSurroundedByParentheses();
    const implicitObject = this.isImplicitObject();
    if (needsParentheses) {
      this.insert(this.contentStart, "(");
    }
    if (implicitObject) {
      if (this.shouldExpandCurlyBraces() && !this.isSurroundedByParentheses()) {
        this.insert(this.innerStart, `{
${this.getIndent()}`);
        this.indent();
      } else {
        this.insert(this.innerStart, "{");
      }
    }
    this.patchMembers();
    if (implicitObject) {
      if (this.shouldExpandCurlyBraces() && !this.isSurroundedByParentheses()) {
        this.appendLineAfter("}", -1);
      } else {
        this.insert(this.innerEnd, "}");
      }
    }
    if (needsParentheses) {
      this.insert(this.contentEnd, ")");
    }
  }
  shouldExpandCurlyBraces() {
    return this.isMultiline() || this.parent instanceof ObjectInitialiserMemberPatcher && notNull(this.parent.parent).isMultiline();
  }
  patchMembers() {
    this.members.forEach((member, i, members) => {
      member.patch();
      if (i !== members.length - 1) {
        if (!member.hasSourceTokenAfter(SourceType17.COMMA)) {
          this.insert(member.outerEnd, ",");
        }
      }
    });
  }
  isImplicitObject() {
    const tokens = this.context.sourceTokens;
    const indexOfFirstToken = notNull(tokens.indexOfTokenStartingAtSourceIndex(this.contentStart));
    return notNull(tokens.tokenAtIndex(indexOfFirstToken)).type !== SourceType17.LBRACE;
  }
  statementNeedsParens() {
    return true;
  }
};

// src/stages/main/patchers/SlicePatcher.ts
import { SourceType as SourceType18 } from "coffee-lex";
import { Int } from "decaffeinate-parser";
var SlicePatcher = class extends NodePatcher {
  constructor(patcherContext, expression, left, right) {
    super(patcherContext);
    this.expression = expression;
    this.left = left;
    this.right = right;
  }
  initialize() {
    this.expression.setRequiresExpression();
    if (this.left) {
      this.left.setRequiresExpression();
    }
    if (this.right) {
      this.right.setRequiresExpression();
    }
  }
  patchAsExpression() {
    this.expression.patch();
    const indexStart = this.getIndexStartSourceToken();
    this.overwrite(this.expression.outerEnd, indexStart.end, this.getSliceStartCode());
    if (this.left) {
      this.left.patch();
    } else if (this.right) {
      this.insert(indexStart.end, "0");
    }
    const slice = this.getSliceSourceToken();
    const right = this.right;
    if (right) {
      if (this.isInclusive()) {
        if (right.node.raw === "-1") {
          this.remove(slice.start, right.outerEnd);
        } else if (right.node instanceof Int) {
          this.overwrite(slice.start, right.outerEnd, `, ${right.node.data + 1}`);
        } else {
          this.overwrite(slice.start, slice.end, ", +");
          if (this.slice(right.contentStart, right.contentStart + 1) === "+") {
            this.insert(slice.end, " ");
          }
          right.patch({ needsParens: true });
          this.insert(right.outerEnd, " + 1 || undefined");
        }
      } else {
        this.overwrite(slice.start, slice.end, ", ");
        right.patch();
      }
    } else {
      this.overwrite(slice.start, slice.end, "");
    }
    const indexEnd = this.getIndexEndSourceToken();
    this.overwrite(indexEnd.start, indexEnd.end, this.getSpliceEndCode());
  }
  getSliceStartCode(splice = false) {
    return splice ? ".splice(" : ".slice(";
  }
  getSpliceEndCode(_splice = false) {
    return ")";
  }
  getSpliceCode(expressionCode) {
    const spliceStart = this.captureCodeForPatchOperation(() => this.patchAsSpliceExpressionStart());
    return `${spliceStart}, ...[].concat(${expressionCode}))`;
  }
  patchAsSpliceExpressionStart() {
    this.expression.patch();
    const indexStart = this.getIndexStartSourceToken();
    this.overwrite(this.expression.outerEnd, indexStart.end, this.getSliceStartCode(true));
    let leftCode;
    if (this.left) {
      leftCode = this.left.patchRepeatable();
    } else {
      this.insert(indexStart.end, "0");
      leftCode = "0";
    }
    const slice = this.getSliceSourceToken();
    const right = this.right;
    if (right) {
      this.overwrite(slice.start, slice.end, ", ");
      right.patch({ needsParens: true });
      if (leftCode !== "0") {
        this.insert(right.outerEnd, ` - ${leftCode}`);
      }
      if (this.isInclusive()) {
        this.insert(right.outerEnd, " + 1");
      }
    } else {
      this.overwrite(slice.start, slice.end, ", 9e9");
    }
    const indexEnd = this.getIndexEndSourceToken();
    this.remove(indexEnd.start, indexEnd.end);
  }
  isInclusive() {
    const slice = this.getSliceSourceToken();
    return slice.end - slice.start === "..".length;
  }
  getIndexStartSourceToken() {
    const tokens = this.context.sourceTokens;
    const index = tokens.indexOfTokenMatchingPredicate((token) => token.type === SourceType18.LBRACKET, this.expression.outerEndTokenIndex);
    if (!index || index.isAfter(this.contentEndTokenIndex)) {
      throw this.error(`could not find INDEX_START after slice expression`);
    }
    return notNull(tokens.tokenAtIndex(index));
  }
  getSliceSourceToken() {
    const tokens = this.context.sourceTokens;
    const { source } = this.context;
    const index = tokens.indexOfTokenMatchingPredicate((token) => {
      if (token.type !== SourceType18.RANGE) {
        return false;
      }
      const operator = source.slice(token.start, token.end);
      return operator === "..." || operator === "..";
    }, this.left ? this.left.outerEndTokenIndex : this.expression.outerEndTokenIndex);
    if (!index || index.isAfter(this.contentEndTokenIndex)) {
      throw this.error(`could not find '..' or '...' in slice`);
    }
    return notNull(tokens.tokenAtIndex(index));
  }
  getIndexEndSourceToken() {
    const tokens = this.context.sourceTokens;
    const index = tokens.lastIndexOfTokenMatchingPredicate((token) => token.type === SourceType18.RBRACKET, this.outerEndTokenIndex);
    if (!index || index.isBefore(this.contentStartTokenIndex)) {
      throw this.error(`could not find ']' ending slice`);
    }
    return notNull(tokens.tokenAtIndex(index));
  }
  statementNeedsParens() {
    return this.expression.statementShouldAddParens();
  }
};

// src/stages/main/patchers/SpreadPatcher.ts
import { SourceType as SourceType19 } from "coffee-lex";
import { Identifier as Identifier7 } from "decaffeinate-parser";
var SpreadPatcher = class extends NodePatcher {
  constructor(patcherContext, expression) {
    super(patcherContext);
    this.expression = expression;
  }
  initialize() {
    this.expression.setRequiresExpression();
  }
  setAssignee() {
    this.expression.setAssignee();
    super.setAssignee();
  }
  patchAsExpression() {
    const needsArrayFrom = this.needsArrayFrom();
    const isEllipsisOnLHS = this.firstToken().type === SourceType19.RANGE;
    if (!isEllipsisOnLHS) {
      this.insert(this.expression.outerStart, "...");
    }
    if (needsArrayFrom) {
      this.insert(this.expression.outerStart, "Array.from(");
    }
    this.expression.patch();
    if (!isEllipsisOnLHS) {
      this.remove(this.expression.outerEnd, this.contentEnd);
    }
    if (needsArrayFrom) {
      if (this.parent instanceof FunctionApplicationPatcher && this.parent.args.length === 1 && this.parent.args[0] === this) {
        this.insert(this.contentEnd, " || []");
      }
      this.insert(this.contentEnd, ")");
    }
  }
  needsArrayFrom() {
    if (this.options.useCS2) {
      return false;
    }
    if (this.isAssignee()) {
      return false;
    }
    if (this.expression.node instanceof Identifier7 && this.expression.node.data === "arguments") {
      return false;
    }
    this.addSuggestion(REMOVE_ARRAY_FROM);
    return true;
  }
};

// src/stages/main/patchers/AssignOpPatcher.ts
var MULTI_ASSIGN_SINGLE_LINE_MAX_LENGTH = 100;
var OBJECT_WITHOUT_KEYS_HELPER = `function __objectWithoutKeys__(object, keys) {
  const result = {...object};
  for (const k of keys) {
    delete result[keys];
  }
  return result;
}`;
var AssignOpPatcher = class extends NodePatcher {
  constructor(patcherContext, assignee, expression) {
    super(patcherContext);
    this.negated = false;
    this.assignee = assignee;
    this.expression = expression;
  }
  initialize() {
    this.assignee.setAssignee();
    this.assignee.setRequiresExpression();
    this.expression.setRequiresExpression();
  }
  negate() {
    this.negated = !this.negated;
  }
  patchAsExpression() {
    if (this.parent instanceof ObjectInitialiserPatcher) {
      this.patchAsObjectDestructureWithDefault();
      return;
    }
    this.markProtoAssignmentRepeatableIfNecessary();
    const shouldAddParens = !this.isAssignee() && (this.negated || this.willResultInSeqExpression() && this.parent instanceof FunctionApplicationPatcher || !this.isSurroundedByParentheses() && !(this.parent instanceof ReturnPatcher || this.parent instanceof DoOpPatcher || this.parent instanceof ConditionalPatcher && this.parent.condition === this && !this.parent.willPatchAsTernary()) && !this.implicitlyReturns());
    if (this.negated) {
      this.insert(this.innerStart, "!");
    }
    if (shouldAddParens) {
      this.insert(this.innerStart, "(");
    }
    if (canPatchAssigneeToJavaScript(this.assignee.node, this.options)) {
      this.patchSimpleAssignment();
    } else {
      this.addSuggestion(SIMPLIFY_COMPLEX_ASSIGNMENTS);
      const assignments = [];
      let expressionCode;
      if (this.expression.isRepeatable()) {
        expressionCode = this.expression.patchAndGetCode();
      } else {
        const fullExpression = this.expression.patchAndGetCode();
        expressionCode = this.claimFreeBinding();
        assignments.push(`${expressionCode} = ${fullExpression}`);
      }
      assignments.push(...this.generateAssignments(this.assignee, expressionCode, true));
      assignments.push(`${expressionCode}`);
      this.overwrite(this.contentStart, this.contentEnd, assignments.join(", "));
    }
    if (shouldAddParens) {
      this.appendDeferredSuffix(")");
    }
  }
  patchAsStatement() {
    this.markProtoAssignmentRepeatableIfNecessary();
    if (canPatchAssigneeToJavaScript(this.assignee.node, this.options)) {
      const shouldAddParens = this.assignee.statementShouldAddParens();
      if (shouldAddParens) {
        this.insert(this.contentStart, "(");
      }
      this.patchSimpleAssignment();
      if (shouldAddParens) {
        this.insert(this.contentEnd, ")");
      }
    } else {
      this.addSuggestion(SIMPLIFY_COMPLEX_ASSIGNMENTS);
      const assignments = this.generateAssignments(this.assignee, this.expression.patchAndGetCode(), this.expression.isRepeatable());
      this.overwriteWithAssignments(assignments);
    }
  }
  patchAsObjectDestructureWithDefault() {
    if (this.assignee instanceof MemberAccessOpPatcher && this.assignee.expression instanceof ThisPatcher) {
      this.insert(this.assignee.outerStart, `${this.assignee.getMemberName()}: `);
    }
    this.assignee.patch();
    this.expression.patch();
  }
  willResultInSeqExpression() {
    return this.willPatchAsExpression() && (!canPatchAssigneeToJavaScript(this.assignee.node, this.options) || this.assignee instanceof ArrayInitialiserPatcher);
  }
  patchSimpleAssignment() {
    const needsArrayFrom = this.shouldUseArrayFrom() && this.assignee instanceof ArrayInitialiserPatcher;
    this.assignee.patch();
    if (needsArrayFrom) {
      this.addSuggestion(REMOVE_ARRAY_FROM);
      this.insert(this.expression.outerStart, "Array.from(");
    }
    if (needsArrayFrom) {
      if (this.willPatchAsExpression()) {
        const expressionRepeatCode = this.expression.patchRepeatable();
        this.insert(this.expression.outerEnd, `), ${expressionRepeatCode}`);
      } else {
        this.expression.patch();
        this.insert(this.expression.outerEnd, `)`);
      }
    } else {
      this.expression.patch();
    }
  }
  overwriteWithAssignments(assignments) {
    let assignmentCode = assignments.join(", ");
    if (assignmentCode.length > MULTI_ASSIGN_SINGLE_LINE_MAX_LENGTH) {
      assignmentCode = assignments.join(`,
${this.getIndent(1)}`);
    }
    if (assignmentCode.startsWith("{")) {
      assignmentCode = `(${assignmentCode})`;
    }
    this.overwrite(this.contentStart, this.contentEnd, assignmentCode);
  }
  generateAssignments(patcher, ref, refIsRepeatable) {
    if (patcher instanceof ExpansionPatcher) {
      return [];
    } else if (patcher instanceof ElisionPatcher) {
      return [];
    } else if (canPatchAssigneeToJavaScript(patcher.node, this.options)) {
      const assigneeCode = patcher.patchAndGetCode();
      if (this.shouldUseArrayFrom() && patcher instanceof ArrayInitialiserPatcher) {
        this.addSuggestion(REMOVE_ARRAY_FROM);
        return [`${assigneeCode} = Array.from(${ref})`];
      } else {
        return [`${assigneeCode} = ${ref}`];
      }
    } else if (patcher instanceof SpreadPatcher) {
      return this.generateAssignments(patcher.expression, ref, refIsRepeatable);
    } else if (patcher instanceof ArrayInitialiserPatcher) {
      if (!refIsRepeatable) {
        const arrReference = this.claimFreeBinding("array");
        return [`${arrReference} = ${ref}`, ...this.generateAssignments(patcher, arrReference, true)];
      }
      const assignees = patcher.members;
      let lengthCode = null;
      const assignments = [];
      for (const [i, assignee] of assignees.entries()) {
        let valueCode;
        if (assignee instanceof ExpansionPatcher || assignee instanceof SpreadPatcher) {
          if (assignee instanceof SpreadPatcher && i < assignees.length - 1) {
            lengthCode = this.claimFreeBinding("adjustedLength");
            assignments.push(`${lengthCode} = Math.max(${ref}.length, ${assignees.length - 1})`);
          } else {
            lengthCode = `${ref}.length`;
          }
          valueCode = `${ref}.slice(${i}, ${lengthCode} - ${assignees.length - i - 1})`;
        } else if (typeof lengthCode === "string") {
          valueCode = `${ref}[${lengthCode} - ${assignees.length - i}]`;
        } else {
          valueCode = `${ref}[${i}]`;
        }
        assignments.push(...this.generateAssignments(assignee, valueCode, false));
      }
      return assignments;
    } else if (patcher instanceof ObjectInitialiserPatcher) {
      if (!refIsRepeatable) {
        const objReference = this.claimFreeBinding("obj");
        return [`${objReference} = ${ref}`, ...this.generateAssignments(patcher, objReference, true)];
      }
      const assignments = [];
      for (const member of patcher.members) {
        if (member instanceof ObjectInitialiserMemberPatcher) {
          const valueCode = `${ref}${this.accessFieldForObjectDestructure(member.key)}`;
          assignments.push(...this.generateAssignments(member.expression || member.key, valueCode, false));
        } else if (member instanceof AssignOpPatcher) {
          const valueCode = `${ref}${this.accessFieldForObjectDestructure(member.assignee)}`;
          assignments.push(...this.generateAssignments(member, valueCode, false));
        } else if (member instanceof SpreadPatcher) {
          const helper = this.registerHelper("__objectWithoutKeys__", OBJECT_WITHOUT_KEYS_HELPER);
          const omittedKeysCode = getObjectAssigneeKeys(patcher).map((key) => `'${key}'`).join(", ");
          assignments.push(...this.generateAssignments(member, `${helper}(${ref}, [${omittedKeysCode}])`, false));
        } else {
          throw this.error(`Unexpected object initializer member: ${member.node.type}`);
        }
      }
      return assignments;
    } else if (patcher instanceof SlicePatcher) {
      return [patcher.getSpliceCode(ref)];
    } else if (patcher instanceof AssignOpPatcher) {
      this.addSuggestion(SHORTEN_NULL_CHECKS);
      if (!refIsRepeatable) {
        const valReference = this.claimFreeBinding("val");
        return [`${valReference} = ${ref}`, ...this.generateAssignments(patcher, valReference, true)];
      }
      const defaultCode = patcher.expression.patchAndGetCode();
      const comparison = this.options.useCS2 ? "!== undefined" : "!= null";
      return this.generateAssignments(patcher.assignee, `${ref} ${comparison} ? ${ref} : ${defaultCode}`, false);
    } else {
      throw this.error(`Invalid assignee type: ${patcher.node.type}`);
    }
  }
  accessFieldForObjectDestructure(patcher) {
    if (patcher instanceof IdentifierPatcher) {
      return `.${patcher.node.data}`;
    } else if (patcher instanceof MemberAccessOpPatcher && patcher.expression instanceof ThisPatcher) {
      return `.${patcher.node.member.data}`;
    } else if (patcher instanceof StringPatcher) {
      return `[${patcher.patchAndGetCode()}]`;
    } else {
      throw this.error(`Unexpected object destructure expression: ${patcher.node.type}`);
    }
  }
  markProtoAssignmentRepeatableIfNecessary() {
    if (!(this.expression instanceof FunctionPatcher && containsSuperCall(this.expression.node))) {
      return;
    }
    const prototypeAssignPatchers = extractPrototypeAssignPatchers(this);
    if (!prototypeAssignPatchers) {
      return;
    }
    const { classRefPatcher, methodAccessPatcher } = prototypeAssignPatchers;
    classRefPatcher.setRequiresRepeatableExpression({
      parens: true,
      ref: "cls",
      forceRepeat: true
    });
    if (methodAccessPatcher instanceof DynamicMemberAccessOpPatcher) {
      methodAccessPatcher.indexingExpr.setRequiresRepeatableExpression({
        ref: "method",
        forceRepeat: true
      });
    }
  }
  shouldUseArrayFrom() {
    return !this.options.useCS2;
  }
};

// src/stages/main/patchers/AwaitPatcher.ts
var AwaitPatcher = class extends NodePatcher {
  constructor(patcherContext, expression) {
    super(patcherContext);
    this.expression = expression;
  }
  initialize() {
    this.awaits();
    this.expression.setRequiresExpression();
  }
  patchAsExpression({ needsParens = true } = {}) {
    this.expression.patch({ needsParens });
  }
};

// src/stages/main/patchers/AwaitReturnPatcher.ts
import { SourceType as SourceType20 } from "coffee-lex";
var AwaitReturnPatcher = class extends ReturnPatcher {
  initialize() {
    this.awaits();
    super.initialize();
  }
  patchAsStatement() {
    const awaitTokenIndex = this.contentStartTokenIndex;
    const returnTokenIndex = notNull(awaitTokenIndex.next());
    const awaitToken = notNull(this.sourceTokenAtIndex(awaitTokenIndex));
    const returnToken = notNull(this.sourceTokenAtIndex(returnTokenIndex));
    if (awaitToken.type !== SourceType20.IDENTIFIER || returnToken.type !== SourceType20.RETURN) {
      throw this.error("Unexpected token types for `await return`.");
    }
    this.remove(awaitToken.start, returnToken.start);
    super.patchAsStatement();
  }
};

// src/stages/main/patchers/BinaryOpPatcher.ts
import { SourceType as SourceType21 } from "coffee-lex";
var BinaryOpPatcher = class extends NodePatcher {
  constructor(patcherContext, left, right) {
    super(patcherContext);
    this.binaryOpNegated = false;
    this.left = left;
    this.right = right;
  }
  initialize() {
    this.left.setRequiresExpression();
    if (!this.rhsMayBeStatement()) {
      this.right.setRequiresExpression();
    }
  }
  rhsMayBeStatement() {
    return false;
  }
  negate() {
    this.binaryOpNegated = !this.binaryOpNegated;
  }
  isPure() {
    return this.left.isPure() && this.right.isPure();
  }
  patchAsExpression({ needsParens = false } = {}) {
    const addParens = needsParens && !this.isSurroundedByParentheses() || this.binaryOpNegated;
    if (this.binaryOpNegated) {
      this.insert(this.innerStart, "!");
    }
    if (addParens) {
      this.insert(this.innerStart, "(");
    }
    if (this.left instanceof BinaryOpPatcher) {
      this.left.patch({ needsParens: this.getOperator() !== this.left.getOperator() });
    } else {
      this.left.patch({ needsParens: true });
    }
    this.patchOperator();
    if (this.right instanceof BinaryOpPatcher) {
      this.right.patch({ needsParens: this.getOperator() !== this.right.getOperator() });
    } else {
      this.right.patch({ needsParens: true });
    }
    if (addParens) {
      this.insert(this.innerEnd, ")");
    }
  }
  patchOperator() {
  }
  getOperator() {
    return this.sourceOfToken(this.getOperatorToken());
  }
  getOperatorToken() {
    const operatorTokenIndex = this.indexOfSourceTokenBetweenPatchersMatching(this.left, this.right, this.operatorTokenPredicate());
    if (!operatorTokenIndex) {
      throw this.error("expected operator between binary operands");
    }
    return notNull(this.sourceTokenAtIndex(operatorTokenIndex));
  }
  operatorTokenPredicate() {
    return (token) => token.type === SourceType21.OPERATOR || token.type === SourceType21.EXISTENCE;
  }
  statementNeedsParens() {
    return this.left.statementShouldAddParens();
  }
};

// src/stages/main/patchers/BoolPatcher.ts
var BoolPatcher = class extends NodePatcher {
  patchAsExpression() {
    switch (this.getOriginalSource()) {
      case "off":
      case "no":
        this.overwrite(this.contentStart, this.contentEnd, "false");
        break;
      case "on":
      case "yes":
        this.overwrite(this.contentStart, this.contentEnd, "true");
        break;
    }
  }
  isRepeatable() {
    return true;
  }
};

// src/stages/main/patchers/BreakPatcher.ts
var BreakPatcher = class extends NodePatcher {
  patchAsStatement() {
  }
  canPatchAsExpression() {
    return false;
  }
};

// src/utils/getCompareOperator.ts
function getCompareOperator(operator, negated) {
  switch (operator) {
    case "==":
    case "is":
      return negated ? "!==" : "===";
    case "!=":
    case "isnt":
      return negated ? "===" : "!==";
    case "<":
      return negated ? ">=" : "<";
    case ">":
      return negated ? "<=" : ">";
    case "<=":
      return negated ? ">" : "<=";
    case ">=":
      return negated ? "<" : ">=";
    default:
      throw new Error(`unsupported equality/inequality type: ${operator}`);
  }
}

// src/utils/isCompareOpNegationUnsafe.ts
function isCompareOpNegationUnsafe(operator) {
  return ["<", ">", "<=", ">="].indexOf(operator) > -1;
}

// src/stages/main/patchers/ChainedComparisonOpPatcher.ts
var ChainedComparisonOpPatcher = class extends NodePatcher {
  constructor(patcherContext, operands) {
    super(patcherContext);
    this.negated = false;
    this.operands = operands;
  }
  initialize() {
    for (const operand of this.operands) {
      operand.setRequiresExpression();
    }
  }
  patchAsExpression({ needsParens = false } = {}) {
    const negateEntireExpression = this.shouldNegateEntireExpression();
    const addParens = negateEntireExpression || needsParens && !this.isSurroundedByParentheses();
    if (negateEntireExpression) {
      this.insert(this.contentStart, "!");
    }
    if (addParens) {
      this.insert(this.contentStart, "(");
    }
    const middle = this.getMiddleOperands();
    const negated = !negateEntireExpression && this.negated;
    const logicalOperator = negated ? "||" : "&&";
    for (const operand of middle) {
      operand.setRequiresRepeatableExpression({ parens: true, ref: "middle" });
    }
    for (const [i, operand] of this.operands.entries()) {
      operand.patch();
      const operator = this.node.operators[i];
      if (operator) {
        const replacement = getCompareOperator(operator.operator, negated);
        if (operator.operator !== replacement) {
          this.overwrite(operator.token.start, operator.token.end, replacement);
        }
      }
    }
    for (const operand of middle) {
      this.insert(operand.outerEnd, ` ${logicalOperator} ${operand.getRepeatCode()}`);
    }
    if (addParens) {
      this.insert(this.contentEnd, ")");
    }
  }
  shouldNegateEntireExpression() {
    return this.negated && this.node.operators.some((operator) => isCompareOpNegationUnsafe(operator.operator)) && !this.options.looseComparisonNegation;
  }
  getMiddleOperands() {
    return this.operands.slice(1, -1);
  }
  negate() {
    this.negated = !this.negated;
  }
  statementNeedsParens() {
    return this.operands[0].statementShouldAddParens();
  }
};

// src/stages/main/patchers/ClassAssignOpPatcher.ts
import { SourceType as SourceType23 } from "coffee-lex";

// src/stages/main/patchers/ClassBoundMethodFunctionPatcher.ts
var ClassBoundMethodFunctionPatcher = class extends FunctionPatcher {
  expectedArrowType() {
    return "=>";
  }
};

// src/stages/main/patchers/ClassPatcher.ts
import { SourceType as SourceType22 } from "coffee-lex";

// src/utils/getBindingCodeForMethod.ts
function getBindingCodeForMethod(method) {
  let accessCode;
  if (method.key instanceof IdentifierPatcher) {
    accessCode = `.${method.key.node.data}`;
  } else {
    accessCode = `[${method.key.getRepeatCode()}]`;
  }
  return `this${accessCode} = this${accessCode}.bind(this)`;
}

// src/utils/stripSharedIndent.ts
function stripSharedIndent(source) {
  let lines = source.split("\n");
  const commonIndent = getCommonIndent(lines);
  lines = lines.map((line) => {
    if (line.startsWith(commonIndent)) {
      return line.substr(commonIndent.length);
    }
    if (/^\s*$/.test(line)) {
      return "";
    }
    return line;
  });
  while (lines.length > 0 && lines[0].length === 0) {
    lines.shift();
  }
  while (lines.length > 0 && lines[lines.length - 1].length === 0) {
    lines.pop();
  }
  return lines.join("\n");
}
function getCommonIndent(lines) {
  let commonIndent = null;
  for (const line of lines) {
    const indent = getIndent(line, 0);
    if (indent === line) {
      continue;
    }
    if (commonIndent === null) {
      commonIndent = indent;
    } else {
      for (let i = 0; i < commonIndent.length; i++) {
        if (i >= indent.length || indent[i] !== commonIndent[i]) {
          commonIndent = commonIndent.substr(0, i);
          break;
        }
      }
    }
  }
  return commonIndent === null ? "" : commonIndent;
}

// src/utils/getInvalidConstructorErrorMessage.ts
function getInvalidConstructorErrorMessage(firstSentence) {
  return stripSharedIndent(`
    ${firstSentence}
    
    JavaScript requires all subclass constructors to call \`super\` and to do so
    before the first use of \`this\`, so the following cases cannot be converted
    automatically:
    * Constructors in subclasses that use \`this\` before \`super\`.
    * Constructors in subclasses that omit the \`super\` call.
    * Subclasses that use \`=>\` method syntax to automatically bind methods.
    
    To convert these cases to JavaScript anyway, remove the option
    --disallow-invalid-constructors when running decaffeinate.
  `);
}

// src/stages/main/patchers/ConstructorPatcher.ts
import { Class as Class3, This } from "decaffeinate-parser";
var ConstructorPatcher = class extends ObjectBodyMemberPatcher {
  constructor(patcherContext, assignee, expression) {
    super(patcherContext, assignee, expression);
    this._bindings = null;
    expression.disableImplicitReturns();
  }
  patch(options = {}) {
    this.checkForConstructorErrors();
    if (this.expression.body) {
      const linesToInsert = this.getLinesToInsert();
      const insertIndex = this.options.useCS2 ? this.getIndexOfSuperStatement() + 1 : 0;
      this.expression.body.insertStatementsAtIndex(linesToInsert, insertIndex);
      super.patch(options);
    } else {
      super.patch(options);
      const linesToInsert = this.getLinesToInsert();
      if (linesToInsert.length > 0) {
        const indent = this.getIndent();
        const bodyIndent = this.getIndent(1);
        const arrowToken = this.expression.getArrowToken();
        const fullLines = linesToInsert.map((line) => `${bodyIndent}${line}
`);
        const bodyCode = `{
${fullLines.join("")}${indent}}`;
        this.overwrite(arrowToken.start, this.expression.outerEnd, bodyCode);
      }
    }
  }
  getLinesToInsert() {
    let lines = [];
    lines = lines.concat(this.getBindings());
    return lines;
  }
  checkForConstructorErrors() {
    const errorMessage = this.getInvalidConstructorMessage();
    if (errorMessage) {
      if (!this.options.disallowInvalidConstructors) {
        this.addSuggestion(FIX_INVALID_CONSTRUCTOR);
      } else {
        throw this.error(getInvalidConstructorErrorMessage(errorMessage));
      }
    }
  }
  getInvalidConstructorMessage() {
    if (!this.getEnclosingClassPatcher().isSubclass()) {
      return null;
    }
    if (this.getBindings().length > 0 && !this.options.useCS2) {
      return "Cannot automatically convert a subclass that uses bound methods.";
    }
    const superIndex = this.getIndexOfSuperStatement();
    const thisIndex = this.getIndexOfFirstThisStatement();
    if (superIndex === -1) {
      return "Cannot automatically convert a subclass with a constructor that does not call super.";
    }
    if (thisIndex >= 0 && thisIndex <= superIndex) {
      return "Cannot automatically convert a subclass with a constructor that uses `this` before `super`.";
    }
    return null;
  }
  getBindings() {
    if (!this._bindings) {
      const boundMethods = this.getEnclosingClassBlockPatcher().boundInstanceMethods();
      const bindings = boundMethods.map(getBindingCodeForMethod);
      this._bindings = bindings;
    }
    return this._bindings;
  }
  getEnclosingClassPatcher() {
    const enclosingClassBlock = this.getEnclosingClassBlockPatcher();
    if (!(enclosingClassBlock.parent instanceof ClassPatcher)) {
      throw this.error("Expected grandparent of ConstructorPatcher to be ClassPatcher.");
    }
    return enclosingClassBlock.parent;
  }
  getEnclosingClassBlockPatcher() {
    if (!(this.parent instanceof ClassBlockPatcher)) {
      throw this.error("Expected parent of ConstructorPatcher to be ClassBlockPatcher.");
    }
    return this.parent;
  }
  getIndexOfSuperStatement() {
    if (!this.expression.body) {
      return -1;
    }
    const statements = this.expression.body.statements;
    for (let i = 0; i < statements.length; i++) {
      if (containsSuperCall(statements[i].node)) {
        return i;
      }
    }
    return -1;
  }
  getIndexOfFirstThisStatement() {
    if (!this.expression.body) {
      return -1;
    }
    const statements = this.expression.body.statements;
    for (let i = 0; i < statements.length; i++) {
      const usesThis = containsDescendant(statements[i].node, (child) => child instanceof This, {
        shouldStopTraversal: (child) => child instanceof Class3 || isFunction(child)
      });
      if (usesThis) {
        return i;
      }
    }
    return -1;
  }
  statementNeedsSemicolon() {
    return false;
  }
};

// src/stages/main/patchers/ClassBlockPatcher.ts
var ClassBlockPatcher = class extends BlockPatcher {
  static patcherClassForChildNode(node, property) {
    if (property === "statements" && node.type === "AssignOp") {
      return ClassAssignOpPatcher;
    }
    return null;
  }
  patch(options = {}) {
    for (const boundMethod of this.boundInstanceMethods()) {
      boundMethod.key.setRequiresRepeatableExpression();
    }
    super.patch(options);
    if (!this.hasConstructor()) {
      const boundMethods = this.boundInstanceMethods();
      if (boundMethods.length > 0) {
        const isSubclass = this.getClassPatcher().isSubclass();
        if (!this.options.useCS2) {
          if (isSubclass && !this.shouldAllowInvalidConstructors()) {
            throw this.error(getInvalidConstructorErrorMessage("Cannot automatically convert a subclass that uses bound methods."));
          } else if (isSubclass) {
            this.addSuggestion(FIX_INVALID_CONSTRUCTOR);
          }
        }
        const { source } = this.context;
        const insertionPoint = this.statements[0].outerStart;
        const methodIndent = adjustIndent(source, insertionPoint, 0);
        const methodBodyIndent = adjustIndent(source, insertionPoint, 1);
        let constructor = "";
        if (isSubclass) {
          constructor += `constructor(...args) {
`;
        } else {
          constructor += `constructor() {
`;
        }
        const bindingStatements = boundMethods.map((method) => `${methodBodyIndent}${getBindingCodeForMethod(method)};
`).join("");
        if (isSubclass) {
          const superStatement = `${methodBodyIndent}super(...args)
`;
          if (this.options.useCS2) {
            constructor += superStatement + bindingStatements;
          } else {
            constructor += bindingStatements + superStatement;
          }
        } else {
          constructor += bindingStatements;
        }
        constructor += `${methodIndent}}

${methodIndent}`;
        this.prependLeft(insertionPoint, constructor);
      }
    }
  }
  shouldAllowInvalidConstructors() {
    return !this.options.disallowInvalidConstructors;
  }
  getClassPatcher() {
    if (!(this.parent instanceof ClassPatcher)) {
      throw this.error("Expected class block parent to be a class.");
    }
    return this.parent;
  }
  canPatchAsExpression() {
    return false;
  }
  hasConstructor() {
    return this.statements.some((statement) => statement instanceof ConstructorPatcher);
  }
  boundInstanceMethods() {
    const boundMethods = [];
    for (const statement of this.statements) {
      if (statement instanceof ClassAssignOpPatcher && statement.isBoundInstanceMethod()) {
        boundMethods.push(statement);
      }
    }
    return boundMethods;
  }
};

// src/stages/main/patchers/ClassPatcher.ts
var ClassPatcher = class extends NodePatcher {
  constructor(patcherContext, nameAssignee, parent, body) {
    super(patcherContext);
    this.nameAssignee = nameAssignee;
    this.superclass = parent;
    this.body = body;
  }
  static patcherClassForChildNode(_node, property) {
    if (property === "body") {
      return ClassBlockPatcher;
    }
    return null;
  }
  initialize() {
    if (this.nameAssignee) {
      this.nameAssignee.setRequiresExpression();
    }
    if (this.superclass) {
      this.superclass.setRequiresExpression();
    }
  }
  patchAsStatement() {
    const hasParens = this.isSurroundedByParentheses();
    const anonymous = this.isAnonymous();
    if (anonymous && !hasParens) {
      this.insert(this.innerStart, "(");
    }
    this.patchAsExpression();
    if (anonymous && !hasParens) {
      this.insert(this.innerEnd, ")");
    }
  }
  patchAsExpression({ skipParens = false } = {}) {
    const needsAssignment = this.nameAssignee && (this.isNamespaced() || this.isNameAlreadyDeclared() || this.willPatchAsExpression());
    const needsParens = !skipParens && needsAssignment && this.willPatchAsExpression() && !this.isSurroundedByParentheses();
    if (needsParens) {
      this.insert(this.contentStart, "(");
    }
    if (needsAssignment && this.nameAssignee) {
      const classToken = this.getClassToken();
      this.remove(classToken.start, this.nameAssignee.outerStart);
      const name = this.getName();
      if (name) {
        this.insert(this.nameAssignee.outerEnd, ` = class ${name}`);
      } else {
        this.insert(this.nameAssignee.outerEnd, ` = class`);
      }
    }
    if (this.nameAssignee) {
      this.nameAssignee.patch();
    }
    if (this.superclass) {
      this.superclass.patch();
    }
    if (!this.body) {
      this.insert(this.innerEnd, " {}");
    } else {
      this.insert(this.getBraceInsertionOffset(), " {");
      this.body.patch({ leftBrace: false });
    }
    if (needsParens) {
      this.insert(this.innerEnd, ")");
    }
  }
  statementNeedsSemicolon() {
    return this.isAnonymous() || this.isNamespaced();
  }
  statementNeedsParens() {
    return this.isAnonymous();
  }
  getClassToken() {
    const tokens = this.context.sourceTokens;
    const classSourceToken = notNull(tokens.tokenAtIndex(this.contentStartTokenIndex));
    if (classSourceToken.type !== SourceType22.CLASS) {
      throw this.error(`expected CLASS token but found ${SourceType22[classSourceToken.type]}`, classSourceToken.start, classSourceToken.end);
    }
    return classSourceToken;
  }
  isAnonymous() {
    return this.nameAssignee === null;
  }
  isNamespaced() {
    return !this.isAnonymous() && !(this.nameAssignee instanceof IdentifierPatcher);
  }
  isNameAlreadyDeclared() {
    const name = this.getName();
    return this.nameAssignee !== null && name !== null && this.getScope().getBinding(name) !== this.nameAssignee.node;
  }
  getName() {
    const { nameAssignee } = this;
    let name;
    if (nameAssignee instanceof IdentifierPatcher) {
      name = nameAssignee.node.data;
    } else if (nameAssignee instanceof MemberAccessOpPatcher) {
      name = nameAssignee.node.member.data;
    } else {
      name = null;
    }
    if (name !== null && isForbiddenJsName(name)) {
      name = `_${name}`;
    }
    return name;
  }
  isSubclass() {
    return this.superclass !== null;
  }
  getBraceInsertionOffset() {
    if (this.superclass) {
      return this.superclass.outerEnd;
    }
    if (this.nameAssignee) {
      return this.nameAssignee.outerEnd;
    }
    return this.getClassToken().end;
  }
};

// src/stages/main/patchers/ClassAssignOpPatcher.ts
var ClassAssignOpPatcher = class extends ObjectBodyMemberPatcher {
  static patcherClassForChildNode(node, property) {
    if (property === "expression" && node.type === "BoundFunction") {
      return ClassBoundMethodFunctionPatcher;
    }
    return null;
  }
  statementNeedsSemicolon() {
    return !this.isMethod();
  }
  patchAsExpression() {
    this.markKeyRepeatableIfNecessary();
    if (this.isStaticMethod()) {
      this.insert(this.key.outerStart, "static ");
    }
    super.patchAsExpression();
    if (this.isStaticMethod()) {
      let replaceEnd;
      if (this.key instanceof MemberAccessOpPatcher) {
        replaceEnd = this.key.getMemberNameSourceToken().start;
      } else if (this.key instanceof DynamicMemberAccessOpPatcher) {
        replaceEnd = this.key.expression.outerEnd;
      } else {
        throw this.error("Unexpected static method key type.");
      }
      this.remove(this.key.outerStart, replaceEnd);
    }
  }
  markKeyRepeatableIfNecessary() {
    if (this.expression instanceof FunctionPatcher && containsSuperCall(this.expression.node)) {
      if (this.isStaticMethod()) {
        if (this.key instanceof DynamicMemberAccessOpPatcher) {
          this.key.indexingExpr.setRequiresRepeatableExpression({
            ref: "method",
            forceRepeat: true
          });
        }
      } else {
        this.key.setRequiresRepeatableExpression({
          ref: "method",
          forceRepeat: this.key instanceof StringPatcher && this.key.expressions.length > 0
        });
      }
    }
  }
  patchKey() {
    if (this.isStaticMethod()) {
      this.key.patch();
    } else {
      super.patchKey();
    }
  }
  patchAsProperty() {
    if (!this.expression) {
      throw this.error("Expected value expression for class assign op.");
    }
    const colonIndex = this.indexOfSourceTokenBetweenPatchersMatching(this.key, this.expression, (token) => token.type === SourceType23.COLON);
    if (!colonIndex) {
      throw this.error("expected a colon between the key and expression of a class property");
    }
    const colonToken = notNull(this.sourceTokenAtIndex(colonIndex));
    this.overwrite(colonToken.start, colonToken.end, " =");
    this.patchExpression();
  }
  isStaticMethod() {
    if (!(this.key instanceof MemberAccessOpPatcher) && !(this.key instanceof DynamicMemberAccessOpPatcher)) {
      return false;
    }
    const memberObject = this.key.expression;
    if (memberObject instanceof ThisPatcher) {
      return true;
    }
    const className = this.getEnclosingClassPatcher().nameAssignee;
    return className instanceof IdentifierPatcher && memberObject instanceof IdentifierPatcher && className.node.data === className.node.data;
  }
  getEnclosingClassPatcher() {
    const enclosingClassPatcher = notNull(this.parent).parent;
    if (!(enclosingClassPatcher instanceof ClassPatcher)) {
      throw this.error("Expected parent's parent to be a class.");
    }
    return enclosingClassPatcher;
  }
  isBoundInstanceMethod() {
    if (!this.expression) {
      throw this.error("Expected value expression for class assign op.");
    }
    return !this.isStaticMethod() && (this.expression.node.type === "BoundFunction" || this.expression.node.type === "BoundGeneratorFunction");
  }
  isMethod() {
    return this.expression instanceof ManuallyBoundFunctionPatcher || super.isMethod();
  }
};

// src/stages/main/patchers/CompoundAssignOpPatcher.ts
import { SourceType as SourceType24 } from "coffee-lex";

// src/utils/nodeContainsSoakOperation.ts
import {
  SoakedDynamicMemberAccessOp as SoakedDynamicMemberAccessOp2,
  SoakedFunctionApplication as SoakedFunctionApplication2,
  SoakedMemberAccessOp as SoakedMemberAccessOp2
} from "decaffeinate-parser";
function nodeContainsSoakOperation(node) {
  return containsDescendant(node, (child) => child instanceof SoakedDynamicMemberAccessOp2 || child instanceof SoakedFunctionApplication2 || child instanceof SoakedMemberAccessOp2);
}

// src/stages/main/patchers/CompoundAssignOpPatcher.ts
var CompoundAssignOpPatcher = class extends AssignOpPatcher {
  getOperatorToken() {
    const operatorIndex = this.indexOfSourceTokenBetweenPatchersMatching(this.assignee, this.expression, (token) => token.type === SourceType24.OPERATOR);
    if (!operatorIndex) {
      throw this.error(`expected OPERATOR token between assignee and expression`, this.assignee.outerEnd, this.expression.outerStart);
    }
    return notNull(this.sourceTokenAtIndex(operatorIndex));
  }
  statementNeedsParens() {
    return this.assignee.statementShouldAddParens();
  }
  lhsHasSoakOperation() {
    return nodeContainsSoakOperation(this.assignee.node);
  }
};

// src/stages/main/patchers/ContinuePatcher.ts
var ContinuePatcher = class extends NodePatcher {
  patchAsStatement() {
  }
  canPatchAsExpression() {
    return false;
  }
};

// src/stages/main/patchers/CSXElementPatcher.ts
import { SourceType as SourceType25 } from "coffee-lex";
var CSXElementPatcher = class extends NodePatcher {
  constructor(patcherContext, properties, children) {
    super(patcherContext);
    this.properties = properties;
    this.children = children;
  }
  initialize() {
    for (const property of this.properties) {
      property.setRequiresExpression();
    }
    for (const child of this.children) {
      if (child) {
        child.setRequiresExpression();
      }
    }
  }
  patchAsExpression() {
    for (const property of this.properties) {
      if (property instanceof StringPatcher && property.expressions.length > 0) {
        this.patchStringProperty(property);
      } else {
        property.patch();
      }
    }
    for (const child of this.children) {
      if (child) {
        child.patch();
      }
    }
  }
  patchStringProperty(property) {
    const prevIndex = property.contentStartTokenIndex.previous();
    if (!prevIndex) {
      throw this.error("Expected index before string property.");
    }
    const prevToken = this.sourceTokenAtIndex(prevIndex);
    if (prevToken && prevToken.type === SourceType25.OPERATOR && this.context.source.slice(prevToken.start, prevToken.end) === "=") {
      this.insert(property.outerStart, "{");
      property.patch();
      this.insert(property.outerEnd, "}");
    } else {
      property.patch();
    }
  }
};

// src/stages/main/patchers/DebuggerPatcher.ts
var DebuggerPatcher = class extends NodePatcher {
  patchAsStatement() {
  }
  canPatchAsExpression() {
    return false;
  }
};

// src/stages/main/patchers/EqualityPatcher.ts
import { SourceType as SourceType26 } from "coffee-lex";
var EqualityPatcher = class extends BinaryOpPatcher {
  constructor() {
    super(...arguments);
    this.negated = false;
  }
  patchOperator() {
    const compareToken = this.getCompareToken();
    this.overwrite(compareToken.start, compareToken.end, this.getCompareOperator());
  }
  getCompareOperator() {
    const token = this.getCompareToken();
    return getCompareOperator(this.sourceOfToken(token), this.negated);
  }
  getCompareToken() {
    const { left, right } = this;
    const compareTokenIndex = this.indexOfSourceTokenBetweenPatchersMatching(left, right, (token) => token.type === SourceType26.OPERATOR);
    if (!compareTokenIndex) {
      throw this.error("expected OPERATOR token but none was found", left.outerEnd, right.outerStart);
    }
    return notNull(this.sourceTokenAtIndex(compareTokenIndex));
  }
  negate() {
    if (isCompareOpNegationUnsafe(this.sourceOfToken(this.getCompareToken())) && !this.options.looseComparisonNegation) {
      return super.negate();
    }
    this.negated = !this.negated;
  }
};

// src/stages/main/patchers/ExistsOpCompoundAssignOpPatcher.ts
var ExistsOpCompoundAssignOpPatcher = class extends CompoundAssignOpPatcher {
  patchAsExpression({ needsParens = false } = {}) {
    this.addSuggestion(SHORTEN_NULL_CHECKS);
    const shouldAddParens = this.negated || needsParens && !this.isSurroundedByParentheses();
    if (this.negated) {
      this.insert(this.contentStart, "!");
    }
    if (shouldAddParens) {
      this.insert(this.contentStart, "(");
    }
    let assigneeAgain;
    if (this.needsTypeofCheck()) {
      this.insert(this.assignee.outerStart, `typeof `);
      assigneeAgain = this.assignee.patchRepeatable({ isForAssignment: true });
      this.insert(this.assignee.outerEnd, ` !== 'undefined' && ${assigneeAgain} !== null ? ${assigneeAgain}`);
    } else {
      assigneeAgain = this.assignee.patchRepeatable({ isForAssignment: true });
      this.insert(this.assignee.outerEnd, ` != null ? ${assigneeAgain}`);
    }
    const operator = this.getOperatorToken();
    this.overwrite(operator.start, operator.end, `: (${assigneeAgain} =`);
    this.expression.patch();
    this.insert(this.expression.outerEnd, ")");
    if (shouldAddParens) {
      this.insert(this.contentEnd, ")");
    }
  }
  patchAsStatement() {
    if (this.lhsHasSoakOperation()) {
      this.patchAsExpression();
      return;
    }
    this.addSuggestion(SHORTEN_NULL_CHECKS);
    let assigneeAgain;
    if (this.needsTypeofCheck()) {
      this.insert(this.assignee.outerStart, `if (typeof `);
      assigneeAgain = this.assignee.patchRepeatable({ isForAssignment: true });
      this.insert(this.assignee.outerEnd, ` === 'undefined' || ${assigneeAgain} === null) {`);
    } else {
      this.insert(this.assignee.outerStart, `if (`);
      assigneeAgain = this.assignee.patchRepeatable({ isForAssignment: true });
      this.insert(this.assignee.outerEnd, ` == null) {`);
    }
    const operator = this.getOperatorToken();
    this.overwrite(operator.start, operator.end, `${assigneeAgain} =`);
    this.expression.patch();
    this.insert(this.expression.outerEnd, `; }`);
  }
  needsTypeofCheck() {
    return this.assignee.mayBeUnboundReference();
  }
  statementNeedsParens() {
    return false;
  }
};

// src/stages/main/patchers/ExistsOpPatcher.ts
var ExistsOpPatcher = class extends BinaryOpPatcher {
  rhsMayBeStatement() {
    return true;
  }
  setExpression(force) {
    this.right.setRequiresExpression();
    return super.setExpression(force);
  }
  patchAsExpression({ needsParens = false } = {}) {
    this.addSuggestion(SHORTEN_NULL_CHECKS);
    const addParens = needsParens && !this.isSurroundedByParentheses() || this.binaryOpNegated;
    if (this.binaryOpNegated) {
      this.insert(this.contentStart, "!");
    }
    if (addParens) {
      this.insert(this.contentStart, "(");
    }
    const needsTypeofCheck = this.left.mayBeUnboundReference();
    if (needsTypeofCheck) {
      this.insert(this.contentStart, `typeof `);
      const leftAgain = this.left.patchRepeatable({ parens: true, ref: "left" });
      this.overwrite(this.left.outerEnd, this.right.outerStart, ` !== 'undefined' && ${leftAgain} !== null ? ${leftAgain} : `);
    } else {
      const leftAgain = this.left.patchRepeatable({ parens: true, ref: "left" });
      this.overwrite(this.left.outerEnd, this.right.outerStart, ` != null ? ${leftAgain} : `);
    }
    this.right.patch();
    if (addParens) {
      this.insert(this.contentEnd, ")");
    }
  }
  patchAsStatement() {
    this.addSuggestion(SHORTEN_NULL_CHECKS);
    const needsTypeofCheck = this.left.mayBeUnboundReference();
    this.insert(this.contentStart, `if (`);
    if (needsTypeofCheck) {
      const leftAgain = this.left.patchRepeatable();
      this.insert(this.contentStart, `typeof `);
      this.overwrite(this.left.outerEnd, this.right.outerStart, ` === 'undefined' || ${leftAgain} === null) { `);
    } else {
      this.left.patch();
      this.overwrite(this.left.outerEnd, this.right.outerStart, ` == null) { `);
    }
    this.right.patch();
    this.insert(this.innerEnd, ` }`);
  }
  statementNeedsParens() {
    return false;
  }
};

// src/stages/main/patchers/ExpOpPatcher.ts
var ExpOpPatcher = class extends BinaryOpPatcher {
  patchAsExpression() {
    this.insert(this.contentStart, `Math.pow(`);
    this.left.patch();
    this.overwrite(this.left.outerEnd, this.right.outerStart, ", ");
    this.right.patch();
    this.insert(this.contentEnd, `)`);
  }
  statementNeedsParens() {
    return false;
  }
};

// src/stages/main/patchers/ExportDefaultDeclarationPatcher.ts
var ExportDefaultDeclarationPatcher = class extends NodePatcher {
  constructor(patcherContext, expression) {
    super(patcherContext);
    this.expression = expression;
  }
  patchAsStatement() {
    this.expression.patch();
  }
};

// src/stages/main/patchers/ExportNamedDeclarationPatcher.ts
var ExportNamedDeclarationPatcher = class extends NodePatcher {
  constructor(patcherContext, expression) {
    super(patcherContext);
    this.expression = expression;
  }
  patchAsStatement() {
    if (this.expression instanceof AssignOpPatcher) {
      const exportToken = this.firstToken();
      this.insert(exportToken.end, " var");
    }
    this.expression.patch();
  }
};

// src/stages/main/patchers/ExtendsOpPatcher.ts
import { SourceType as SourceType27 } from "coffee-lex";
var EXTENDS_HELPER = `
function __extends__(child, parent) {
  Object.getOwnPropertyNames(parent).forEach(
    name => child[name] = parent[name]
  );
  child.prototype = Object.create(parent.prototype);
  child.__super__ = parent.prototype;
  return child;
}
`;
var ExtendsOpPatcher = class extends BinaryOpPatcher {
  patchAsExpression() {
    const helper = this.registerHelper("__extends__", EXTENDS_HELPER);
    this.insert(this.left.outerStart, `${helper}(`);
    this.left.patch();
    this.overwrite(this.left.outerEnd, this.right.outerStart, ", ");
    this.right.patch();
    this.insert(this.right.outerEnd, ")");
  }
  statementNeedsParens() {
    return false;
  }
  operatorTokenPredicate() {
    return (token) => token.type === SourceType27.EXTENDS;
  }
};

// src/stages/main/patchers/FloorDivideOpCompoundAssignOpPatcher.ts
var FloorDivideOpCompoundAssignOpPatcher = class extends CompoundAssignOpPatcher {
  patchAsExpression({ needsParens = false } = {}) {
    const shouldAddParens = this.negated || needsParens && !this.isSurroundedByParentheses();
    if (this.negated) {
      this.insert(this.contentStart, "!");
    }
    if (shouldAddParens) {
      this.insert(this.contentStart, "(");
    }
    const assigneeAgain = this.assignee.patchRepeatable({ isForAssignment: true });
    this.overwrite(this.assignee.outerEnd, this.expression.outerStart, ` = Math.floor(${assigneeAgain} / `);
    this.expression.patch({ needsParens: true });
    this.insert(this.expression.outerEnd, ")");
    if (shouldAddParens) {
      this.insert(this.contentEnd, ")");
    }
  }
  patchAsStatement() {
    this.patchAsExpression();
  }
};

// src/stages/main/patchers/FloorDivideOpPatcher.ts
var FloorDivideOpPatcher = class extends BinaryOpPatcher {
  patchAsExpression() {
    const operator = this.getOperatorToken();
    this.insert(this.contentStart, "Math.floor(");
    this.left.patch({ needsParens: true });
    this.overwrite(operator.start, operator.end, "/");
    this.right.patch({ needsParens: true });
    this.insert(this.contentEnd, ")");
  }
  statementNeedsParens() {
    return false;
  }
};

// src/stages/main/patchers/ForInPatcher.ts
import { traverse as traverse5, Float, Int as Int3, Number, UnaryNegateOp } from "decaffeinate-parser";

// src/stages/main/patchers/ForPatcher.ts
import { SourceType as SourceType28 } from "coffee-lex";

// src/utils/getAssigneeBindings.ts
import {
  ArrayInitialiser as ArrayInitialiser4,
  AssignOp as AssignOp4,
  Identifier as Identifier8,
  ObjectInitialiser as ObjectInitialiser5,
  ObjectInitialiserMember as ObjectInitialiserMember6,
  Rest as Rest3,
  Spread as Spread3
} from "decaffeinate-parser";
function getAssigneeBindings(node) {
  if (node instanceof Identifier8) {
    return [node.data];
  } else if (node instanceof ArrayInitialiser4) {
    const bindings = [];
    for (const member of node.members) {
      bindings.push(...getAssigneeBindings(member));
    }
    return bindings;
  } else if (node instanceof ObjectInitialiser5) {
    const bindings = [];
    for (const member of node.members) {
      bindings.push(...getAssigneeBindings(member));
    }
    return bindings;
  } else if (node instanceof ObjectInitialiserMember6) {
    return getAssigneeBindings(node.expression || node.key);
  } else if (node instanceof AssignOp4) {
    return getAssigneeBindings(node.assignee);
  } else if (node instanceof Spread3 || node instanceof Rest3) {
    return getAssigneeBindings(node.expression);
  }
  return [];
}

// src/stages/main/patchers/LoopPatcher.ts
var LoopPatcher = class extends NodePatcher {
  constructor(patcherContext, body) {
    super(patcherContext);
    this._resultArrayBinding = null;
    this._resultArrayElementBinding = null;
    this.body = body;
  }
  initialize() {
    getEnclosingScopeBlock(this).markIIFEPatcherDescendant(this);
  }
  patchAsExpression() {
    const baseIndent = notNull(this.parent).getIndent(0);
    const iifeBodyIndent = this.getLoopIndent();
    if (this.body !== null) {
      this.body.setShouldPatchInline(false);
      this.body.setImplicitlyReturns();
      this.body.setIndent(this.getLoopBodyIndent());
    }
    const resultBinding = this.getResultArrayBinding();
    this.patchInIIFE(() => {
      this.insert(this.innerStart, `
${iifeBodyIndent}${resultBinding} = [];
${iifeBodyIndent}`);
      this.patchAsStatement();
      this.insert(this.innerEnd, `
${iifeBodyIndent}return ${resultBinding};
${baseIndent}`);
    });
  }
  getLoopIndent() {
    if (this.willPatchAsExpression()) {
      return notNull(this.parent).getIndent(1);
    } else {
      return this.getIndent();
    }
  }
  getOuterLoopBodyIndent() {
    return this.getLoopIndent() + this.getProgramIndentString();
  }
  getLoopBodyIndent() {
    throw this.error(`'getLoopBodyIndent' must be overridden in subclasses`);
  }
  patchPossibleNewlineAfterLoopHeader(loopHeaderEndIndex) {
    if (this.body && this.shouldConvertInlineBodyToNonInline()) {
      this.overwrite(loopHeaderEndIndex, this.body.contentStart, `
`);
    }
  }
  patchBody() {
    if (this.body) {
      if (this.shouldConvertInlineBodyToNonInline()) {
        this.body.insert(this.body.outerStart, this.getLoopBodyIndent());
      }
      this.body.patch({ leftBrace: false, rightBrace: false });
    }
  }
  shouldConvertInlineBodyToNonInline() {
    return this.willPatchAsExpression() && this.body !== null && this.body.node.inline;
  }
  canHandleImplicitReturn() {
    return this.willPatchAsIIFE();
  }
  willPatchAsIIFE() {
    throw this.error(`'willPatchAsIIFE' must be overridden in subclasses`);
  }
  implicitReturnWillBreak() {
    return false;
  }
  patchImplicitReturnStart(patcher) {
    if (!patcher.canPatchAsExpression()) {
      return;
    }
    patcher.setRequiresExpression();
    this.insert(patcher.outerStart, `${this.getResultArrayBinding()}.push(`);
  }
  patchImplicitReturnEnd(patcher) {
    if (!patcher.canPatchAsExpression()) {
      return;
    }
    this.insert(patcher.outerEnd, `)`);
  }
  getEmptyImplicitReturnCode() {
    return `${this.getResultArrayBinding()}.push(undefined)`;
  }
  getResultArrayBinding() {
    if (!this._resultArrayBinding) {
      this._resultArrayBinding = this.claimFreeBinding("result");
    }
    return this._resultArrayBinding;
  }
  getResultArrayElementBinding() {
    if (!this._resultArrayElementBinding) {
      this._resultArrayElementBinding = this.claimFreeBinding("item");
    }
    return this._resultArrayElementBinding;
  }
  statementNeedsSemicolon() {
    return false;
  }
};

// src/stages/main/patchers/ForPatcher.ts
var ForPatcher = class extends LoopPatcher {
  constructor(patcherContext, keyAssignee, valAssignee, target, filter, body) {
    super(patcherContext, body);
    this._filterCode = null;
    this._targetCode = null;
    this._indexBinding = null;
    this._targetReference = null;
    this.keyAssignee = keyAssignee;
    this.valAssignee = valAssignee;
    this.target = target;
    this.filter = filter;
  }
  initialize() {
    super.initialize();
    if (this.keyAssignee) {
      this.keyAssignee.setAssignee();
      this.keyAssignee.setRequiresExpression();
    }
    if (this.valAssignee) {
      this.valAssignee.setAssignee();
      this.valAssignee.setRequiresExpression();
    }
    this.target.setRequiresExpression();
    if (this.filter) {
      this.filter.setRequiresExpression();
    }
  }
  getIIFEAssignments() {
    if (this.willPatchAsIIFE()) {
      const iifeAssignments = [];
      if (this.keyAssignee) {
        iifeAssignments.push(...getAssigneeBindings(this.keyAssignee.node));
      }
      if (this.valAssignee) {
        iifeAssignments.push(...getAssigneeBindings(this.valAssignee.node));
      }
      return iifeAssignments;
    } else {
      return [];
    }
  }
  getFilterCode() {
    const filter = this.filter;
    if (!filter) {
      return null;
    }
    if (!this._filterCode) {
      this._filterCode = filter.patchAndGetCode({ needsParens: false });
    }
    return this._filterCode;
  }
  getLoopBodyIndent() {
    if (this.filter) {
      return this.getOuterLoopBodyIndent() + this.getProgramIndentString();
    } else {
      return this.getOuterLoopBodyIndent();
    }
  }
  patchBodyAndFilter() {
    const filterCode = this.getFilterCode();
    if (this.body) {
      if (filterCode) {
        this.body.insertLineBefore(`if (${filterCode}) {`, this.getOuterLoopBodyIndent());
        this.patchBody();
        this.body.insertLineAfter("}", this.getOuterLoopBodyIndent());
        this.body.insertLineAfter("}", this.getLoopIndent());
      } else {
        this.patchBody();
        this.body.insertLineAfter("}", this.getLoopIndent());
      }
    } else {
      if (filterCode) {
        this.insert(this.contentEnd, `if (${filterCode}) {} }`);
      } else {
        this.insert(this.contentEnd, `}`);
      }
    }
  }
  getRelationToken() {
    let searchStart = -1;
    if (this.keyAssignee) {
      searchStart = Math.max(this.keyAssignee.outerEnd);
    }
    if (this.valAssignee) {
      searchStart = Math.max(this.valAssignee.outerEnd);
    }
    if (searchStart === -1) {
      throw this.error("Expected to find a good starting point to search for relation token.");
    }
    const tokenIndex = this.indexOfSourceTokenBetweenSourceIndicesMatching(searchStart, this.target.outerStart, (token) => token.type === SourceType28.RELATION || token.type === SourceType28.IDENTIFIER);
    if (!tokenIndex) {
      throw this.error(`cannot find relation keyword in 'for' loop`);
    }
    return notNull(this.sourceTokenAtIndex(tokenIndex));
  }
  getIndexBinding() {
    if (!this._indexBinding) {
      this._indexBinding = this.computeIndexBinding();
    }
    return this._indexBinding;
  }
  computeIndexBinding() {
    const keyAssignee = this.keyAssignee;
    if (keyAssignee) {
      if (keyAssignee instanceof MemberAccessOpPatcher) {
        return `this.${keyAssignee.member.node.data}`;
      } else if (keyAssignee instanceof IdentifierPatcher) {
        return this.slice(keyAssignee.contentStart, keyAssignee.contentEnd);
      } else {
        throw keyAssignee.error(`expected loop index to be an identifier or this-assignment`);
      }
    } else {
      return this.claimFreeBinding(this.indexBindingCandidates());
    }
  }
  isThisAssignIndexBinding() {
    return this.keyAssignee instanceof MemberAccessOpPatcher;
  }
  indexBindingCandidates() {
    return ["i", "j", "k"];
  }
  removeThenToken() {
    const searchStart = this.getLoopHeaderEnd();
    let searchEnd;
    if (this.body) {
      searchEnd = this.body.outerStart;
    } else {
      const nextToken = this.nextSemanticToken();
      if (nextToken) {
        searchEnd = nextToken.end;
      } else {
        searchEnd = this.contentEnd;
      }
    }
    const index = this.indexOfSourceTokenBetweenSourceIndicesMatching(searchStart, searchEnd, (token) => token.type === SourceType28.THEN);
    if (index) {
      const thenToken = notNull(this.sourceTokenAtIndex(index));
      const nextIndex = index.next();
      const nextToken = nextIndex && this.sourceTokenAtIndex(nextIndex);
      if (nextToken) {
        this.remove(thenToken.start, nextToken.start);
      } else {
        this.remove(thenToken.start, thenToken.end);
      }
    }
  }
  getLoopHeaderEnd() {
    return Math.max(this.filter ? this.filter.outerEnd : -1, this.target.outerEnd);
  }
  getTargetCode() {
    this.computeTargetCodeIfNecessary();
    return notNull(this._targetCode);
  }
  getTargetReference() {
    this.computeTargetCodeIfNecessary();
    return notNull(this._targetReference);
  }
  computeTargetCodeIfNecessary() {
    if (!this._targetReference || !this._targetCode) {
      this._targetCode = this.target.patchAndGetCode();
      if (this.requiresExtractingTarget()) {
        this._targetReference = this.claimFreeBinding(this.targetBindingCandidate());
      } else {
        this._targetReference = this._targetCode;
      }
    }
  }
};

// src/stages/main/patchers/RangePatcher.ts
import { SourceType as SourceType29 } from "coffee-lex";
import { Int as Int2 } from "decaffeinate-parser";
var RANGE_HELPER = `function __range__(left, right, inclusive) {
  let range = [];
  let ascending = left < right;
  let end = !inclusive ? right : ascending ? right + 1 : right - 1;
  for (let i = left; ascending ? i < end : i > end; ascending ? i++ : i--) {
    range.push(i);
  }
  return range;
}`;
var MAXIMUM_LITERAL_RANGE_ELEMENTS = 21;
var RangePatcher = class extends BinaryOpPatcher {
  patchAsExpression() {
    if (this.canBecomeLiteralArray()) {
      this.patchAsLiteralArray();
    } else {
      this.patchAsIIFE();
    }
  }
  patchAsLiteralArray() {
    if (!(this.left.node instanceof Int2) || !(this.right.node instanceof Int2)) {
      throw this.error("Expected ints on both sides for a literal array.");
    }
    const start = this.left.node.data;
    let end = this.right.node.data;
    const inclusive = this.isInclusive();
    const ascending = start < end;
    if (inclusive) {
      end += ascending ? 1 : -1;
    }
    let list = "";
    for (let i = start; ascending ? i < end : i > end; ascending ? i++ : i--) {
      const isLast = ascending ? i === end - 1 : i === end + 1;
      if (isLast) {
        list += `${i}`;
      } else {
        list += `${i}, `;
      }
    }
    this.overwrite(this.contentStart, this.contentEnd, `[${list}]`);
  }
  patchAsIIFE() {
    const helper = this.registerHelper("__range__", RANGE_HELPER);
    this.overwrite(this.contentStart, this.left.outerStart, `${helper}(`);
    this.left.patch();
    this.overwrite(this.left.outerEnd, this.right.outerStart, ", ");
    this.right.patch();
    this.overwrite(this.right.outerEnd, this.contentEnd, `, ${JSON.stringify(this.isInclusive())})`);
  }
  canBecomeLiteralArray() {
    const range = this.getLiteralRange();
    if (!range) {
      return false;
    }
    const [first, last] = range;
    return Math.abs(last - first) <= MAXIMUM_LITERAL_RANGE_ELEMENTS;
  }
  getLiteralRange() {
    const left = this.left.node;
    const right = this.right.node;
    if (!(left instanceof Int2) || !(right instanceof Int2)) {
      return null;
    }
    const first = left.data;
    const last = right.data;
    if (first < last) {
      return [first, last + (this.isInclusive() ? 1 : 0)];
    } else {
      return [first, last - (this.isInclusive() ? 1 : 0)];
    }
  }
  isInclusive() {
    return this.node.isInclusive;
  }
  operatorTokenPredicate() {
    return (token) => token.type === SourceType29.RANGE;
  }
};

// src/stages/main/patchers/UnaryOpPatcher.ts
var UnaryOpPatcher = class extends NodePatcher {
  constructor(patcherContext, expression) {
    super(patcherContext);
    this.expression = expression;
  }
  initialize() {
    this.expression.setRequiresExpression();
  }
  patchAsExpression(options = {}) {
    this.expression.patch({ ...options, needsParens: true });
  }
  statementNeedsParens() {
    return this.expression.statementShouldAddParens();
  }
};

// src/stages/main/patchers/UnaryMathOpPatcher.ts
var UnaryMathOpPatcher = class extends UnaryOpPatcher {
  isRepeatable() {
    return this.expression.isRepeatable();
  }
};

// src/stages/main/patchers/ForInPatcher.ts
var UP = "UP";
var DOWN = "DOWN";
var UNKNOWN = "UNKNOWN";
var ForInPatcher = class extends ForPatcher {
  constructor(patcherContext, keyAssignee, valAssignee, target, step, filter, body) {
    super(patcherContext, keyAssignee, valAssignee, target, filter, body);
    this._ascReference = null;
    this._endCode = null;
    this._endReference = null;
    this._internalIndexBinding = null;
    this._startCode = null;
    this._startReference = null;
    this._valueBinding = null;
    this._step = null;
    this.step = step;
  }
  initialize() {
    super.initialize();
    if (this.step) {
      this.step.setRequiresExpression();
    }
  }
  patchAsExpression() {
    if (!this.canPatchAsMapExpression()) {
      return super.patchAsExpression();
    }
    if (!this.body) {
      throw this.error("Expected non-null body.");
    }
    this.removeThenToken();
    let assigneeCode = this.getValueBinding();
    if (this.keyAssignee) {
      assigneeCode += `, ${this.getIndexBinding()}`;
    }
    this.remove(this.contentStart, this.target.outerStart);
    if (this.shouldWrapMapExpressionTargetInArrayFrom()) {
      this.insert(this.target.contentStart, "Array.from(");
      this.target.patch();
      this.insert(this.target.contentEnd, ")");
    } else {
      this.target.patch({ needsParens: true });
    }
    let mapInsertPoint;
    if (this.filter !== null) {
      this.overwrite(this.target.outerEnd, this.filter.outerStart, `.filter((${assigneeCode}) => `);
      this.filter.patch();
      this.insert(this.filter.outerEnd, `)`);
      mapInsertPoint = this.filter.outerEnd;
    } else {
      mapInsertPoint = this.target.outerEnd;
    }
    if (this.isMapBodyNoOp()) {
      this.remove(mapInsertPoint, this.body.outerEnd);
    } else {
      this.insert(mapInsertPoint, `.map((${assigneeCode}) =>`);
      this.patchBodyForExpressionLoop();
      this.insert(this.body.outerEnd, ")");
    }
  }
  isMapBodyNoOp() {
    if (this.valAssignee instanceof IdentifierPatcher) {
      const varName = this.valAssignee.node.data;
      if (this.body instanceof BlockPatcher && this.body.statements.length === 1) {
        const statement = this.body.statements[0];
        if (statement instanceof IdentifierPatcher && statement.node.data === varName) {
          return true;
        }
      }
    }
    return false;
  }
  patchBodyForExpressionLoop() {
    if (!this.body) {
      throw this.error("Expected non-null body.");
    }
    this.body.setRequiresExpression();
    const bodyNeedsParens = blockStartsWithObjectInitialiser(this.body) && !this.body.isSurroundedByParentheses();
    if (bodyNeedsParens) {
      const insertPoint = this.filter ? this.filter.outerEnd : this.target.outerEnd;
      if (this.slice(insertPoint, insertPoint + 1) === " ") {
        this.body.insert(insertPoint + 1, "(");
      } else {
        this.body.insert(insertPoint, " (");
      }
    }
    this.body.patch();
    if (bodyNeedsParens) {
      this.body.insert(this.body.outerEnd, ")");
    }
  }
  canPatchAsMapExpression() {
    if (!this.canAssigneesBecomeParams()) {
      return false;
    }
    if (this.step !== null) {
      return false;
    }
    if (this.body === null || !this.body.prefersToPatchAsExpression()) {
      return false;
    }
    if (this.filter !== null && this.keyAssignee !== null) {
      return false;
    }
    if (this.filter !== null && !this.filter.isPure()) {
      return false;
    }
    if (this.body.containsYield() || this.filter && this.filter.containsYield()) {
      return false;
    }
    if (this.body.containsAwait() || this.filter && this.filter.containsAwait()) {
      return false;
    }
    return true;
  }
  canAssigneesBecomeParams() {
    const assignees = [this.valAssignee, this.keyAssignee].filter((assignee) => assignee);
    for (const assignee of assignees) {
      if (!(assignee instanceof IdentifierPatcher)) {
        return false;
      }
      const name = assignee.node.data;
      const assignmentNode = this.getScope().getBinding(name);
      if (!assignmentNode) {
        throw this.error("Expected loop assignee to have a binding in its scope.");
      }
      const containerNode = this.context.getScope(assignmentNode).containerNode;
      if (countVariableUsages(containerNode, name) !== countVariableUsages(this.node, name)) {
        return false;
      }
    }
    return true;
  }
  willPatchAsIIFE() {
    return this.willPatchAsExpression() && !this.canPatchAsMapExpression();
  }
  patchAsStatement() {
    if (this.body && !this.body.inline()) {
      this.body.setIndent(this.getLoopBodyIndent());
    }
    if (this.shouldPatchAsForOf()) {
      this.getFilterCode();
      this.patchForOfLoop();
    } else {
      this.getIndexBinding();
      this.getValueBinding();
      this.getFilterCode();
      this.patchForLoopHeader();
      this.patchForLoopBody();
    }
  }
  shouldPatchAsForOf() {
    return !this.shouldPatchAsInitTestUpdateLoop() && this.step === null && this.keyAssignee === null;
  }
  getValueBinding() {
    if (!this._valueBinding) {
      if (this.valAssignee) {
        this._valueBinding = this.valAssignee.patchAndGetCode();
      } else if (this.shouldPatchAsInitTestUpdateLoop()) {
        this._valueBinding = this.claimFreeBinding(this.indexBindingCandidates());
      } else {
        this._valueBinding = this.claimFreeBinding("value");
      }
    }
    return this._valueBinding;
  }
  computeIndexBinding() {
    if (this.shouldPatchAsInitTestUpdateLoop()) {
      return this.getValueBinding();
    } else {
      return super.computeIndexBinding();
    }
  }
  patchForLoopHeader() {
    if (this.requiresExtractingTarget()) {
      this.insert(this.innerStart, `${this.getTargetReference()} = ${this.getTargetCode()}
${this.getLoopIndent()}`);
    }
    const firstHeaderPatcher = this.valAssignee || this.target;
    const lastHeaderPatcher = this.getLastHeaderPatcher();
    this.overwrite(firstHeaderPatcher.outerStart, lastHeaderPatcher.outerEnd, `(${this.getInitCode()}; ${this.getTestCode()}; ${this.getUpdateCode()}) {`);
  }
  getLastHeaderPatcher() {
    let resultPatcher = null;
    for (const patcher of [this.step, this.filter, this.target]) {
      if (patcher && (resultPatcher === null || patcher.contentEnd > resultPatcher.contentEnd)) {
        resultPatcher = patcher;
      }
    }
    return notNull(resultPatcher);
  }
  patchForLoopBody() {
    this.removeThenToken();
    this.patchPossibleNewlineAfterLoopHeader(this.getLastHeaderPatcher().outerEnd);
    if (this.body && !this.shouldPatchAsInitTestUpdateLoop() && this.valAssignee) {
      let valueAssignment = `${this.getValueBinding()} = ${this.getTargetReference()}[${this.getIndexBinding()}]`;
      if (this.valAssignee.statementNeedsParens()) {
        valueAssignment = `(${valueAssignment})`;
      }
      this.body.insertLineBefore(valueAssignment, this.getOuterLoopBodyIndent());
    }
    this.patchBodyAndFilter();
  }
  patchForOfLoop() {
    this.getFilterCode();
    if (this.filter) {
      this.remove(this.target.outerEnd, this.filter.outerEnd);
    }
    if (this.valAssignee) {
      const relationToken = this.getRelationToken();
      this.valAssignee.patch();
      this.insert(this.valAssignee.outerStart, "(");
      this.overwrite(relationToken.start, relationToken.end, "of");
    } else {
      const valueBinding = this.getValueBinding();
      this.insert(this.target.outerStart, `(let ${valueBinding} of `);
    }
    if (this.shouldWrapForOfStatementTargetInArrayFrom()) {
      this.insert(this.target.outerStart, "Array.from(");
    }
    this.target.patch();
    if (this.shouldWrapForOfStatementTargetInArrayFrom()) {
      this.insert(this.target.outerEnd, ")");
    }
    this.insert(this.target.outerEnd, ") {");
    this.removeThenToken();
    this.patchBodyAndFilter();
  }
  getLoopHeaderEnd() {
    return Math.max(this.step ? this.step.outerEnd : -1, super.getLoopHeaderEnd());
  }
  requiresExtractingTarget() {
    return !this.shouldPatchAsInitTestUpdateLoop() && !this.target.isRepeatable() && !this.shouldPatchAsForOf();
  }
  targetBindingCandidate() {
    return "iterable";
  }
  getInternalIndexBinding() {
    if (!this._internalIndexBinding) {
      if (this.needsUniqueIndexName()) {
        this._internalIndexBinding = this.claimFreeBinding(this.indexBindingCandidates());
      } else {
        this._internalIndexBinding = this.getIndexBinding();
      }
    }
    return this._internalIndexBinding;
  }
  needsUniqueIndexName() {
    if (this.isThisAssignIndexBinding()) {
      return true;
    }
    const userIndex = this.getIndexBinding();
    if (this.getScope().hasInnerClosureModification(userIndex)) {
      return true;
    }
    const fakeScope = new Scope(this.node, null);
    traverse5(this.node, (child) => {
      fakeScope.processNode(child);
    });
    return fakeScope.hasModificationAfterDeclaration(userIndex);
  }
  getInitCode() {
    const step = this.getStep();
    if (this.shouldPatchAsInitTestUpdateLoop()) {
      const assignments = [];
      if (this.shouldExtractStart()) {
        assignments.push(`${this.getStartReference()} = ${this.getStartCode()}`);
      }
      assignments.push(`${this.getInternalIndexBinding()} = ${this.getStartReference()}`);
      if (this.getInternalIndexBinding() !== this.getIndexBinding()) {
        assignments.push(`${this.getIndexBinding()} = ${this.getInternalIndexBinding()}`);
      }
      if (!this.isEndFixed()) {
        assignments.push(`${this.getEndReference()} = ${this.getEndCode()}`);
      }
      if (!step.isLiteral) {
        assignments.push(`${step.update} = ${step.init}`);
      }
      if (this.getIndexDirection() === UNKNOWN) {
        assignments.push(`${this.getAscReference()} = ${this.getAscCode()}`);
      }
      return assignments.join(", ");
    } else {
      const direction = this.getIndexDirection();
      const descInit = `${this.getTargetReference()}.length - 1`;
      const assignments = [];
      if (!step.isLiteral) {
        assignments.push(`${step.update} = ${step.init}`);
      }
      if (direction === DOWN) {
        assignments.push(`${this.getInternalIndexBinding()} = ${descInit}`);
      } else if (direction === UP) {
        assignments.push(`${this.getInternalIndexBinding()} = 0`);
      } else {
        assignments.push(`${this.getAscReference()} = ${this.getAscCode()}`);
        assignments.push(`${this.getInternalIndexBinding()} = ${this.getAscReference()} ? 0 : ${descInit}`);
      }
      if (this.getInternalIndexBinding() !== this.getIndexBinding()) {
        assignments.push(`${this.getIndexBinding()} = ${this.getInternalIndexBinding()}`);
      }
      return assignments.join(", ");
    }
  }
  getTestCode() {
    const direction = this.getIndexDirection();
    if (this.shouldPatchAsInitTestUpdateLoop()) {
      if (!(this.target instanceof RangePatcher)) {
        throw this.error("Expected range patcher for target.");
      }
      const inclusive = this.target.isInclusive();
      const gt = inclusive ? ">=" : ">";
      const lt = inclusive ? "<=" : "<";
      const index = this.getInternalIndexBinding();
      const end = this.getEndReference();
      if (direction === DOWN) {
        return `${index} ${gt} ${end}`;
      } else if (direction === UP) {
        return `${index} ${lt} ${end}`;
      } else {
        return `${this.getAscReference()} ? ${index} ${lt} ${end} : ${index} ${gt} ${end}`;
      }
    } else {
      const downComparison = `${this.getInternalIndexBinding()} >= 0`;
      const upComparison = `${this.getInternalIndexBinding()} < ${this.getTargetReference()}.length`;
      if (direction === DOWN) {
        return downComparison;
      } else if (direction === UP) {
        return upComparison;
      } else {
        return `${this.getAscReference()} ? ${upComparison} : ${downComparison}`;
      }
    }
  }
  getUpdateCode() {
    const assignments = [this.getUpdateAssignment()];
    if (this.getInternalIndexBinding() !== this.getIndexBinding()) {
      assignments.push(`${this.getIndexBinding()} = ${this.getInternalIndexBinding()}`);
    }
    return assignments.join(", ");
  }
  getUpdateAssignment() {
    const index = this.getInternalIndexBinding();
    const step = this.getStep();
    if (step.number === null) {
      return `${index} += ${step.update}`;
    }
    const direction = this.getIndexDirection();
    const incCode = step.number === 1 ? "++" : ` += ${step.update}`;
    const decCode = step.number === 1 ? "--" : ` -= ${step.update}`;
    if (direction === DOWN) {
      return `${index}${decCode}`;
    } else if (direction === UP) {
      return `${index}${incCode}`;
    } else {
      return `${this.getAscReference()} ? ${index}${incCode} : ${index}${decCode}`;
    }
  }
  getStartReference() {
    if (!this.shouldExtractStart()) {
      return this.getStartCode();
    }
    if (!this._startReference) {
      this._startReference = this.claimFreeBinding("start");
    }
    return this._startReference;
  }
  isStartFixed() {
    if (!(this.target instanceof RangePatcher)) {
      throw this.error("Expected target to be a range.");
    }
    return this.target.left.node.type === "Int" || this.target.left.node.type === "Float";
  }
  shouldExtractStart() {
    if (!(this.target instanceof RangePatcher)) {
      throw this.error("Expected target to be a range.");
    }
    return !this.target.left.isRepeatable() && this.getIndexDirection() === UNKNOWN && this.getStep().isVirtual;
  }
  getStartCode() {
    if (!(this.target instanceof RangePatcher)) {
      throw this.error("Expected target to be a range.");
    }
    if (!this._startCode) {
      this._startCode = this.target.left.patchAndGetCode();
    }
    return this._startCode;
  }
  getEndReference() {
    if (this.isEndFixed()) {
      return this.getEndCode();
    }
    if (!this._endReference) {
      this._endReference = this.claimFreeBinding("end");
    }
    return this._endReference;
  }
  isEndFixed() {
    if (!(this.target instanceof RangePatcher)) {
      throw this.error("Expected target to be a range.");
    }
    return this.target.right.node.type === "Int" || this.target.right.node.type === "Float";
  }
  getEndCode() {
    if (!(this.target instanceof RangePatcher)) {
      throw this.error("Expected target to be a range.");
    }
    if (!this._endCode) {
      this._endCode = this.target.right.patchAndGetCode();
    }
    return this._endCode;
  }
  getAscReference() {
    if (!this._ascReference) {
      this._ascReference = this.claimFreeBinding("asc");
    }
    return this._ascReference;
  }
  getAscCode() {
    const step = this.getStep();
    if (step.isVirtual) {
      if (!this.shouldPatchAsInitTestUpdateLoop()) {
        throw new Error("Should not be getting asc code when the target is not a range and the step is unspecified.");
      }
      return `${this.getStartReference()} <= ${this.getEndReference()}`;
    } else {
      return `${step.update} > 0`;
    }
  }
  getStep() {
    if (this._step === null) {
      this._step = new Step(this.step);
    }
    return this._step;
  }
  shouldPatchAsInitTestUpdateLoop() {
    return this.target instanceof RangePatcher;
  }
  shouldWrapMapExpressionTargetInArrayFrom() {
    const shouldWrap = !this.options.looseForExpressions && !this.isTargetAlreadyArray();
    if (shouldWrap) {
      this.addSuggestion(REMOVE_ARRAY_FROM);
    }
    return shouldWrap;
  }
  shouldWrapForOfStatementTargetInArrayFrom() {
    const shouldWrap = !this.options.looseForOf && !this.isTargetAlreadyArray();
    if (shouldWrap) {
      this.addSuggestion(REMOVE_ARRAY_FROM);
    }
    return shouldWrap;
  }
  isTargetAlreadyArray() {
    return this.target instanceof RangePatcher || this.target instanceof ArrayInitialiserPatcher;
  }
  hasExplicitStep() {
    return !this.getStep().isVirtual;
  }
  getIndexDirection() {
    const step = this.getStep();
    if (this.shouldPatchAsInitTestUpdateLoop()) {
      if (!(this.target instanceof RangePatcher)) {
        throw this.error("Expected target to be a range.");
      }
      if (!step.isVirtual && step.isLiteral) {
        return step.negated ? DOWN : UP;
      } else if (this.hasFixedRange()) {
        if (!(this.target.left.node instanceof Number) || !(this.target.right.node instanceof Number)) {
          throw this.error("Expected numbers for the left and right of the range.");
        }
        const left = this.target.left.node.data;
        const right = this.target.right.node.data;
        return left > right ? DOWN : UP;
      } else {
        this.addSuggestion(SIMPLIFY_DYNAMIC_RANGE_LOOPS);
        return UNKNOWN;
      }
    } else {
      if (step.isLiteral) {
        return step.negated ? DOWN : UP;
      } else {
        this.addSuggestion(SIMPLIFY_DYNAMIC_RANGE_LOOPS);
        return UNKNOWN;
      }
    }
  }
  hasFixedRange() {
    return this.target instanceof RangePatcher && this.isStartFixed() && this.isEndFixed();
  }
};
var Step = class {
  constructor(patcher) {
    let negated = false;
    let root = patcher;
    const apply = (patcher2) => {
      if (patcher2.node instanceof UnaryNegateOp && patcher2 instanceof UnaryMathOpPatcher) {
        negated = !negated;
        apply(patcher2.expression);
      } else {
        root = patcher2;
      }
    };
    if (patcher) {
      apply(patcher);
      if (!root) {
        throw new Error("Expected non-null root.");
      }
      this.isLiteral = root.node instanceof Int3 || root.node instanceof Float;
      this.init = patcher.patchAndGetCode();
      if (root.node instanceof Int3 || root.node instanceof Float) {
        this.isLiteral = true;
        this.update = root.slice(root.contentStart, root.contentEnd);
        this.number = root.node.data;
      } else {
        this.isLiteral = false;
        this.update = root.claimFreeBinding("step");
        this.number = null;
      }
    } else {
      this.isLiteral = true;
      this.init = "1";
      this.update = "1";
      this.number = 1;
    }
    this.negated = negated;
    this.isVirtual = !patcher;
  }
};

// src/stages/main/patchers/ForFromPatcher.ts
var ForFromPatcher = class extends ForInPatcher {
  constructor(patcherContext, keyAssignee, valAssignee, target, filter, body) {
    super(patcherContext, keyAssignee, valAssignee, target, null, filter, body);
  }
  shouldPatchAsForOf() {
    return true;
  }
  shouldWrapForOfStatementTargetInArrayFrom() {
    return false;
  }
};

// src/stages/main/patchers/ForOfPatcher.ts
import { SourceType as SourceType30 } from "coffee-lex";
var ForOfPatcher = class extends ForPatcher {
  patchAsStatement() {
    if (this.body && !this.body.inline()) {
      this.body.setIndent(this.getLoopBodyIndent());
    }
    const { keyAssignee } = this;
    this.getFilterCode();
    if (this.filter) {
      this.remove(this.target.outerEnd, this.filter.outerEnd);
    }
    this.removeOwnTokenIfExists();
    const shouldExtractTarget = this.requiresExtractingTarget();
    if (shouldExtractTarget) {
      this.insert(this.innerStart, `${this.getTargetReference()} = ${this.getTargetCode()}
${this.getLoopIndent()}`);
    }
    const keyBinding = this.getIndexBinding();
    this.insert(keyAssignee.outerStart, "(");
    this.overwrite(this.keyAssignee.contentStart, this.keyAssignee.contentEnd, keyBinding);
    const targetReference = this.getTargetReference();
    const { valAssignee } = this;
    let valueAssignment = null;
    if (valAssignee) {
      valAssignee.patch();
      const valAssigneeString = this.slice(valAssignee.contentStart, valAssignee.contentEnd);
      this.remove(keyAssignee.outerEnd, valAssignee.outerEnd);
      valueAssignment = `${valAssigneeString} = ${this.getTargetReference()}[${keyBinding}]`;
      if (valAssignee.statementNeedsParens()) {
        valueAssignment = `(${valueAssignment})`;
      }
    }
    const relationToken = this.getRelationToken();
    if (this.node.isOwn) {
      this.addSuggestion(CLEAN_UP_FOR_OWN_LOOPS);
      if (shouldExtractTarget) {
        this.overwrite(relationToken.end, this.target.outerEnd, ` Object.keys(${targetReference} || {})) {`);
      } else {
        this.insert(this.target.outerStart, "Object.keys(");
        this.insert(this.target.outerEnd, " || {})) {");
      }
    } else {
      if (shouldExtractTarget) {
        this.overwrite(relationToken.start, this.target.outerEnd, `in ${targetReference}) {`);
      } else {
        this.overwrite(relationToken.start, relationToken.end, "in");
        this.insert(this.target.outerEnd, ") {");
      }
    }
    this.removeThenToken();
    this.patchPossibleNewlineAfterLoopHeader(this.target.outerEnd);
    if (valueAssignment !== null && this.body !== null) {
      this.body.insertLineBefore(valueAssignment, this.getOuterLoopBodyIndent());
    }
    this.patchBodyAndFilter();
  }
  removeOwnTokenIfExists() {
    if (this.node.isOwn) {
      const ownIndex = this.indexOfSourceTokenAfterSourceTokenIndex(this.contentStartTokenIndex, SourceType30.OWN);
      if (!ownIndex) {
        throw this.error("Expected to find own token in for-own.");
      }
      const ownToken = notNull(this.sourceTokenAtIndex(ownIndex));
      this.remove(ownToken.start, this.keyAssignee.outerStart);
    }
  }
  requiresExtractingTarget() {
    return !this.target.isRepeatable() && this.valAssignee !== null;
  }
  targetBindingCandidate() {
    return "object";
  }
  indexBindingCandidates() {
    return ["key"];
  }
  willPatchAsIIFE() {
    return this.willPatchAsExpression();
  }
};

// src/stages/main/patchers/HeregexPatcher.ts
var CLOSE_TOKEN_BASE_LENGTH = 3;
var HeregexPatcher = class extends InterpolatedPatcher {
  patchAsExpression() {
    const openToken = this.firstToken();
    const closeToken = this.lastToken();
    this.overwrite(openToken.start, openToken.end, "new RegExp(`");
    if (closeToken.end - closeToken.start > CLOSE_TOKEN_BASE_LENGTH) {
      this.overwrite(closeToken.start, closeToken.start + CLOSE_TOKEN_BASE_LENGTH, "`, '");
      this.insert(closeToken.end, "')");
    } else {
      this.overwrite(closeToken.start, closeToken.end, "`)");
    }
    this.patchInterpolations();
    this.processContents();
    this.escapeQuasis(/^\\\s/, ["`", "${", "\\"]);
  }
  shouldExcapeZeroChars() {
    return true;
  }
  shouldDowngradeUnicodeCodePointEscapes() {
    return !this.node.flags.unicode;
  }
};

// src/stages/main/patchers/IncrementDecrementPatcher.ts
var IncrementDecrementPatcher = class extends NodePatcher {
  constructor(patcherContext, expression) {
    super(patcherContext);
    this.expression = expression;
  }
  initialize() {
    this.expression.setRequiresExpression();
  }
  patchAsExpression() {
    this.expression.patch();
  }
  isRepeatable() {
    return false;
  }
};

// src/stages/main/patchers/InOpPatcher.ts
import { SourceType as SourceType31 } from "coffee-lex";
var IN_HELPER = `function __in__(needle, haystack) {
  return Array.from(haystack).indexOf(needle) >= 0;
}`;
var InOpPatcher = class extends BinaryOpPatcher {
  constructor(patcherContext, left, right) {
    super(patcherContext, left, right);
    this.negated = patcherContext.node.isNot;
  }
  negate() {
    this.negated = !this.negated;
  }
  operatorTokenPredicate() {
    return (token) => token.type === SourceType31.RELATION;
  }
  patchAsExpression() {
    if (this.options.noArrayIncludes) {
      this.patchAsIndexLookup();
      return;
    }
    if (!this.left.isPure() || !this.right.isPure()) {
      this.patchWithLHSExtracted();
      return;
    }
    let rightCode = this.right.patchAndGetCode();
    if (this.shouldWrapInArrayFrom()) {
      rightCode = `Array.from(${rightCode})`;
    } else if (this.rhsNeedsParens()) {
      rightCode = `(${rightCode})`;
    }
    this.remove(this.left.outerEnd, this.right.outerEnd);
    if (this.negated) {
      this.insert(this.left.outerStart, "!");
    }
    this.insert(this.left.outerStart, `${rightCode}.includes(`);
    this.left.patch();
    this.insert(this.left.outerEnd, ")");
  }
  patchWithLHSExtracted() {
    this.addSuggestion(FIX_INCLUDES_EVALUATION_ORDER);
    this.insert(this.contentStart, "(");
    const leftRef = this.left.patchRepeatable({ ref: "needle", forceRepeat: true });
    this.insert(this.left.outerEnd, `, `);
    this.remove(this.left.outerEnd, this.right.outerStart);
    if (this.negated) {
      this.insert(this.right.outerStart, "!");
    }
    const wrapInArrayFrom = this.shouldWrapInArrayFrom();
    const rhsNeedsParens = wrapInArrayFrom || this.rhsNeedsParens();
    if (wrapInArrayFrom) {
      this.insert(this.right.outerStart, "Array.from");
    }
    if (rhsNeedsParens) {
      this.insert(this.right.outerStart, "(");
    }
    this.right.patch();
    if (rhsNeedsParens) {
      this.insert(this.right.outerEnd, ")");
    }
    this.insert(this.right.outerEnd, `.includes(${leftRef}))`);
  }
  shouldWrapInArrayFrom() {
    if (this.options.looseIncludes) {
      return false;
    }
    const shouldWrap = !(this.right instanceof ArrayInitialiserPatcher);
    if (shouldWrap) {
      this.addSuggestion(REMOVE_ARRAY_FROM);
    }
    return shouldWrap;
  }
  rhsNeedsParens() {
    return !(this.right instanceof IdentifierPatcher) && !(this.right instanceof MemberAccessOpPatcher) && !(this.right instanceof DynamicMemberAccessOpPatcher) && !(this.right instanceof FunctionApplicationPatcher) && !(this.right instanceof ArrayInitialiserPatcher) && !(this.right instanceof StringPatcher);
  }
  patchAsIndexLookup() {
    const helper = this.registerHelper("__in__", IN_HELPER);
    if (this.negated) {
      this.insert(this.left.outerStart, "!");
    }
    this.insert(this.left.outerStart, `${helper}(`);
    this.left.patch();
    this.overwrite(this.left.outerEnd, this.right.outerStart, ", ");
    this.right.patch();
    this.insert(this.right.outerEnd, ")");
  }
  statementNeedsParens() {
    return false;
  }
};

// src/stages/main/patchers/NegatableBinaryOpPatcher.ts
var NegatableBinaryOpPatcher = class extends BinaryOpPatcher {
  constructor(patcherContext, left, right) {
    super(patcherContext, left, right);
    this.negated = patcherContext.node.isNot;
  }
  negate() {
    this.negated = !this.negated;
  }
  javaScriptOperator() {
    throw new Error(`'javaScriptOperator' should be implemented in subclass`);
  }
  patchAsExpression() {
    const { negated } = this;
    if (negated) {
      this.insert(this.innerStart, "!(");
    }
    super.patchAsExpression();
    if (negated) {
      this.insert(this.innerEnd, ")");
    }
    const token = this.getOperatorToken();
    this.overwrite(token.start, token.end, this.javaScriptOperator());
  }
  statementNeedsParens() {
    if (this.negated) {
      return false;
    } else {
      return super.statementNeedsParens();
    }
  }
};

// src/stages/main/patchers/InstanceofOpPatcher.ts
var InstanceofOpPatcher = class extends NegatableBinaryOpPatcher {
  javaScriptOperator() {
    return "instanceof";
  }
};

// src/stages/main/patchers/JavaScriptPatcher.ts
import { SourceType as SourceType32 } from "coffee-lex";
var JavaScriptPatcher = class extends NodePatcher {
  patchAsExpression() {
    const shouldRemoveSurroundingParens = !(this.parent instanceof FunctionApplicationPatcher && this === this.parent.args[0]);
    if (shouldRemoveSurroundingParens) {
      this.remove(this.outerStart, this.contentStart);
    }
    const token = this.firstToken();
    let tokenLength;
    if (token.type === SourceType32.JS) {
      tokenLength = 1;
    } else if (token.type === SourceType32.HEREJS) {
      tokenLength = 3;
    } else {
      throw this.error(`Unexpected token in JavaScript node: ${token.type}`);
    }
    this.remove(this.contentStart, this.contentStart + tokenLength);
    for (let i = this.contentStart + tokenLength; i < this.contentEnd - tokenLength - 1; i++) {
      if (this.slice(i, i + 2) === "\\`") {
        let startBackslash = i;
        while (this.context.source[startBackslash] === "\\") {
          startBackslash--;
        }
        startBackslash++;
        const numBackslashesToKeep = Math.floor((i + 1 - startBackslash) / 2);
        this.remove(startBackslash, i + 1 - numBackslashesToKeep);
      }
    }
    this.remove(this.contentEnd - tokenLength, this.contentEnd);
    if (shouldRemoveSurroundingParens) {
      this.remove(this.contentEnd, this.outerEnd);
    }
  }
};

// src/stages/main/patchers/LogicalNotOpPatcher.ts
var LogicalNotOpPatcher = class extends UnaryOpPatcher {
  isRepeatable() {
    return this.expression.isRepeatable();
  }
  patchAsExpression(options = {}) {
    if (this.expression.canHandleNegationInternally()) {
      this.expression.negate();
      this.remove(this.contentStart, this.expression.outerStart);
    } else {
      this.overwrite(this.contentStart, this.expression.outerStart, "!");
    }
    super.patchAsExpression(options);
  }
};

// src/stages/main/patchers/LogicalOpCompoundAssignOpPatcher.ts
var LogicalOpCompoundAssignOpPatcher = class extends CompoundAssignOpPatcher {
  patchAsExpression({ needsParens = false } = {}) {
    const shouldAddParens = this.negated || needsParens && !this.isSurroundedByParentheses();
    if (this.negated) {
      this.insert(this.contentStart, "!");
    }
    if (shouldAddParens) {
      this.insert(this.contentStart, "(");
    }
    const operator = this.getOperatorToken();
    this.overwrite(operator.start, operator.end, this.isOrOp() ? `||` : `&&`);
    const assigneeAgain = this.assignee.patchRepeatable({ isForAssignment: true });
    this.insert(this.expression.outerStart, `(${assigneeAgain} = `);
    this.expression.patch();
    this.insert(this.expression.outerEnd, ")");
    if (shouldAddParens) {
      this.insert(this.contentEnd, ")");
    }
  }
  patchAsStatement(options = {}) {
    if (this.lhsHasSoakOperation()) {
      this.patchAsExpression(options);
      return;
    }
    this.insert(this.contentStart, "if (");
    if (this.isOrOp()) {
      this.assignee.negate();
    }
    const assigneeAgain = this.assignee.patchRepeatable({ isForAssignment: true });
    this.overwrite(this.assignee.outerEnd, this.expression.outerStart, `) { ${assigneeAgain} = `);
    this.expression.patch();
    this.insert(this.expression.outerEnd, " }");
  }
  isOrOp() {
    const operator = this.getOperatorToken();
    const op = this.sourceOfToken(operator);
    return op.substr(0, 2) === "||" || op.substr(0, 2) === "or";
  }
  statementNeedsParens() {
    return false;
  }
};

// src/stages/main/patchers/LogicalOpPatcher.ts
var LogicalOpPatcher = class extends BinaryOpPatcher {
  constructor() {
    super(...arguments);
    this.negated = false;
  }
  patchOperator() {
    const operatorToken = this.getOperatorToken();
    this.overwrite(operatorToken.start, operatorToken.end, this.getOperator());
  }
  getOperator() {
    const operatorToken = this.getOperatorToken();
    let operator = this.context.source.slice(operatorToken.start, operatorToken.end);
    if (operator === "and") {
      operator = "&&";
    } else if (operator === "or") {
      operator = "||";
    }
    if (this.negated) {
      return operator === "&&" ? "||" : "&&";
    } else {
      return operator;
    }
  }
  negate() {
    this.negated = !this.negated;
    this.left.negate();
    this.right.negate();
  }
};

// src/stages/main/patchers/ModuleSpecifierPatcher.ts
var ModuleSpecifierPatcher = class extends NodePatcher {
  constructor(patcherContext, original, alias) {
    super(patcherContext);
    this.original = original;
    this.alias = alias;
  }
  patchAsStatement() {
    this.original.patch();
    if (this.alias) {
      this.alias.patch();
    }
  }
};

// src/utils/registerModHelper.ts
var MOD_HELPER = `function __mod__(a, b) {
  a = +a;
  b = +b;
  return (a % b + b) % b;
}`;
function registerModHelper(patcher) {
  return patcher.registerHelper("__mod__", MOD_HELPER);
}

// src/stages/main/patchers/ModuloOpCompoundAssignOpPatcher.ts
var ModuloOpCompoundAssignOpPatcher = class extends CompoundAssignOpPatcher {
  patchAsExpression({ needsParens = false } = {}) {
    const helper = registerModHelper(this);
    const shouldAddParens = this.negated || needsParens && !this.isSurroundedByParentheses();
    if (this.negated) {
      this.insert(this.contentStart, "!");
    }
    if (shouldAddParens) {
      this.insert(this.contentStart, "(");
    }
    const assigneeAgain = this.assignee.patchRepeatable({ isForAssignment: true });
    this.overwrite(this.assignee.outerEnd, this.expression.outerStart, ` = ${helper}(${assigneeAgain}, `);
    this.expression.patch();
    this.insert(this.expression.outerEnd, ")");
    if (shouldAddParens) {
      this.insert(this.contentEnd, ")");
    }
  }
  patchAsStatement() {
    this.patchAsExpression();
  }
};

// src/stages/main/patchers/ModuloOpPatcher.ts
var ModuloOpPatcher = class extends BinaryOpPatcher {
  constructor(patcherContext, left, right) {
    super(patcherContext, left, right);
  }
  patchAsExpression() {
    const helper = registerModHelper(this);
    this.insert(this.left.outerStart, `${helper}(`);
    this.left.patch();
    this.overwrite(this.left.outerEnd, this.right.outerStart, ", ");
    this.right.patch();
    this.insert(this.right.outerEnd, ")");
  }
  statementNeedsParens() {
    return false;
  }
};

// src/stages/main/patchers/NewOpPatcher.ts
var NewOpPatcher = class extends FunctionApplicationPatcher {
  patchAsExpression() {
    const fnNeedsParens = !this.fn.isSurroundedByParentheses() && !(this.fn instanceof IdentifierPatcher) && !(this.fn instanceof MemberAccessOpPatcher);
    super.patchAsExpression({ fnNeedsParens });
  }
};

// src/stages/main/patchers/NullishCoalescingCompoundAssignOpPatcher.ts
var NullishCoalescingCompoundAssignOpPatcher = class extends CompoundAssignOpPatcher {
  patchAsExpression() {
    this.patchOperator();
    super.patchAsExpression();
  }
  patchAsStatement() {
    this.patchOperator();
    super.patchAsStatement();
  }
  patchOperator() {
    const operator = this.getOperatorToken();
    this.overwrite(operator.start, operator.end, "??=");
  }
};

// src/stages/main/patchers/NullishCoalescingExistsOpPatcher.ts
var NullishCoalescingExistsOpPatcher = class extends ExistsOpPatcher {
  patchAsExpression(options = {}) {
    const needsTypeofCheck = this.left.mayBeUnboundReference();
    if (needsTypeofCheck) {
      super.patchAsExpression(options);
    } else {
      BinaryOpPatcher.prototype.patchAsExpression.call(this, options);
    }
  }
  patchOperator() {
    const opToken = this.getOperatorToken();
    this.overwrite(opToken.start, opToken.end, "??");
  }
};

// src/stages/main/patchers/OfOpPatcher.ts
import { SourceType as SourceType33 } from "coffee-lex";
var OfOpPatcher = class extends NegatableBinaryOpPatcher {
  operatorTokenPredicate() {
    return (token) => token.type === SourceType33.RELATION;
  }
  javaScriptOperator() {
    return "in";
  }
};

// src/stages/main/patchers/OptionalChainingSoakedDynamicMemberAccessOpPatcher.ts
import { SourceType as SourceType34 } from "coffee-lex";

// src/utils/isInsideAssignee.ts
import {
  AssignOp as AssignOp5,
  CompoundAssignOp as CompoundAssignOp2,
  DynamicMemberAccessOp as DynamicMemberAccessOp2,
  FunctionApplication as FunctionApplication2,
  MemberAccessOp as MemberAccessOp4,
  PostDecrementOp as PostDecrementOp2,
  PostIncrementOp as PostIncrementOp2,
  PreDecrementOp as PreDecrementOp2,
  PreIncrementOp as PreIncrementOp2,
  SoakedDynamicMemberAccessOp as SoakedDynamicMemberAccessOp3,
  SoakedMemberAccessOp as SoakedMemberAccessOp3,
  SoakedNewOp,
  SoakedFunctionApplication as SoakedFunctionApplication3
} from "decaffeinate-parser";
function isInsideAssignee(node) {
  if (!node.parentNode) {
    return false;
  }
  if (node.parentNode instanceof AssignOp5 || node.parentNode instanceof CompoundAssignOp2) {
    return node.parentNode.assignee === node;
  }
  if (node.parentNode instanceof PostIncrementOp2 || node.parentNode instanceof PreIncrementOp2 || node.parentNode instanceof PostDecrementOp2 || node.parentNode instanceof PreDecrementOp2) {
    return true;
  }
  if (node.parentNode instanceof MemberAccessOp4 || node.parentNode instanceof DynamicMemberAccessOp2 || node.parentNode instanceof SoakedMemberAccessOp3 || node.parentNode instanceof SoakedDynamicMemberAccessOp3) {
    return node.parentNode.expression === node && isInsideAssignee(node.parentNode);
  }
  if (node.parentNode instanceof SoakedNewOp) {
    return node.parentNode.ctor === node && isInsideAssignee(node.parentNode);
  }
  if (node.parentNode instanceof FunctionApplication2 || node.parentNode instanceof SoakedFunctionApplication3) {
    return node.parentNode.function === node && isInsideAssignee(node.parentNode);
  }
  return false;
}

// src/stages/main/patchers/OptionalChainingSoakedDynamicMemberAccessOpPatcher.ts
var OptionalChainingSoakedDynamicMemberAccessOpPatcher = class extends DynamicMemberAccessOpPatcher {
  patchAsExpression() {
    if (isInsideAssignee(this.node)) {
      throw this.error(`JavaScript does not allow an optional chaining expression in an assignment position. Run without --optional-chaining or edit the original source to remove the assignment of an optional chaining expression.`);
    }
    const bracketTokenIndex = this.indexOfSourceTokenBetweenPatchersMatching(this.expression, this.indexingExpr, (token) => token.type === SourceType34.LBRACKET);
    if (!bracketTokenIndex) {
      throw this.error(`Expected left bracket between expression and index expression, but none was found`);
    }
    const bracketToken = this.sourceTokenAtIndex(bracketTokenIndex);
    if (!bracketToken) {
      throw this.error(`Expected left bracket between expression and index expression, but none was found`);
    }
    this.expression.patch();
    this.insert(bracketToken.start, ".");
    this.indexingExpr.patch();
  }
};

// src/stages/main/patchers/OptionalChainingSoakedFunctionApplicationPatcher.ts
import { SourceType as SourceType35 } from "coffee-lex";
var OptionalChainingSoakedFunctionApplicationPatcher = class extends FunctionApplicationPatcher {
  patchAsExpression() {
    if (isInsideAssignee(this.node)) {
      throw this.error(`JavaScript does not allow an optional chaining expression in an assignment position. Run without --optional-chaining or edit the original source to remove the assignment of an optional chaining expression.`);
    }
    const existenceTokenIndex = notNull(this.indexOfSourceTokenAfterSourceTokenIndex(this.fn.outerEndTokenIndex, SourceType35.EXISTENCE));
    const lparenTokenIndex = notNull(this.indexOfSourceTokenAfterSourceTokenIndex(existenceTokenIndex, SourceType35.CALL_START));
    const lparenToken = notNull(this.sourceTokenAtIndex(lparenTokenIndex));
    this.insert(lparenToken.start, ".");
    super.patchAsExpression();
  }
};

// src/stages/main/patchers/OptionalChainingSoakedMemberAccessOpPatcher.ts
var OptionalChainingSoakedMemberAccessOpPatcher = class extends MemberAccessOpPatcher {
  patchAsExpression() {
    if (isInsideAssignee(this.node)) {
      throw this.error(`JavaScript does not allow an optional chaining expression in an assignment position. Run without --optional-chaining or edit the original source to remove the assignment of an optional chaining expression.`);
    }
    this.expression.patch();
  }
  hasImplicitOperator() {
    return false;
  }
};

// src/stages/main/patchers/OptionalChainingSoakedNewOpPatcher.ts
var OptionalChainingSoakedNewOpPatcher = class extends OptionalChainingSoakedFunctionApplicationPatcher {
  patchAsExpression() {
    throw this.error(`JavaScript does not allow constructors with optional chaining. Run without --optional-chaining or edit the original source to manually invoke the constructor conditionally.`);
  }
};

// src/stages/main/patchers/OptionalChainingSoakedSlicePatcher.ts
var OptionalChainingSoakedSlicePatcher = class extends SlicePatcher {
  getSliceStartCode(splice = false) {
    return splice ? "?.splice(" : "?.slice(";
  }
};

// src/stages/main/patchers/ProgramPatcher.ts
import { SourceType as SourceType36 } from "coffee-lex";

// src/patchers/SharedProgramPatcher.ts
var ProgramPatcher = class extends NodePatcher {
  constructor(patcherContext, body) {
    super(patcherContext);
    this.helpers = /* @__PURE__ */ new Map();
    this._indentString = null;
    this.body = body;
  }
  shouldTrimContentRange() {
    return true;
  }
  registerHelper(name, code) {
    code = code.trim();
    if (this.helpers.has(name)) {
      if (this.helpers.get(name) !== code) {
        throw new Error(`BUG: cannot override helper '${name}'`);
      }
    } else {
      this.helpers.set(name, code);
    }
    return name;
  }
  patchHelpers() {
    for (const helper of this.helpers.values()) {
      this.editor.append(`
${helper}`);
    }
  }
  getProgramIndentString() {
    if (!this._indentString) {
      this._indentString = determineIndent(this.context.source);
    }
    return this._indentString;
  }
};

// src/stages/main/patchers/ProgramPatcher.ts
var BLOCK_COMMENT_DELIMITER = "###";
var ProgramPatcher2 = class extends ProgramPatcher {
  canPatchAsExpression() {
    return false;
  }
  patchAsStatement() {
    this.patchComments();
    this.patchContinuations();
    if (this.body) {
      this.body.patch({ leftBrace: false, rightBrace: false });
    }
    this.patchHelpers();
  }
  patchContinuations() {
    this.getProgramSourceTokens().forEach((token) => {
      if (token.type === SourceType36.CONTINUATION) {
        this.remove(token.start, token.end);
      }
    });
  }
  patchComments() {
    const { source } = this.context;
    this.getProgramSourceTokens().forEach((token) => {
      if (token.type === SourceType36.COMMENT) {
        if (token.start === 0 && source[1] === "!") {
          this.patchShebangComment(token);
        } else {
          this.patchLineComment(token);
        }
      } else if (token.type === SourceType36.HERECOMMENT) {
        this.patchBlockComment(token);
      }
    });
  }
  patchBlockComment(comment) {
    const { start, end } = comment;
    this.overwrite(start, start + BLOCK_COMMENT_DELIMITER.length, "/*");
    let atStartOfLine = false;
    let lastStartOfLine = null;
    let lineUpAsterisks = true;
    let isMultiline = false;
    const { source } = this.context;
    const expectedIndent = getIndent(source, start);
    const leadingHashIndexes = [];
    for (let index = start + BLOCK_COMMENT_DELIMITER.length; index < end - BLOCK_COMMENT_DELIMITER.length; index++) {
      switch (source[index]) {
        case "\n":
          isMultiline = true;
          atStartOfLine = true;
          lastStartOfLine = index + "\n".length;
          break;
        case " ":
        case "	":
          break;
        case "#":
          if (atStartOfLine) {
            leadingHashIndexes.push(index);
            atStartOfLine = false;
            if (source.slice(notNull(lastStartOfLine), index) !== expectedIndent) {
              lineUpAsterisks = false;
            }
          }
          break;
        default:
          if (atStartOfLine) {
            atStartOfLine = false;
            lineUpAsterisks = false;
          }
          break;
      }
    }
    leadingHashIndexes.forEach((index) => {
      this.overwrite(index, index + "#".length, lineUpAsterisks ? " *" : "*");
    });
    this.overwrite(end - BLOCK_COMMENT_DELIMITER.length, end, isMultiline && lineUpAsterisks ? " */" : "*/");
  }
  patchLineComment(comment) {
    const { start } = comment;
    this.overwrite(start, start + "#".length, "//");
  }
  patchShebangComment(comment) {
    const { start, end } = comment;
    const commentBody = this.slice(start, end);
    const coffeeIndex = commentBody.indexOf("coffee");
    if (coffeeIndex >= 0) {
      this.overwrite(start + coffeeIndex, start + coffeeIndex + "coffee".length, "node");
    }
  }
  canHandleImplicitReturn() {
    return true;
  }
};

// src/stages/main/patchers/QuasiPatcher.ts
var QuasiPatcher = class extends PassthroughPatcher {
};

// src/stages/main/patchers/RegexPatcher.ts
var RegexPatcher = class extends NodePatcher {
  patchAsExpression() {
    escapeSpecialWhitespaceInRange(this.contentStart + 1, this.contentEnd - 1, this);
    if (!this.node.flags.unicode) {
      downgradeUnicodeCodePointEscapesInRange(this.contentStart + 1, this.contentEnd - 1, this);
    }
  }
};

// src/stages/main/patchers/RestPatcher.ts
var RestPatcher = class extends SpreadPatcher {
};

// src/stages/main/patchers/SeqOpPatcher.ts
import { SourceType as SourceType37 } from "coffee-lex";
var SeqOpPatcher = class extends NodePatcher {
  constructor(patcherContext, left, right) {
    super(patcherContext);
    this.negated = false;
    this.left = left;
    this.right = right;
  }
  negate() {
    this.negated = !this.negated;
  }
  patchAsExpression() {
    this.left.setRequiresExpression();
    this.right.setRequiresExpression();
    if (this.negated) {
      this.insert(this.innerStart, "!(");
    }
    this.left.patch();
    const token = this.getOperatorToken();
    if (token.type === SourceType37.SEMICOLON) {
      this.overwrite(token.start, token.end, ",");
    } else if (token.type === SourceType37.NEWLINE) {
      this.insert(this.left.outerEnd, ",");
    }
    this.right.patch();
    if (this.negated) {
      this.insert(this.innerEnd, ")");
    }
  }
  patchAsStatement() {
    this.left.patch();
    this.right.patch();
  }
  getOperatorToken() {
    const operatorTokenIndex = this.indexOfSourceTokenBetweenPatchersMatching(this.left, this.right, (token) => token.type === SourceType37.SEMICOLON || token.type === SourceType37.NEWLINE);
    if (!operatorTokenIndex) {
      throw this.error("expected operator between binary operands");
    }
    return notNull(this.sourceTokenAtIndex(operatorTokenIndex));
  }
  statementNeedsParens() {
    return this.left.statementShouldAddParens();
  }
};

// src/stages/main/patchers/SoakedFunctionApplicationPatcher.ts
import { SourceType as SourceType39 } from "coffee-lex";

// src/stages/main/patchers/SoakedMemberAccessOpPatcher.ts
var GUARD_HELPER = `function __guard__(value, transform) {
  return (typeof value !== 'undefined' && value !== null) ? transform(value) : undefined;
}`;
var SoakedMemberAccessOpPatcher = class extends MemberAccessOpPatcher {
  constructor() {
    super(...arguments);
    this._shouldSkipSoakPatch = false;
  }
  patchAsExpression() {
    if (!this._shouldSkipSoakPatch) {
      if (this.shouldPatchAsConditional()) {
        this.patchAsConditional();
      } else {
        this.patchAsGuardCall();
      }
    } else {
      this.expression.patch();
    }
  }
  shouldPatchAsConditional() {
    return this.expression.isRepeatable() && !nodeContainsSoakOperation(this.expression.node);
  }
  patchAsConditional() {
    this.addSuggestion(SHORTEN_NULL_CHECKS);
    const soakContainer = findSoakContainer(this);
    const memberNameToken = this.getMemberNameSourceToken();
    const expressionCode = this.expression.patchRepeatable();
    let conditionCode;
    if (this.expression.mayBeUnboundReference()) {
      conditionCode = `typeof ${expressionCode} !== 'undefined' && ${expressionCode} !== null`;
    } else {
      conditionCode = `${expressionCode} != null`;
    }
    this.overwrite(this.expression.outerEnd, memberNameToken.start, ".");
    if (soakContainer.willPatchAsExpression()) {
      const containerNeedsParens = ternaryNeedsParens(soakContainer);
      if (containerNeedsParens) {
        soakContainer.insert(soakContainer.contentStart, "(");
      }
      soakContainer.insert(soakContainer.contentStart, `${conditionCode} ? `);
      soakContainer.appendDeferredSuffix(" : undefined");
      if (containerNeedsParens) {
        soakContainer.appendDeferredSuffix(")");
      }
    } else {
      soakContainer.insert(soakContainer.contentStart, `if (${conditionCode}) {
${soakContainer.getIndent(1)}`);
      soakContainer.appendDeferredSuffix(`
${soakContainer.getIndent()}}`);
    }
  }
  patchAsGuardCall() {
    this.registerHelper("__guard__", GUARD_HELPER);
    this.addSuggestion(REMOVE_GUARD);
    const soakContainer = findSoakContainer(this);
    const varName = soakContainer.claimFreeBinding("x");
    const prefix = this.slice(soakContainer.contentStart, this.contentStart);
    if (prefix.length > 0) {
      this.remove(soakContainer.contentStart, this.contentStart);
    }
    const memberNameToken = this.getMemberNameSourceToken();
    this.overwrite(this.expression.outerEnd, memberNameToken.start, `, ${varName} => ${prefix}${varName}.`);
    soakContainer.insert(soakContainer.contentStart, "__guard__(");
    soakContainer.appendDeferredSuffix(")");
    this.expression.patch();
  }
  setShouldSkipSoakPatch() {
    this._shouldSkipSoakPatch = true;
  }
  hasImplicitOperator() {
    return false;
  }
};

// src/stages/main/patchers/WhilePatcher.ts
import { SourceType as SourceType38 } from "coffee-lex";
var WhilePatcher = class extends LoopPatcher {
  constructor(patcherContext, condition, guard, body) {
    super(patcherContext, body);
    this.condition = condition;
    this.guard = guard;
  }
  initialize() {
    super.initialize();
    this.condition.setRequiresExpression();
    if (this.guard !== null) {
      this.guard.setRequiresExpression();
    }
  }
  patchAsStatement() {
    if (this.body && !this.body.inline()) {
      this.body.setIndent(this.getLoopBodyIndent());
    }
    const whileToken = notNull(this.sourceTokenAtIndex(this.getWhileTokenIndex()));
    this.overwrite(whileToken.start, whileToken.end, "while");
    const conditionNeedsParens = !this.condition.isSurroundedByParentheses();
    if (conditionNeedsParens) {
      this.insert(this.condition.outerStart, "(");
    }
    if (this.node.isUntil) {
      this.condition.negate();
    }
    this.condition.patch({ needsParens: false });
    if (this.guard) {
      const guardNeedsParens = !this.guard.isSurroundedByParentheses();
      if (this.body && this.body.inline()) {
        this.overwrite(this.condition.outerEnd, this.guard.outerStart, `${conditionNeedsParens ? ")" : ""} { if ${guardNeedsParens ? "(" : ""}`);
      } else {
        this.overwrite(this.condition.outerEnd, this.guard.outerStart, `${conditionNeedsParens ? ")" : ""} {
${this.getOuterLoopBodyIndent()}if ${guardNeedsParens ? "(" : ""}`);
      }
      this.guard.patch({ needsParens: false });
      this.insert(this.guard.outerEnd, `${guardNeedsParens ? ")" : ""} {`);
    } else {
      this.insert(this.condition.outerEnd, `${conditionNeedsParens ? ")" : ""} {`);
    }
    const thenIndex = this.getThenTokenIndex();
    if (thenIndex) {
      const thenToken = notNull(this.sourceTokenAtIndex(thenIndex));
      const nextToken = this.sourceTokenAtIndex(notNull(thenIndex.next()));
      if (nextToken) {
        this.remove(thenToken.start, nextToken.start);
      } else {
        this.remove(thenToken.start, thenToken.end);
      }
    }
    this.patchPossibleNewlineAfterLoopHeader(this.guard ? this.guard.outerEnd : this.condition.outerEnd);
    this.patchBody();
    if (this.guard) {
      if (this.body) {
        this.body.insertLineAfter("}", this.getOuterLoopBodyIndent());
      } else {
        this.insert(this.contentEnd, "}");
      }
    }
    if (this.body) {
      this.body.insertLineAfter("}", this.getLoopIndent());
    } else {
      this.insert(this.contentEnd, "}");
    }
  }
  getWhileTokenIndex() {
    const whileTokenIndex = this.contentStartTokenIndex;
    const whileToken = this.sourceTokenAtIndex(whileTokenIndex);
    if (!whileToken || whileToken.type !== SourceType38.WHILE) {
      throw this.error(`could not get first token of 'while' loop`);
    }
    return whileTokenIndex;
  }
  getThenTokenIndex() {
    const whileTokenIndex = this.getWhileTokenIndex();
    if (!whileTokenIndex) {
      throw this.error(`could not get first token of 'while' loop`);
    }
    let searchStart;
    if (this.guard) {
      searchStart = this.guard.outerEnd;
    } else {
      searchStart = this.condition.outerEnd;
    }
    let searchEnd;
    if (this.body) {
      searchEnd = this.body.outerStart;
    } else {
      const nextToken = this.nextSemanticToken();
      if (nextToken) {
        searchEnd = nextToken.end;
      } else {
        searchEnd = this.contentEnd;
      }
    }
    return this.indexOfSourceTokenBetweenSourceIndicesMatching(searchStart, searchEnd, (token) => token.type === SourceType38.THEN);
  }
  getLoopBodyIndent() {
    if (this.guard) {
      return this.getOuterLoopBodyIndent() + this.getProgramIndentString();
    } else {
      return this.getOuterLoopBodyIndent();
    }
  }
  willPatchAsIIFE() {
    return this.willPatchAsExpression();
  }
};

// src/utils/ternaryNeedsParens.ts
function ternaryNeedsParens(patcher) {
  if (patcher.hadUnparenthesizedNegation()) {
    return true;
  }
  const { parent } = patcher;
  return !(patcher.isSurroundedByParentheses() || parent instanceof FunctionApplicationPatcher && patcher !== parent.fn || parent instanceof DynamicMemberAccessOpPatcher && patcher === parent.indexingExpr || parent instanceof ConditionalPatcher && !parent.node.isUnless && patcher === parent.condition && !parent.willPatchAsTernary() || parent instanceof WhilePatcher && !parent.node.isUntil && patcher === parent.condition || parent instanceof InOpPatcher || parent instanceof AssignOpPatcher && patcher === parent.expression || parent instanceof SoakedMemberAccessOpPatcher && patcher === parent.expression || parent instanceof SoakedDynamicMemberAccessOpPatcher && patcher === parent.expression || parent instanceof SoakedFunctionApplicationPatcher && patcher === parent.fn);
}

// src/stages/main/patchers/SoakedFunctionApplicationPatcher.ts
var GUARD_FUNC_HELPER = `function __guardFunc__(func, transform) {
  return typeof func === 'function' ? transform(func) : undefined;
}`;
var GUARD_METHOD_HELPER = `function __guardMethod__(obj, methodName, transform) {
  if (typeof obj !== 'undefined' && obj !== null && typeof obj[methodName] === 'function') {
    return transform(obj, methodName);
  } else {
    return undefined;
  }
}`;
var SoakedFunctionApplicationPatcher = class extends FunctionApplicationPatcher {
  patchAsExpression() {
    if (this.shouldPatchAsConditional()) {
      this.patchAsConditional();
    } else {
      if (this.fn instanceof MemberAccessOpPatcher) {
        this.patchMethodCall(this.fn);
      } else if (this.fn instanceof DynamicMemberAccessOpPatcher) {
        this.patchDynamicMethodCall(this.fn);
      } else {
        this.patchNonMethodCall();
      }
      super.patchAsExpression();
    }
  }
  shouldPatchAsConditional() {
    return this.fn.isRepeatable() && !nodeContainsSoakOperation(this.fn.node);
  }
  patchAsConditional() {
    const soakContainer = findSoakContainer(this);
    this.fn.setRequiresRepeatableExpression();
    super.patchAsExpression();
    const fnCode = this.fn.getRepeatCode();
    const conditionCode = `typeof ${fnCode} === 'function'`;
    const callStartToken = this.getCallStartToken();
    this.remove(this.fn.outerEnd, callStartToken.start);
    if (soakContainer.willPatchAsExpression()) {
      const containerNeedsParens = ternaryNeedsParens(soakContainer);
      if (containerNeedsParens) {
        soakContainer.insert(soakContainer.contentStart, "(");
      }
      soakContainer.insert(soakContainer.contentStart, `${conditionCode} ? `);
      soakContainer.appendDeferredSuffix(" : undefined");
      if (containerNeedsParens) {
        soakContainer.appendDeferredSuffix(")");
      }
    } else {
      soakContainer.insert(soakContainer.contentStart, `if (${conditionCode}) {
${soakContainer.getIndent(1)}`);
      soakContainer.appendDeferredSuffix(`
${soakContainer.getIndent()}}`);
    }
  }
  patchMethodCall(fn) {
    const memberName = fn.getMemberName();
    if (fn.hasImplicitOperator()) {
      fn.setSkipImplicitDotCreation();
    }
    this.registerHelper("__guardMethod__", GUARD_METHOD_HELPER);
    this.addSuggestion(REMOVE_GUARD);
    if (fn instanceof SoakedMemberAccessOpPatcher) {
      fn.setShouldSkipSoakPatch();
    }
    const callStartToken = this.getCallStartToken();
    const soakContainer = findSoakContainer(this);
    const varName = soakContainer.claimFreeBinding("o");
    const prefix = this.slice(soakContainer.contentStart, this.fn.outerStart);
    if (prefix.length > 0) {
      this.remove(soakContainer.contentStart, this.fn.outerStart);
    }
    this.overwrite(fn.expression.outerEnd, callStartToken.end, `, '${memberName}', ${varName} => ${prefix}${varName}.${memberName}(`);
    soakContainer.insert(soakContainer.contentStart, "__guardMethod__(");
    soakContainer.appendDeferredSuffix(")");
  }
  patchDynamicMethodCall(fn) {
    const { expression, indexingExpr } = fn;
    this.registerHelper("__guardMethod__", GUARD_METHOD_HELPER);
    this.addSuggestion(REMOVE_GUARD);
    if (fn instanceof SoakedDynamicMemberAccessOpPatcher) {
      fn.setShouldSkipSoakPatch();
    }
    const callStartToken = this.getCallStartToken();
    const soakContainer = findSoakContainer(this);
    const objVarName = soakContainer.claimFreeBinding("o");
    const methodVarName = soakContainer.claimFreeBinding("m");
    const prefix = this.slice(soakContainer.contentStart, this.fn.outerStart);
    if (prefix.length > 0) {
      this.remove(soakContainer.contentStart, this.fn.outerStart);
    }
    this.overwrite(expression.outerEnd, indexingExpr.outerStart, `, `);
    this.overwrite(indexingExpr.outerEnd, callStartToken.end, `, (${objVarName}, ${methodVarName}) => ${prefix}${objVarName}[${methodVarName}](`);
    soakContainer.insert(soakContainer.contentStart, "__guardMethod__(");
    soakContainer.appendDeferredSuffix(")");
  }
  patchNonMethodCall() {
    this.registerHelper("__guardFunc__", GUARD_FUNC_HELPER);
    this.addSuggestion(REMOVE_GUARD);
    const callStartToken = this.getCallStartToken();
    const soakContainer = findSoakContainer(this);
    const varName = soakContainer.claimFreeBinding("f");
    const prefix = this.slice(soakContainer.contentStart, this.fn.outerStart);
    if (prefix.length > 0) {
      this.remove(soakContainer.contentStart, this.fn.outerStart);
    }
    this.overwrite(this.fn.outerEnd, callStartToken.end, `, ${varName} => ${prefix}${varName}(`);
    soakContainer.insert(soakContainer.contentStart, "__guardFunc__(");
    soakContainer.appendDeferredSuffix(")");
  }
  getCallStartToken() {
    const tokens = this.context.sourceTokens;
    const index = tokens.indexOfTokenMatchingPredicate((token) => token.type === SourceType39.CALL_START, this.fn.outerEndTokenIndex);
    if (!index || index.isAfter(this.contentEndTokenIndex)) {
      throw this.error(`unable to find open-paren for function call`);
    }
    return notNull(tokens.tokenAtIndex(index));
  }
};

// src/utils/findSoakContainer.ts
function findSoakContainer(patcher) {
  let result = patcher;
  while (canParentHandleSoak(result)) {
    result = notNull(result.parent);
  }
  return result;
}
function canParentHandleSoak(patcher) {
  if (patcher.parent === null) {
    return false;
  }
  if (patcher.isSurroundedByParentheses()) {
    return false;
  }
  if ((patcher.parent instanceof MemberAccessOpPatcher || patcher.parent instanceof DynamicMemberAccessOpPatcher) && patcher.parent.parent !== null && patcher.parent.parent instanceof SoakedFunctionApplicationPatcher && patcher.parent.parent.fn === patcher.parent) {
    return false;
  }
  if (patcher.parent instanceof MemberAccessOpPatcher && !(patcher.parent instanceof SoakedMemberAccessOpPatcher)) {
    return true;
  }
  if (patcher.parent instanceof DynamicMemberAccessOpPatcher && !(patcher.parent instanceof SoakedDynamicMemberAccessOpPatcher) && patcher.parent.expression === patcher) {
    return true;
  }
  if (patcher.parent instanceof FunctionApplicationPatcher && !(patcher.parent instanceof SoakedFunctionApplicationPatcher) && patcher.parent.fn === patcher) {
    return true;
  }
  if (patcher.parent instanceof AssignOpPatcher && patcher.parent.assignee === patcher) {
    return true;
  }
  if (["PostIncrementOp", "PostDecrementOp", "PreIncrementOp", "PreDecrementOp", "DeleteOp"].indexOf(patcher.parent.node.type) >= 0) {
    return true;
  }
  return false;
}

// src/stages/main/patchers/SoakedDynamicMemberAccessOpPatcher.ts
var GUARD_HELPER2 = `function __guard__(value, transform) {
  return (typeof value !== 'undefined' && value !== null) ? transform(value) : undefined;
}`;
var SoakedDynamicMemberAccessOpPatcher = class extends DynamicMemberAccessOpPatcher {
  constructor(patcherContext, expression, indexingExpr) {
    super(patcherContext, expression, indexingExpr);
    this._shouldSkipSoakPatch = false;
  }
  patchAsExpression() {
    if (!this._shouldSkipSoakPatch) {
      if (this.shouldPatchAsConditional()) {
        this.patchAsConditional();
      } else {
        this.patchAsGuardCall();
      }
    } else {
      this.expression.patch();
      this.indexingExpr.patch();
    }
  }
  shouldPatchAsConditional() {
    return this.expression.isRepeatable() && !nodeContainsSoakOperation(this.expression.node);
  }
  patchAsConditional() {
    this.addSuggestion(SHORTEN_NULL_CHECKS);
    const soakContainer = findSoakContainer(this);
    const expressionCode = this.expression.patchRepeatable();
    let conditionCode;
    if (this.expression.mayBeUnboundReference()) {
      conditionCode = `typeof ${expressionCode} !== 'undefined' && ${expressionCode} !== null`;
    } else {
      conditionCode = `${expressionCode} != null`;
    }
    this.overwrite(this.expression.outerEnd, this.indexingExpr.outerStart, "[");
    this.indexingExpr.patch();
    if (soakContainer.willPatchAsExpression()) {
      const containerNeedsParens = ternaryNeedsParens(soakContainer);
      if (containerNeedsParens) {
        soakContainer.insert(soakContainer.contentStart, "(");
      }
      soakContainer.insert(soakContainer.contentStart, `${conditionCode} ? `);
      soakContainer.appendDeferredSuffix(" : undefined");
      if (containerNeedsParens) {
        soakContainer.appendDeferredSuffix(")");
      }
    } else {
      soakContainer.insert(soakContainer.contentStart, `if (${conditionCode}) {
${soakContainer.getIndent(1)}`);
      soakContainer.appendDeferredSuffix(`
${soakContainer.getIndent()}}`);
    }
  }
  patchAsGuardCall() {
    this.registerHelper("__guard__", GUARD_HELPER2);
    this.addSuggestion(REMOVE_GUARD);
    const soakContainer = findSoakContainer(this);
    const varName = soakContainer.claimFreeBinding("x");
    const prefix = this.slice(soakContainer.contentStart, this.contentStart);
    if (prefix.length > 0) {
      this.remove(soakContainer.contentStart, this.contentStart);
    }
    this.overwrite(this.expression.outerEnd, this.indexingExpr.outerStart, `, ${varName} => ${prefix}${varName}[`);
    soakContainer.insert(soakContainer.contentStart, "__guard__(");
    soakContainer.appendDeferredSuffix(")");
    this.expression.patch();
    this.indexingExpr.patch();
  }
  setShouldSkipSoakPatch() {
    this._shouldSkipSoakPatch = true;
  }
};

// src/stages/main/patchers/SoakedNewOpPatcher.ts
var SoakedNewOpPatcher = class extends SoakedFunctionApplicationPatcher {
  patchMethodCall() {
    this.patchNonMethodCall();
  }
  patchDynamicMethodCall() {
    this.patchNonMethodCall();
  }
};

// src/stages/main/patchers/SoakedSlicePatcher.ts
var GUARD_HELPER3 = `function __guard__(value, transform) {
  return (typeof value !== 'undefined' && value !== null) ? transform(value) : undefined;
}`;
var SoakedSlicePatcher = class extends SlicePatcher {
  patchAsExpression() {
    this.registerHelper("__guard__", GUARD_HELPER3);
    this.addSuggestion(REMOVE_GUARD);
    const soakContainer = findSoakContainer(this);
    const varName = soakContainer.claimFreeBinding("x");
    const prefix = this.slice(soakContainer.contentStart, this.contentStart);
    if (prefix.length > 0) {
      this.remove(soakContainer.contentStart, this.contentStart);
    }
    this.insert(this.expression.outerEnd, `, ${varName} => ${prefix}${varName}`);
    soakContainer.insert(soakContainer.contentStart, "__guard__(");
    super.patchAsExpression();
    soakContainer.appendDeferredSuffix(")");
  }
  getSpliceCode(expressionCode) {
    const spliceStart = this.captureCodeForPatchOperation(() => {
      this.registerHelper("__guard__", GUARD_HELPER3);
      this.addSuggestion(REMOVE_GUARD);
      const varName = this.claimFreeBinding("x");
      this.insert(this.expression.outerEnd, `, ${varName} => ${varName}`);
      this.patchAsSpliceExpressionStart();
    });
    return `__guard__(${spliceStart}, ...[].concat(${expressionCode})))`;
  }
};

// src/stages/main/patchers/SuperPatcher.ts
import { SourceType as SourceType40 } from "coffee-lex";
var SuperPatcher = class extends NodePatcher {
  patchAsExpression() {
    const { classCode, accessCode } = this.getEnclosingMethodInfo();
    if (this.canConvertToJsSuper()) {
      if (accessCode) {
        this.insert(this.contentEnd, accessCode);
      }
    } else {
      if (!accessCode) {
        throw this.error("Cannot handle a super call in an inner function in a constructor. Please either rewrite your CoffeeScript code to not use this construct or file a bug to discuss ways that decaffeinate could handle this case.");
      }
      if (!classCode) {
        throw this.error("Complex super calls within anonymous classes are not yet supported.");
      }
      const openParenToken = this.getFollowingOpenParenToken();
      this.overwrite(this.contentStart, openParenToken.end, `${classCode}.prototype.__proto__${accessCode}.call(this, `);
    }
  }
  getEnclosingMethodInfo() {
    const methodAssignment = this.getEnclosingMethodAssignment();
    if (methodAssignment instanceof ClassAssignOpPatcher) {
      let accessCode;
      if (methodAssignment.isStaticMethod()) {
        if (methodAssignment.key instanceof MemberAccessOpPatcher) {
          accessCode = `.${methodAssignment.key.node.member.data}`;
        } else if (methodAssignment.key instanceof DynamicMemberAccessOpPatcher) {
          accessCode = `[${methodAssignment.key.indexingExpr.getRepeatCode()}]`;
        } else {
          throw this.error("Unexpected key type for static method.");
        }
      } else {
        if (methodAssignment.key instanceof IdentifierPatcher) {
          accessCode = `.${methodAssignment.key.node.data}`;
        } else {
          accessCode = `[${methodAssignment.key.getRepeatCode()}]`;
        }
      }
      return {
        classCode: this.getEnclosingClassName(methodAssignment),
        accessCode
      };
    } else if (methodAssignment instanceof ConstructorPatcher) {
      return {
        classCode: this.getEnclosingClassName(methodAssignment),
        accessCode: null
      };
    } else {
      const methodInfo = this.getPrototypeAssignInfo(methodAssignment);
      if (!methodInfo) {
        throw this.error("Expected a valid method assignment from getEnclosingMethodAssignment.");
      }
      return methodInfo;
    }
  }
  getEnclosingClassName(patcher) {
    let { parent } = patcher;
    while (parent) {
      if (parent instanceof ClassPatcher) {
        return parent.getName();
      }
      parent = parent.parent;
    }
    throw this.error("Expected super expression to be in a class body.");
  }
  getEnclosingMethodAssignment() {
    let { parent } = this;
    while (parent) {
      if (parent instanceof ClassAssignOpPatcher || parent instanceof ConstructorPatcher || this.getPrototypeAssignInfo(parent) !== null) {
        return parent;
      }
      parent = parent.parent;
    }
    throw this.error("super called in a context where we cannot determine the class and method name.");
  }
  getPrototypeAssignInfo(patcher) {
    const prototypeAssignPatchers = extractPrototypeAssignPatchers(patcher);
    if (!prototypeAssignPatchers) {
      return null;
    }
    const { classRefPatcher, methodAccessPatcher } = prototypeAssignPatchers;
    if (methodAccessPatcher instanceof MemberAccessOpPatcher) {
      return {
        classCode: classRefPatcher.getRepeatCode(),
        accessCode: `.${methodAccessPatcher.member.node.data}`
      };
    } else if (methodAccessPatcher instanceof DynamicMemberAccessOpPatcher) {
      return {
        classCode: classRefPatcher.getRepeatCode(),
        accessCode: `[${methodAccessPatcher.indexingExpr.getRepeatCode()}]`
      };
    } else {
      throw this.error("Expected the method access patcher to be either MemberAccessOpPatcher or DynamicMemberAccessOpPatcher.");
    }
  }
  canConvertToJsSuper() {
    const methodAssignment = this.getEnclosingMethodAssignment();
    if (methodAssignment instanceof ConstructorPatcher || methodAssignment instanceof ClassAssignOpPatcher) {
      return methodAssignment.expression === this.getEnclosingFunction();
    }
    return false;
  }
  getEnclosingFunction() {
    let { parent } = this;
    while (parent) {
      if (parent instanceof FunctionPatcher) {
        return parent;
      }
      parent = parent.parent;
    }
    throw this.error("super called outside of a function.");
  }
  getFollowingOpenParenToken() {
    const openParenTokenIndex = this.indexOfSourceTokenAfterSourceTokenIndex(this.contentEndTokenIndex, SourceType40.CALL_START);
    if (!openParenTokenIndex) {
      throw this.error("Expected open-paren after super.");
    }
    return notNull(this.sourceTokenAtIndex(openParenTokenIndex));
  }
};

// src/stages/main/patchers/SwitchCasePatcher.ts
import { SourceType as SourceType41 } from "coffee-lex";
var SwitchCasePatcher = class extends NodePatcher {
  constructor(patcherContext, conditions, consequent) {
    super(patcherContext);
    this.conditions = conditions;
    this.consequent = consequent;
    this.negated = false;
  }
  initialize() {
    this.conditions.forEach((condition) => condition.setRequiresExpression());
  }
  patchAsStatement() {
    const lastCondition = this.conditions[this.conditions.length - 1];
    const whenToken = this.getWhenToken();
    this.remove(whenToken.start, this.conditions[0].outerStart);
    this.getCommaTokens().forEach((comma) => {
      this.remove(comma.start, comma.end);
    });
    this.conditions.forEach((condition) => {
      this.insert(condition.outerStart, "case ");
      if (this.negated) {
        condition.negate();
      }
      condition.patch({ leftBrace: false, rightBrace: false });
      this.insert(condition.outerEnd, ":");
    });
    const thenToken = this.getThenToken();
    if (thenToken) {
      if (this.consequent !== null) {
        this.remove(thenToken.start, this.consequent.outerStart);
      } else {
        this.remove(lastCondition.outerEnd, thenToken.end);
      }
    }
    if (this.consequent !== null) {
      this.consequent.patch({ leftBrace: false, rightBrace: false });
    }
    const implicitReturnWillBreak = this.implicitlyReturns() && this.implicitReturnPatcher().implicitReturnWillBreak() && (!this.consequent || this.consequent.allCodePathsPresent());
    const shouldAddBreak = !this.hasExistingBreak() && !implicitReturnWillBreak;
    if (shouldAddBreak) {
      if (thenToken) {
        if (this.consequent !== null) {
          this.insert(this.consequent.contentEnd, " break");
        } else {
          this.insert(lastCondition.outerEnd, `
${this.getIndent(1)}break`);
        }
      } else {
        this.appendLineAfter("break", 1);
      }
    }
  }
  setImplicitlyReturns() {
    super.setImplicitlyReturns();
    if (this.consequent !== null) {
      this.consequent.setImplicitlyReturns();
    }
  }
  patchAsExpression() {
    this.patchAsStatement();
  }
  negate() {
    this.negated = !this.negated;
  }
  getWhenToken() {
    const whenToken = this.sourceTokenAtIndex(this.contentStartTokenIndex);
    if (!whenToken) {
      throw this.error(`bad token index for start of 'when'`);
    }
    if (whenToken.type !== SourceType41.WHEN) {
      throw this.error(`unexpected ${SourceType41[whenToken.type]} at start of 'switch' case`);
    }
    return whenToken;
  }
  getCommaTokens() {
    const result = [];
    for (let i = 1; i < this.conditions.length; i++) {
      const left = this.conditions[i - 1];
      const right = this.conditions[i];
      const commaIndex = this.indexOfSourceTokenBetweenPatchersMatching(left, right, (token) => token.type === SourceType41.COMMA);
      if (!commaIndex) {
        throw this.error(`unable to find comma between 'when' conditions`, left.contentEnd, right.contentStart);
      }
      result.push(notNull(this.sourceTokenAtIndex(commaIndex)));
    }
    return result;
  }
  hasExistingBreak() {
    if (!(this.consequent instanceof BlockPatcher)) {
      return false;
    }
    const lastStatement = this.consequent.statements[this.consequent.statements.length - 1];
    return lastStatement && lastStatement instanceof BreakPatcher;
  }
  getThenToken() {
    const thenTokenIndex = this.indexOfSourceTokenBetweenSourceIndicesMatching(this.conditions[0].outerEnd, this.consequent !== null ? this.consequent.outerStart : this.contentEnd, (token) => token.type === SourceType41.THEN);
    if (thenTokenIndex) {
      return this.sourceTokenAtIndex(thenTokenIndex);
    }
    const nextToken = this.nextSemanticToken();
    if (nextToken && nextToken.type === SourceType41.THEN) {
      return nextToken;
    }
    return null;
  }
};

// src/stages/main/patchers/TaggedTemplateLiteralPatcher.ts
var TaggedTemplateLiteralPatcher = class extends NodePatcher {
  constructor(patcherContext, tag, template) {
    super(patcherContext);
    this.tag = tag;
    this.template = template;
  }
  patchAsExpression() {
    this.tag.patch();
    this.template.patch({ forceTemplateLiteral: true });
  }
};

// src/stages/main/patchers/ThrowPatcher.ts
import { SourceType as SourceType42 } from "coffee-lex";
var ThrowPatcher = class extends NodePatcher {
  constructor(patcherContext, expression) {
    super(patcherContext);
    this.expression = expression;
  }
  initialize() {
    this.expression.setRequiresExpression();
  }
  prefersToPatchAsExpression() {
    return false;
  }
  setImplicitlyReturns() {
  }
  patchAsExpression() {
    const hasParens = this.isSurroundedByParentheses();
    if (!hasParens) {
      this.insert(this.outerStart, "(");
    }
    this.insert(this.innerStart, "() => { ");
    this.patchAsStatement();
    this.insert(this.innerEnd, " }");
    if (!hasParens) {
      this.insert(this.outerEnd, ")");
    }
    this.insert(this.outerEnd, "()");
  }
  patchAsStatement() {
    const throwToken = notNull(this.sourceTokenAtIndex(this.contentStartTokenIndex));
    if (throwToken.type !== SourceType42.THROW) {
      throw this.error("Expected to find throw token at the start of throw statement.");
    }
    const spacing = this.slice(throwToken.end, this.expression.outerStart);
    if (spacing.indexOf("\n") !== -1) {
      this.overwrite(throwToken.end, this.expression.outerStart, " ");
    }
    this.expression.patch();
  }
  allowPatchingOuterBounds() {
    return true;
  }
};

// src/stages/main/patchers/TryPatcher.ts
import { SourceType as SourceType43 } from "coffee-lex";
var TryPatcher = class extends NodePatcher {
  constructor(patcherContext, body, catchAssignee, catchBody, finallyBody) {
    super(patcherContext);
    this._errorBinding = null;
    this.body = body;
    this.catchAssignee = catchAssignee;
    this.catchBody = catchBody;
    this.finallyBody = finallyBody;
  }
  initialize() {
    if (this.catchAssignee) {
      this.catchAssignee.setAssignee();
      this.catchAssignee.setRequiresExpression();
    }
    getEnclosingScopeBlock(this).markIIFEPatcherDescendant(this);
  }
  canPatchAsExpression() {
    if (this.body && !this.body.canPatchAsExpression()) {
      return false;
    }
    if (this.catchBody && !this.catchBody.canPatchAsExpression()) {
      return false;
    }
    if (this.finallyBody && !this.finallyBody.canPatchAsExpression()) {
      return false;
    }
    return true;
  }
  patchAsStatement() {
    const tryToken = this.getTryToken();
    const catchToken = this.getCatchToken();
    const thenTokenIndex = this.getThenTokenIndex();
    const finallyToken = this.getFinallyToken();
    this.insert(tryToken.end, ` {`);
    if (this.body) {
      if (this.body.inline()) {
        this.body.patch({ leftBrace: false });
      } else {
        if (catchToken || finallyToken) {
          this.body.patch({ leftBrace: false, rightBrace: false });
          this.insert(notNull(catchToken || finallyToken).start, "} ");
        } else {
          this.body.patch({ leftBrace: false });
        }
      }
    } else {
      this.insert(tryToken.end, "}");
    }
    if (thenTokenIndex) {
      const thenToken = notNull(this.sourceTokenAtIndex(thenTokenIndex));
      const nextToken = this.sourceTokenAtIndex(notNull(thenTokenIndex.next()));
      if (nextToken) {
        this.remove(thenToken.start, nextToken.start);
      } else {
        this.remove(thenToken.start, thenToken.end);
      }
    }
    if (catchToken) {
      const afterCatchHeader = this.catchAssignee ? this.catchAssignee.outerEnd : catchToken.end;
      if (this.catchAssignee) {
        const addErrorParens = !this.catchAssignee.isSurroundedByParentheses();
        if (addErrorParens) {
          this.insert(this.catchAssignee.outerStart, "(");
        }
        this.catchAssignee.patch();
        if (addErrorParens) {
          this.insert(this.catchAssignee.outerEnd, ")");
        }
      } else {
        this.insert(afterCatchHeader, ` (${this.getErrorBinding()})`);
      }
      if (this.catchBody) {
        this.insert(afterCatchHeader, " {");
        this.catchBody.patch({ leftBrace: false });
      } else {
        this.insert(afterCatchHeader, " {}");
      }
    } else if (!finallyToken) {
      const insertPos = this.body ? this.body.innerEnd : tryToken.end;
      this.insert(insertPos, ` catch (${this.getErrorBinding()}) {}`);
    }
    if (finallyToken) {
      if (!this.finallyBody) {
        this.insert(finallyToken.end, " {}");
      } else if (this.finallyBody.inline()) {
        this.finallyBody.patch();
      } else {
        this.insert(finallyToken.end, " {");
        this.finallyBody.patch({ leftBrace: false });
      }
    }
  }
  patchAsExpression() {
    this.setImplicitlyReturns();
    this.patchInIIFE(() => {
      this.insert(this.innerStart, " ");
      this.patchAsStatement();
      this.insert(this.innerEnd, " ");
    });
  }
  willPatchAsIIFE() {
    return this.willPatchAsExpression();
  }
  canHandleImplicitReturn() {
    return this.willPatchAsExpression();
  }
  implicitlyReturns() {
    return super.implicitlyReturns() && this.willPatchAsExpression();
  }
  setImplicitlyReturns() {
    super.setImplicitlyReturns();
    if (this.body) {
      this.body.setImplicitlyReturns();
    }
    if (this.catchBody) {
      this.catchBody.setImplicitlyReturns();
    }
  }
  statementNeedsSemicolon() {
    return false;
  }
  getTryToken() {
    const tryTokenIndex = this.contentStartTokenIndex;
    const tryToken = this.sourceTokenAtIndex(tryTokenIndex);
    if (!tryToken || tryToken.type !== SourceType43.TRY) {
      throw this.error(`expected 'try' keyword at start of 'try' statement`);
    }
    return tryToken;
  }
  getCatchToken() {
    let searchStart;
    if (this.body) {
      searchStart = this.body.outerEnd;
    } else {
      searchStart = this.getTryToken().end;
    }
    let searchEnd;
    if (this.catchAssignee) {
      searchEnd = this.catchAssignee.outerStart;
    } else if (this.catchBody) {
      searchEnd = this.catchBody.outerStart;
    } else if (this.finallyBody) {
      searchEnd = this.finallyBody.outerStart;
    } else {
      searchEnd = this.contentEnd;
    }
    const catchTokenIndex = this.indexOfSourceTokenBetweenSourceIndicesMatching(searchStart, searchEnd, (token) => token.type === SourceType43.CATCH);
    if (!catchTokenIndex) {
      return null;
    }
    return this.sourceTokenAtIndex(catchTokenIndex);
  }
  getThenTokenIndex() {
    let searchStart;
    if (this.catchAssignee) {
      searchStart = this.catchAssignee.outerEnd;
    } else if (this.body) {
      searchStart = this.body.outerEnd;
    } else {
      searchStart = this.getTryToken().end;
    }
    let searchEnd;
    if (this.catchBody) {
      searchEnd = this.catchBody.outerStart;
    } else if (this.finallyBody) {
      searchEnd = this.finallyBody.outerStart;
    } else {
      const nextToken = this.nextSemanticToken();
      if (nextToken) {
        searchEnd = nextToken.end;
      } else {
        searchEnd = this.contentEnd;
      }
    }
    return this.indexOfSourceTokenBetweenSourceIndicesMatching(searchStart, searchEnd, (token) => token.type === SourceType43.THEN);
  }
  getFinallyToken() {
    let searchStart;
    if (this.catchBody) {
      searchStart = this.catchBody.outerEnd;
    } else if (this.catchAssignee) {
      searchStart = this.catchAssignee.outerEnd;
    } else if (this.body) {
      searchStart = this.body.outerEnd;
    } else {
      searchStart = this.getTryToken().end;
    }
    let searchEnd;
    if (this.finallyBody) {
      searchEnd = this.finallyBody.outerStart;
    } else {
      searchEnd = this.contentEnd;
    }
    const finallyTokenIndex = this.indexOfSourceTokenBetweenSourceIndicesMatching(searchStart, searchEnd, (token) => token.type === SourceType43.FINALLY);
    if (!finallyTokenIndex) {
      return null;
    }
    return this.sourceTokenAtIndex(finallyTokenIndex);
  }
  getErrorBinding() {
    if (!this._errorBinding) {
      this._errorBinding = this.claimFreeBinding("error");
    }
    return this._errorBinding;
  }
};

// src/stages/main/patchers/UnaryExistsOpPatcher.ts
var UnaryExistsOpPatcher = class extends UnaryOpPatcher {
  constructor() {
    super(...arguments);
    this.negated = false;
  }
  patchAsExpression({ needsParens = true } = {}) {
    const addParens = needsParens && !this.isSurroundedByParentheses();
    if (addParens) {
      this.insert(this.contentStart, "(");
    }
    this.patchAsStatement();
    if (addParens) {
      this.insert(this.contentEnd, ")");
    }
  }
  patchAsStatement() {
    this.addSuggestion(SHORTEN_NULL_CHECKS);
    const {
      node: { expression },
      negated
    } = this;
    const needsTypeofCheck = this.needsTypeofCheck();
    this.expression.patch();
    if (needsTypeofCheck) {
      if (negated) {
        this.overwrite(this.contentStart, this.contentEnd, `typeof ${expression.raw} === 'undefined' || ${expression.raw} === null`);
      } else {
        this.overwrite(this.contentStart, this.contentEnd, `typeof ${expression.raw} !== 'undefined' && ${expression.raw} !== null`);
      }
    } else {
      if (negated) {
        this.overwrite(this.expression.outerEnd, this.contentEnd, " == null");
      } else {
        this.overwrite(this.expression.outerEnd, this.contentEnd, " != null");
      }
    }
  }
  canHandleNegationInternally() {
    return true;
  }
  negate() {
    this.negated = !this.negated;
  }
  needsTypeofCheck() {
    return this.expression.mayBeUnboundReference();
  }
  statementNeedsParens() {
    if (this.needsTypeofCheck()) {
      return false;
    } else {
      return this.expression.statementShouldAddParens();
    }
  }
};

// src/stages/main/patchers/UnaryTypeofOpPatcher.ts
var UnaryTypeofOpPatcher = class extends UnaryOpPatcher {
  isRepeatable() {
    return this.expression.isRepeatable();
  }
  statementNeedsParens() {
    return false;
  }
};

// src/stages/main/patchers/YieldPatcher.ts
var YieldPatcher = class extends NodePatcher {
  constructor(patcherContext, expression) {
    super(patcherContext);
    this.expression = expression;
  }
  initialize() {
    this.yields();
    if (this.expression) {
      this.expression.setRequiresExpression();
    }
  }
  patchAsExpression({ needsParens = true } = {}) {
    const surroundInParens = this.needsParens() && !this.isSurroundedByParentheses();
    if (surroundInParens) {
      this.insert(this.contentStart, "(");
    }
    if (this.expression) {
      this.expression.patch({ needsParens });
    }
    if (surroundInParens) {
      this.insert(this.contentEnd, ")");
    }
  }
  needsParens() {
    return !(this.parent instanceof BlockPatcher || this.parent instanceof ReturnPatcher || this.parent instanceof AssignOpPatcher && this.parent.expression === this);
  }
};

// src/stages/main/patchers/YieldFromPatcher.ts
var YieldFromPatcher = class extends YieldPatcher {
  patchAsExpression({ needsParens = true } = {}) {
    const firstToken = this.firstToken();
    this.overwrite(firstToken.start, firstToken.end, "yield*");
    super.patchAsExpression({ needsParens });
  }
};

// src/stages/main/patchers/YieldReturnPatcher.ts
import { SourceType as SourceType44 } from "coffee-lex";
var YieldReturnPatcher = class extends ReturnPatcher {
  initialize() {
    this.yields();
    super.initialize();
  }
  patchAsStatement() {
    const yieldTokenIndex = this.contentStartTokenIndex;
    const returnTokenIndex = notNull(yieldTokenIndex.next());
    const yieldToken = notNull(this.sourceTokenAtIndex(yieldTokenIndex));
    const returnToken = notNull(this.sourceTokenAtIndex(returnTokenIndex));
    if (yieldToken.type !== SourceType44.YIELD || returnToken.type !== SourceType44.RETURN) {
      throw this.error("Unexpected token types for `yield return`.");
    }
    this.remove(yieldToken.start, returnToken.start);
    super.patchAsStatement();
  }
};

// src/stages/main/index.ts
var MainStage = class extends TransformCoffeeScriptStage {
  patcherConstructorForNode(node) {
    switch (node.type) {
      case "Identifier":
        return IdentifierPatcher;
      case "String":
        return StringPatcher;
      case "TaggedTemplateLiteral":
        return TaggedTemplateLiteralPatcher;
      case "Int":
      case "Float":
      case "Null":
      case "Undefined":
        return PassthroughPatcher;
      case "Break":
        return BreakPatcher;
      case "Continue":
        return ContinuePatcher;
      case "Debugger":
        return DebuggerPatcher;
      case "Quasi":
        return QuasiPatcher;
      case "FunctionApplication":
        return FunctionApplicationPatcher;
      case "SoakedFunctionApplication":
        if (this.options.optionalChaining) {
          return OptionalChainingSoakedFunctionApplicationPatcher;
        } else {
          return SoakedFunctionApplicationPatcher;
        }
      case "MemberAccessOp":
        return MemberAccessOpPatcher;
      case "DynamicMemberAccessOp":
        return DynamicMemberAccessOpPatcher;
      case "EQOp":
      case "NEQOp":
      case "LTOp":
      case "GTOp":
      case "LTEOp":
      case "GTEOp":
        return EqualityPatcher;
      case "PostIncrementOp":
      case "PostDecrementOp":
      case "PreIncrementOp":
      case "PreDecrementOp":
        return IncrementDecrementPatcher;
      case "ObjectInitialiserMember":
        return ObjectInitialiserMemberPatcher;
      case "ObjectInitialiser":
        return ObjectInitialiserPatcher;
      case "This":
        return ThisPatcher;
      case "Yield":
        return YieldPatcher;
      case "YieldFrom":
        return YieldFromPatcher;
      case "YieldReturn":
        return YieldReturnPatcher;
      case "Await":
        return AwaitPatcher;
      case "AwaitReturn":
        return AwaitReturnPatcher;
      case "GeneratorFunction":
        return GeneratorFunctionPatcher;
      case "Function":
        return FunctionPatcher;
      case "BoundFunction":
        return BoundFunctionPatcher;
      case "BoundGeneratorFunction":
        return BoundGeneratorFunctionPatcher;
      case "AsyncFunction":
        return AsyncFunctionPatcher;
      case "BoundAsyncFunction":
        return BoundAsyncFunctionPatcher;
      case "Bool":
        return BoolPatcher;
      case "Conditional":
        return ConditionalPatcher;
      case "ArrayInitialiser":
        return ArrayInitialiserPatcher;
      case "Block":
        return BlockPatcher;
      case "AssignOp":
        return AssignOpPatcher;
      case "DefaultParam":
        return DefaultParamPatcher;
      case "CompoundAssignOp":
        switch (node.op) {
          case "LogicalAndOp":
          case "LogicalOrOp":
            if (this.options.logicalAssignment) {
              return CompoundAssignOpPatcher;
            } else {
              return LogicalOpCompoundAssignOpPatcher;
            }
          case "ExistsOp":
            if (this.options.logicalAssignment) {
              return NullishCoalescingCompoundAssignOpPatcher;
            } else {
              return ExistsOpCompoundAssignOpPatcher;
            }
          case "ModuloOp":
            return ModuloOpCompoundAssignOpPatcher;
          case "FloorDivideOp":
            return FloorDivideOpCompoundAssignOpPatcher;
          default:
            return CompoundAssignOpPatcher;
        }
      case "Return":
        return ReturnPatcher;
      case "PlusOp":
      case "SubtractOp":
      case "DivideOp":
      case "MultiplyOp":
      case "RemOp":
      case "BitAndOp":
      case "BitOrOp":
      case "BitXorOp":
      case "LeftShiftOp":
      case "SignedRightShiftOp":
      case "UnsignedRightShiftOp":
        return BinaryOpPatcher;
      case "ModuloOp":
        return ModuloOpPatcher;
      case "Regex":
        return RegexPatcher;
      case "Heregex":
        return HeregexPatcher;
      case "ExistsOp":
        if (this.options.nullishCoalescing) {
          return NullishCoalescingExistsOpPatcher;
        } else {
          return ExistsOpPatcher;
        }
      case "LogicalAndOp":
      case "LogicalOrOp":
        return LogicalOpPatcher;
      case "LogicalNotOp":
        return LogicalNotOpPatcher;
      case "SoakedMemberAccessOp":
        if (this.options.optionalChaining) {
          return OptionalChainingSoakedMemberAccessOpPatcher;
        } else {
          return SoakedMemberAccessOpPatcher;
        }
      case "SoakedDynamicMemberAccessOp":
        if (this.options.optionalChaining) {
          return OptionalChainingSoakedDynamicMemberAccessOpPatcher;
        } else {
          return SoakedDynamicMemberAccessOpPatcher;
        }
      case "ForIn":
        return ForInPatcher;
      case "ForFrom":
        return ForFromPatcher;
      case "ForOf":
        return ForOfPatcher;
      case "While":
        return WhilePatcher;
      case "NewOp":
        return NewOpPatcher;
      case "SoakedNewOp":
        if (this.options.optionalChaining) {
          return OptionalChainingSoakedNewOpPatcher;
        } else {
          return SoakedNewOpPatcher;
        }
      case "InOp":
        return InOpPatcher;
      case "Slice":
        return SlicePatcher;
      case "SoakedSlice":
        if (this.options.optionalChaining) {
          return OptionalChainingSoakedSlicePatcher;
        } else {
          return SoakedSlicePatcher;
        }
      case "Expansion":
        return ExpansionPatcher;
      case "Elision":
        return ElisionPatcher;
      case "Rest":
        return RestPatcher;
      case "Spread":
        return SpreadPatcher;
      case "Range":
        return RangePatcher;
      case "Throw":
        return ThrowPatcher;
      case "UnaryPlusOp":
      case "UnaryNegateOp":
      case "BitNotOp":
        return UnaryMathOpPatcher;
      case "TypeofOp":
        return UnaryTypeofOpPatcher;
      case "DeleteOp":
        return UnaryOpPatcher;
      case "UnaryExistsOp":
        return UnaryExistsOpPatcher;
      case "ClassProtoAssignOp":
        return ClassAssignOpPatcher;
      case "Super":
        return SuperPatcher;
      case "Class":
        return ClassPatcher;
      case "Constructor":
        return ConstructorPatcher;
      case "Try":
        return TryPatcher;
      case "Switch":
        return SwitchPatcher;
      case "SwitchCase":
        return SwitchCasePatcher;
      case "DoOp":
        return DoOpPatcher;
      case "Program":
        return ProgramPatcher2;
      case "InstanceofOp":
        return InstanceofOpPatcher;
      case "OfOp":
        return OfOpPatcher;
      case "ChainedComparisonOp":
        return ChainedComparisonOpPatcher;
      case "SeqOp":
        return SeqOpPatcher;
      case "JavaScript":
        return JavaScriptPatcher;
      case "FloorDivideOp":
        return FloorDivideOpPatcher;
      case "ExpOp":
        return ExpOpPatcher;
      case "ExtendsOp":
        return ExtendsOpPatcher;
      case "CSXElement":
        return CSXElementPatcher;
      case "ExportBindingsDeclaration":
        return ExportBindingsDeclarationPatcher;
      case "ExportDefaultDeclaration":
        return ExportDefaultDeclarationPatcher;
      case "ExportAllDeclaration":
        return ExportAllDeclarationPatcher;
      case "ModuleSpecifier":
        return ModuleSpecifierPatcher;
      case "ImportDeclaration":
        return ImportDeclarationPatcher;
      case "ExportNamedDeclaration":
        return ExportNamedDeclarationPatcher;
      default:
        return null;
    }
  }
};

// src/utils/normalizeListItem.ts
import { SourceType as SourceType45 } from "coffee-lex";
import { Block, Elision as Elision2 } from "decaffeinate-parser";
function normalizeListItem(patcher, listItemPatcher, nextListItemPatcher) {
  const lastToken = listItemPatcher.lastToken();
  if (lastToken.type === SourceType45.COMMA && !(listItemPatcher.node instanceof Elision2) && nextListItemPatcher) {
    patcher.remove(lastToken.start, lastToken.end);
  }
  const nextToken = listItemPatcher.nextSemanticToken();
  if (nextToken && nextToken.type === SourceType45.SEMICOLON && nextToken.end <= patcher.contentEnd) {
    if (patcherEndsInStatement(listItemPatcher)) {
      patcher.remove(nextToken.start, nextToken.end);
    } else {
      patcher.overwrite(nextToken.start, nextToken.end, ",");
    }
  }
  if (nextListItemPatcher) {
    const commaTokens = patcher.getProgramSourceTokens().slice(notNull(listItemPatcher.outerEndTokenIndex.next()), nextListItemPatcher.outerStartTokenIndex).filter((token) => token.type === SourceType45.COMMA || token.type === SourceType45.SEMICOLON).toArray();
    if (nextListItemPatcher.node.type === "ObjectInitialiser" && !isNestedListItem(listItemPatcher) && commaTokens.length === 0) {
      patcher.insert(listItemPatcher.outerEnd, ",");
    }
  }
}
function isNestedListItem(patcher) {
  return [
    "BoundFunction",
    "BoundGeneratorFunction",
    "Conditional",
    "ForIn",
    "ForOf",
    "Function",
    "GeneratorFunction",
    "Switch",
    "Try",
    "While"
  ].indexOf(patcher.node.type) > -1;
}
function patcherEndsInStatement(patcher) {
  return containsDescendant(patcher.node, (child) => child instanceof Block && child.end === patcher.contentEnd);
}

// src/stages/normalize/patchers/ArrayInitialiserPatcher.ts
var ArrayInitialiserPatcher2 = class extends NodePatcher {
  constructor(patcherContext, members) {
    super(patcherContext);
    this.members = members;
  }
  patchAsExpression() {
    for (const [i, member] of this.members.entries()) {
      member.patch();
      normalizeListItem(this, member, this.members[i + 1]);
    }
  }
};

// src/stages/normalize/patchers/AssignOpPatcher.ts
import { SourceType as SourceType49 } from "coffee-lex";

// src/stages/normalize/patchers/ClassPatcher.ts
import { SourceType as SourceType48 } from "coffee-lex";

// src/stages/normalize/patchers/BlockPatcher.ts
import { SourceType as SourceType46 } from "coffee-lex";
var BlockPatcher2 = class extends SharedBlockPatcher {
  patchAsExpression() {
    this.patchAsStatement();
  }
  patchAsStatement() {
    if (this.node.inline) {
      for (const statement of this.statements) {
        statement.patch();
        this.normalizeAfterStatement(statement);
      }
      return;
    }
    let blockIndentLength = null;
    for (const statement of this.statements) {
      const indentLength = this.getIndentLength(statement);
      if (indentLength !== null) {
        if (blockIndentLength === null) {
          blockIndentLength = indentLength;
        } else {
          const charsToRemove = indentLength - blockIndentLength;
          if (charsToRemove < 0) {
            throw this.error("Unexpected statement at an earlier indentation level than an earlier statement in the block.");
          }
          if (charsToRemove > 0) {
            this.removePrecedingSpaceChars(statement.outerStart, charsToRemove);
          }
        }
      }
      statement.patch();
      this.normalizeAfterStatement(statement);
    }
  }
  removePrecedingSpaceChars(index, numToRemove) {
    let numRemaining = numToRemove;
    for (let i = index; numRemaining > 0 && i > 0; i--) {
      const contents = this.slice(i - 1, i);
      if (contents.includes("\n")) {
        throw this.error("Found start of line before removing enough indentation.");
      }
      if (contents === " " || contents === "	") {
        this.remove(i - 1, i);
        numRemaining -= 1;
      }
    }
  }
  getIndentLength(statement) {
    const startOfLine = getStartOfLine(this.context.source, statement.outerStart);
    const indentText = this.context.source.slice(startOfLine, statement.outerStart);
    if (/^[ \t]*$/.test(indentText)) {
      return indentText.length;
    } else {
      return null;
    }
  }
  normalizeAfterStatement(statement) {
    const followingComma = statement.nextSemanticToken();
    if (!followingComma || followingComma.type !== SourceType46.COMMA || followingComma.start >= this.contentEnd) {
      return;
    }
    this.overwrite(followingComma.start, followingComma.end, ";");
  }
};

// src/stages/normalize/patchers/FunctionPatcher.ts
import { SourceType as SourceType47 } from "coffee-lex";
import { Constructor } from "decaffeinate-parser";

// src/stages/normalize/patchers/DoOpPatcher.ts
var DoOpPatcher2 = class extends PassthroughPatcher {
};

// src/stages/normalize/patchers/DefaultParamPatcher.ts
var DefaultParamPatcher2 = class extends PassthroughPatcher {
  constructor(patcherContext, param, value) {
    super(patcherContext, param, value);
    this.param = param;
    this.value = value;
  }
  patch() {
    super.patch();
    if (this.shouldExtractToConditionalAssign()) {
      const callback = this.findAddDefaultParamAssignmentCallback();
      if (callback) {
        const paramCode = this.slice(this.param.contentStart, this.param.contentEnd);
        const valueCode = this.slice(this.value.contentStart, this.value.contentEnd);
        const newParamCode = callback(paramCode, valueCode, this.param.node);
        this.overwrite(this.param.contentStart, this.param.contentEnd, newParamCode);
        this.remove(this.param.outerEnd, this.value.outerEnd);
      }
    }
  }
  findAddDefaultParamAssignmentCallback() {
    let patcher = this;
    while (patcher) {
      if (patcher.addDefaultParamAssignmentAtScopeHeader) {
        return patcher.addDefaultParamAssignmentAtScopeHeader;
      }
      if (patcher.parent instanceof DefaultParamPatcher2 && patcher.parent.value === patcher) {
        break;
      }
      patcher = patcher.parent;
    }
    return null;
  }
  shouldExtractToConditionalAssign() {
    if (this.options.looseDefaultParams) {
      return false;
    }
    if (this.value.node.type === "Null") {
      return false;
    }
    if (this.parent instanceof FunctionPatcher2 && this.parent.parent instanceof DoOpPatcher2) {
      return false;
    }
    if (this.parent instanceof FunctionPatcher2 && this.parent.parent instanceof AssignOpPatcher2 && this.parent.parent.parent instanceof DoOpPatcher2) {
      return false;
    }
    return true;
  }
};

// src/stages/normalize/patchers/ExpansionPatcher.ts
var ExpansionPatcher2 = class extends PassthroughPatcher {
};

// src/stages/normalize/patchers/IdentifierPatcher.ts
var IdentifierPatcher2 = class extends PassthroughPatcher {
};

// src/stages/normalize/patchers/SpreadPatcher.ts
var SpreadPatcher2 = class extends PassthroughPatcher {
  constructor(patcherContext, expression) {
    super(patcherContext, expression);
    this.expression = expression;
  }
};

// src/stages/normalize/patchers/FunctionPatcher.ts
var FunctionPatcher2 = class extends NodePatcher {
  constructor(patcherContext, parameters, body) {
    super(patcherContext);
    this._thisAfterSuper = false;
    this.parameters = parameters;
    this.body = body;
  }
  patchAsExpression() {
    if (this.body && !this.slice(this.body.contentStart - 1, this.body.contentStart).match(/\s/)) {
      this.insert(this.body.contentStart, " ");
    }
    const neededExplicitBindings = [];
    const firstRestParamIndex = this.getFirstRestParamIndex();
    const assignments = [];
    const thisAssignmentsAfterSuper = [];
    for (const [i, parameter] of this.parameters.entries()) {
      if (firstRestParamIndex === -1 || i < firstRestParamIndex) {
        const { newAssignments, newThisAssignments, newBindings } = this.patchParameterAndGetAssignments(parameter);
        assignments.push(...newAssignments);
        thisAssignmentsAfterSuper.push(...newThisAssignments);
        neededExplicitBindings.push(...newBindings);
      } else {
        parameter.patch();
      }
      normalizeListItem(this, parameter, this.parameters[i + 1]);
      if (i === this.parameters.length - 1) {
        const nextToken = parameter.nextSemanticToken();
        if (nextToken && nextToken.type === SourceType47.SEMICOLON) {
          this.remove(nextToken.start, nextToken.end);
        }
      }
    }
    if (firstRestParamIndex !== -1) {
      if (firstRestParamIndex === this.parameters.length - 1 && this.parameters[this.parameters.length - 1] instanceof ExpansionPatcher2) {
        if (firstRestParamIndex === 0) {
          this.remove(this.parameters[0].contentStart, this.parameters[0].contentEnd);
        } else {
          this.remove(this.parameters[firstRestParamIndex - 1].outerEnd, this.parameters[this.parameters.length - 1].outerEnd);
        }
      } else {
        const candidateName = firstRestParamIndex === 0 ? "args" : "rest";
        const paramName = this.claimFreeBinding(candidateName);
        const restParamsStart = this.parameters[firstRestParamIndex].contentStart;
        const restParamsEnd = this.parameters[this.parameters.length - 1].contentEnd;
        let paramCode = this.slice(restParamsStart, restParamsEnd);
        paramCode = this.fixGeneratedAssigneeWhitespace(paramCode);
        this.overwrite(restParamsStart, restParamsEnd, `${paramName}...`);
        assignments.push(`[${paramCode}] = ${paramName}`);
        for (let i = firstRestParamIndex; i < this.parameters.length; i++) {
          neededExplicitBindings.push(...getAssigneeBindings(this.parameters[i].node));
        }
      }
    }
    let uniqueExplicitBindings = [...new Set(neededExplicitBindings)];
    uniqueExplicitBindings = uniqueExplicitBindings.filter((name) => notNull(this.parent).getScope().hasBinding(name));
    if (uniqueExplicitBindings.length > 0) {
      assignments.unshift(`\`var ${uniqueExplicitBindings.join(", ")};\``);
    }
    if (this.body) {
      if (thisAssignmentsAfterSuper.length > 0) {
        this.body.insertStatementsAtIndex(thisAssignmentsAfterSuper, this.getIndexOfSuperStatement() + 1);
      }
      for (const assignment of assignments) {
        this.body.insertLineBefore(assignment);
      }
      this.body.patch();
    } else if (assignments.length) {
      if (!(this.context.getParent(this.node) instanceof Constructor)) {
        assignments.push("return");
      }
      const indent = this.getIndent(1);
      const text = assignments.join(`
${indent}`);
      this.insert(this.contentEnd, `
${indent}${text}`);
    }
  }
  patchParameterAndGetAssignments(parameter) {
    const thisAssignments = [];
    const defaultParamAssignments = [];
    const newBindings = [];
    this.addThisAssignmentAtScopeHeader = (memberName) => {
      const varName = this.claimFreeBinding(memberName);
      thisAssignments.push(`@${memberName} = ${varName}`);
      this.log(`Replacing parameter @${memberName} with ${varName}`);
      return varName;
    };
    this.addDefaultParamAssignmentAtScopeHeader = (assigneeCode, initCode, assigneeNode) => {
      if (assigneeNode.type === "Identifier" || assigneeNode.type === "MemberAccessOp") {
        defaultParamAssignments.push(`(${assigneeCode} ?= ${initCode})`);
        return assigneeCode;
      } else {
        assigneeCode = this.fixGeneratedAssigneeWhitespace(assigneeCode);
        const paramName = this.claimFreeBinding("param");
        defaultParamAssignments.push(`(${paramName} ?= ${initCode})`);
        defaultParamAssignments.push(`${assigneeCode} = ${paramName}`);
        newBindings.push(...getAssigneeBindings(assigneeNode));
        return paramName;
      }
    };
    parameter.patch();
    delete this.addDefaultParamAssignmentAtScopeHeader;
    delete this.addThisAssignmentAtScopeHeader;
    if (this.options.useCS2 && this.thisAfterSuperEnabled() && this.getIndexOfSuperStatement() >= 0) {
      return {
        newAssignments: defaultParamAssignments,
        newThisAssignments: thisAssignments,
        newBindings
      };
    } else {
      return {
        newAssignments: [...defaultParamAssignments, ...thisAssignments],
        newThisAssignments: [],
        newBindings
      };
    }
  }
  fixGeneratedAssigneeWhitespace(assigneeCode) {
    const firstNewlineIndex = assigneeCode.indexOf("\n");
    if (firstNewlineIndex < 0) {
      return assigneeCode;
    }
    const indent = this.body ? this.body.getIndent(0) : this.getIndent(1);
    const firstLine = assigneeCode.substr(0, firstNewlineIndex);
    let otherLines = assigneeCode.substr(firstNewlineIndex + 1);
    otherLines = stripSharedIndent(otherLines);
    otherLines = otherLines.replace(/\n/g, `
${indent}`);
    return `${firstLine}
${indent}${otherLines}`;
  }
  getFirstRestParamIndex() {
    if (this.parameters.some((param, i) => i < this.parameters.length - 1 && param instanceof ExpansionPatcher2)) {
      return 0;
    }
    for (let i = 0; i < this.parameters.length; i++) {
      const parameter = this.parameters[i];
      if (parameter instanceof DefaultParamPatcher2 && canPatchAssigneeToJavaScript(parameter.param.node, this.options)) {
        continue;
      }
      if (i === this.parameters.length - 1 && parameter instanceof SpreadPatcher2 && parameter.expression instanceof IdentifierPatcher2) {
        continue;
      }
      if (!this.options.useCS2 && parameter instanceof ArrayInitialiserPatcher2 || !canPatchAssigneeToJavaScript(parameter.node, this.options)) {
        return i;
      }
    }
    return -1;
  }
  getIndexOfSuperStatement() {
    if (!this.body) {
      return -1;
    }
    const statements = this.body.statements;
    for (let i = 0; i < statements.length; i++) {
      if (containsSuperCall(statements[i].node)) {
        return i;
      }
    }
    return -1;
  }
  enableThisAfterSuper() {
    this._thisAfterSuper = true;
  }
  thisAfterSuperEnabled() {
    return this._thisAfterSuper;
  }
};

// src/stages/normalize/patchers/ConstructorPatcher.ts
var ConstructorPatcher2 = class extends PassthroughPatcher {
  constructor(patcherContext, assignee, expression) {
    super(patcherContext, assignee, expression);
    this.assignee = assignee;
    this.expression = expression;
    if (expression instanceof FunctionPatcher2) {
      expression.enableThisAfterSuper();
    }
  }
};

// src/stages/normalize/patchers/ProgramPatcher.ts
var ProgramPatcher3 = class extends ProgramPatcher {
  patchAsStatement() {
    if (this.body) {
      this.body.patch();
    }
    this.patchHelpers();
  }
};

// src/stages/normalize/patchers/ClassPatcher.ts
import {
  AssignOp as AssignOp6,
  BaseFunction as BaseFunction2,
  ClassProtoAssignOp,
  DynamicMemberAccessOp as DynamicMemberAccessOp3,
  Identifier as Identifier9,
  MemberAccessOp as MemberAccessOp5,
  This as This2
} from "decaffeinate-parser";
var ClassPatcher2 = class extends NodePatcher {
  constructor(patcherContext, nameAssignee, parent, body) {
    super(patcherContext);
    this.nameAssignee = nameAssignee;
    this.superclass = parent;
    this.body = body;
  }
  patchAsStatement() {
    if (this.needsIndent()) {
      this.indent(1, { skipFirstLine: true });
    }
    this.removeThenTokenIfNecessary();
    const indent = this.getIndent();
    if (this.nameAssignee) {
      this.nameAssignee.patch();
    }
    if (this.superclass) {
      this.superclass.patch();
    }
    if (this.body) {
      this.body.patch();
    }
    if (!this.needsInitClass()) {
      return;
    }
    this.addSuggestion(AVOID_INITCLASS);
    const insertPoint = this.getInitClassInsertPoint();
    const nonMethodPatchers = this.getNonMethodPatchers();
    const customConstructorInfo = this.extractCustomConstructorInfo();
    const shouldUseIIFE = this.shouldUseIIFE();
    if (shouldUseIIFE) {
      if (this.nameAssignee instanceof IdentifierPatcher2) {
        this.insert(this.outerStart, `${this.nameAssignee.node.data} = `);
      }
      this.insert(this.outerStart, `do ->
${indent}`);
    }
    let needsTmpName = false;
    let classRef;
    if (this.nameAssignee instanceof IdentifierPatcher2) {
      classRef = this.nameAssignee.node.data;
    } else {
      classRef = this.claimFreeBinding("Cls");
      needsTmpName = true;
    }
    const assignmentNames = this.generateInitClassMethod(nonMethodPatchers, customConstructorInfo, insertPoint);
    this.insert(this.outerEnd, `
${indent}${classRef}.initClass()`);
    if (shouldUseIIFE) {
      this.insert(this.outerEnd, `
${indent}return ${classRef}`);
    }
    for (const assignmentName of assignmentNames) {
      this.insert(this.outerStart, `${assignmentName} = undefined
${indent}`);
    }
    if (needsTmpName) {
      this.insert(this.outerStart, `${classRef} = `);
    }
  }
  patchAsExpression() {
    if (this.body) {
      this.body.patch();
    }
  }
  needsIndent() {
    return this.needsInitClass() && this.shouldUseIIFE();
  }
  needsInitClass() {
    if (!this.body) {
      return false;
    }
    if (this.body.statements.length === 0) {
      return false;
    }
    const nonMethodPatchers = this.getNonMethodPatchers();
    if (nonMethodPatchers.length === 0 && !this.needsCustomConstructor()) {
      return false;
    }
    return true;
  }
  removeThenTokenIfNecessary() {
    let searchStart;
    if (this.superclass) {
      searchStart = this.superclass.outerEnd;
    } else if (this.nameAssignee) {
      searchStart = this.nameAssignee.outerEnd;
    } else {
      searchStart = this.firstToken().end;
    }
    let searchEnd;
    if (this.body) {
      searchEnd = this.body.outerStart;
    } else {
      searchEnd = this.contentEnd;
    }
    const index = this.indexOfSourceTokenBetweenSourceIndicesMatching(searchStart, searchEnd, (token) => token.type === SourceType48.THEN);
    if (index) {
      this.overwrite(searchStart, searchEnd, `
${this.getIndent(1)}`);
    }
  }
  shouldUseIIFE() {
    const nonMethodPatchers = this.getNonMethodPatchers();
    if (this.hasAnyAssignments(nonMethodPatchers)) {
      return true;
    }
    if (this.parent instanceof BlockPatcher2) {
      const { statements } = this.parent;
      if (!(this.parent.parent instanceof ProgramPatcher3) && this === statements[statements.length - 1]) {
        return true;
      }
      return false;
    }
    return true;
  }
  getInitClassInsertPoint() {
    if (this.superclass) {
      return this.superclass.outerEnd;
    }
    if (this.nameAssignee) {
      return this.nameAssignee.outerEnd;
    }
    return this.firstToken().end;
  }
  getNonMethodPatchers() {
    if (!this.body) {
      throw this.error("Expected non-null body.");
    }
    const nonMethodPatchers = [];
    let deleteStart = this.getInitClassInsertPoint();
    for (const patcher of this.body.statements) {
      if (!this.isClassMethod(patcher)) {
        nonMethodPatchers.push({
          patcher,
          deleteStart
        });
      }
      deleteStart = patcher.outerEnd;
    }
    return nonMethodPatchers;
  }
  isClassMethod(patcher) {
    if (patcher instanceof ConstructorPatcher2) {
      return true;
    }
    const node = patcher.node;
    if (this.isClassAssignment(node)) {
      if (node instanceof AssignOp6 && ["BoundFunction", "BoundGeneratorFunction"].indexOf(node.expression.type) >= 0) {
        return false;
      }
      if (node.expression instanceof BaseFunction2) {
        return true;
      }
    }
    return false;
  }
  isClassAssignment(node) {
    if (node instanceof ClassProtoAssignOp) {
      return true;
    }
    if (node instanceof AssignOp6) {
      const { assignee } = node;
      if (assignee instanceof MemberAccessOp5 || assignee instanceof DynamicMemberAccessOp3) {
        if (assignee.expression instanceof This2) {
          return true;
        }
        if (this.nameAssignee && this.nameAssignee instanceof IdentifierPatcher2) {
          const className = this.nameAssignee.node.data;
          if (assignee.expression instanceof Identifier9 && assignee.expression.data === className) {
            return true;
          }
        }
      }
    }
    return false;
  }
  needsCustomConstructor() {
    if (!this.body) {
      throw this.error("Expected non-null body.");
    }
    for (const patcher of this.body.statements) {
      if (patcher instanceof ConstructorPatcher2 && !(patcher.expression instanceof FunctionPatcher2)) {
        return true;
      }
    }
    return false;
  }
  extractCustomConstructorInfo() {
    if (!this.body) {
      throw this.error("Expected non-null body.");
    }
    for (const patcher of this.body.statements) {
      if (patcher instanceof ConstructorPatcher2) {
        if (!(patcher.expression instanceof FunctionPatcher2)) {
          const expressionCode = this.slice(patcher.expression.contentStart, patcher.expression.contentEnd);
          let ctorName;
          if (this.nameAssignee instanceof IdentifierPatcher2) {
            const className = this.nameAssignee.node.data;
            ctorName = this.claimFreeBinding(`create${className}`);
          } else {
            ctorName = this.claimFreeBinding("createInstance");
          }
          const bodyIndent = this.getBodyIndent();
          const indentString = this.getProgramIndentString();
          this.overwrite(patcher.expression.outerStart, patcher.expression.outerEnd, `->
${bodyIndent}${indentString}return ${ctorName}.apply(this, arguments)`);
          return {
            ctorName,
            expressionCode
          };
        }
      }
    }
    return null;
  }
  generateInitClassMethod(nonMethodPatchers, customConstructorInfo, insertPoint) {
    const bodyIndent = this.getBodyIndent();
    const indentString = this.getProgramIndentString();
    this.insert(insertPoint, `
${bodyIndent}@initClass: ->`);
    const assignmentNames = [];
    for (const { patcher, deleteStart } of nonMethodPatchers) {
      const assignmentName = this.getAssignmentName(patcher);
      if (assignmentName) {
        assignmentNames.push(assignmentName);
      }
      let statementCode = this.slice(deleteStart, patcher.outerEnd);
      statementCode = statementCode.replace(/\n/g, `
${indentString}`);
      this.insert(insertPoint, statementCode);
      this.remove(deleteStart, patcher.outerEnd);
    }
    if (customConstructorInfo) {
      const { ctorName, expressionCode } = customConstructorInfo;
      this.insert(insertPoint, `
${bodyIndent}${indentString}${ctorName} = ${expressionCode}`);
      assignmentNames.push(ctorName);
    }
    this.insert(insertPoint, `
${bodyIndent}${indentString}return`);
    return assignmentNames;
  }
  hasAnyAssignments(nonMethodPatchers) {
    for (const { patcher } of nonMethodPatchers) {
      if (this.getAssignmentName(patcher)) {
        return true;
      }
    }
    return false;
  }
  getBodyIndent() {
    if (!this.body) {
      throw this.error("Expected non-null body.");
    }
    const bodyNodeIndent = this.body.getIndent();
    if (bodyNodeIndent === this.getIndent()) {
      return this.getIndent(1);
    } else {
      return bodyNodeIndent;
    }
  }
  getAssignmentName(statementPatcher) {
    if (statementPatcher.node instanceof AssignOp6 && statementPatcher.node.assignee instanceof Identifier9) {
      return statementPatcher.node.assignee.data;
    }
    if (statementPatcher instanceof ClassPatcher2 && statementPatcher.nameAssignee instanceof IdentifierPatcher2) {
      return statementPatcher.nameAssignee.node.data;
    }
    return null;
  }
};

// src/stages/normalize/patchers/DynamicMemberAccessOpPatcher.ts
var DynamicMemberAccessOpPatcher2 = class extends PassthroughPatcher {
  constructor(patcherContext, expression, indexingExpr) {
    super(patcherContext, expression, indexingExpr);
    this.expression = expression;
    this.indexingExpr = indexingExpr;
  }
};

// src/stages/normalize/patchers/MemberAccessOpPatcher.ts
var MemberAccessOpPatcher2 = class extends PassthroughPatcher {
  constructor(patcherContext, expression, member) {
    super(patcherContext, expression, member);
    this.expression = expression;
    this.member = member;
  }
  shouldTrimContentRange() {
    return true;
  }
  patch() {
    super.patch();
    const callback = this.findAddThisAssignmentCallback();
    if (callback) {
      this.overwrite(this.contentStart, this.contentEnd, callback(this.node.member.data));
    }
  }
  findAddThisAssignmentCallback() {
    let patcher = this;
    while (patcher) {
      if (patcher.addThisAssignmentAtScopeHeader) {
        return patcher.addThisAssignmentAtScopeHeader;
      }
      if (patcher.parent instanceof DefaultParamPatcher2 && patcher.parent.value === patcher) {
        break;
      }
      patcher = patcher.parent;
    }
    return null;
  }
};

// src/stages/normalize/patchers/AssignOpPatcher.ts
var AssignOpPatcher2 = class extends NodePatcher {
  constructor(patcherContext, assignee, expression) {
    super(patcherContext);
    this.assignee = assignee;
    this.expression = expression;
  }
  patchAsExpression() {
    this.prepareEarlySuperTransform();
    const isDynamicallyCreatedClassAssignment = this.isDynamicallyCreatedClassAssignment();
    if (isDynamicallyCreatedClassAssignment) {
      this.patchClassAssignmentPrefix();
    }
    this.assignee.patch();
    if (isDynamicallyCreatedClassAssignment) {
      this.patchClassAssignmentOperator();
    }
    this.removeUnnecessaryThenToken();
    this.expression.patch();
  }
  isDynamicallyCreatedClassAssignment() {
    const classParent = this.getClassParent();
    return classParent !== null && classParent.isClassAssignment(this.node) && !(classParent.isClassMethod(this) && notNull(classParent.body).statements.indexOf(this) > -1);
  }
  patchClassAssignmentPrefix() {
    if (this.node.type === "ClassProtoAssignOp") {
      this.insert(this.assignee.outerStart, "@prototype.");
    }
  }
  patchClassAssignmentOperator() {
    const colonIndex = this.indexOfSourceTokenBetweenPatchersMatching(this.assignee, this.expression, (token) => token.type === SourceType49.COLON);
    if (colonIndex) {
      const colonToken = notNull(this.sourceTokenAtIndex(colonIndex));
      this.overwrite(colonToken.start, colonToken.end, " =");
    }
  }
  getClassParent() {
    let parent = this;
    while (parent) {
      if (parent instanceof FunctionPatcher2) {
        return null;
      } else if (parent instanceof ClassPatcher2) {
        return parent;
      }
      parent = parent.parent;
    }
    return null;
  }
  needsEarlySuperTransform() {
    if (!this.isDynamicallyCreatedClassAssignment()) {
      return false;
    }
    return this.node.type !== "ClassProtoAssignOp" && this.expression instanceof FunctionPatcher2 && containsSuperCall(this.expression.node);
  }
  prepareEarlySuperTransform() {
    if (this.needsEarlySuperTransform()) {
      if (this.assignee instanceof MemberAccessOpPatcher2) {
        this.assignee.expression.setRequiresRepeatableExpression({
          parens: true,
          ref: "cls",
          forceRepeat: true
        });
      } else if (this.assignee instanceof DynamicMemberAccessOpPatcher2) {
        this.assignee.expression.setRequiresRepeatableExpression({
          parens: true,
          ref: "cls",
          forceRepeat: true
        });
        this.assignee.indexingExpr.setRequiresRepeatableExpression({
          ref: "method",
          forceRepeat: true
        });
      } else {
        throw this.error("Unexpected assignee type for early super transform.");
      }
    }
  }
  getEarlySuperTransformInfo() {
    if (this.needsEarlySuperTransform()) {
      if (this.assignee instanceof MemberAccessOpPatcher2) {
        return {
          classCode: this.assignee.expression.getRepeatCode(),
          accessCode: `.${this.assignee.member.node.data}`
        };
      } else if (this.assignee instanceof DynamicMemberAccessOpPatcher2) {
        return {
          classCode: this.assignee.expression.getRepeatCode(),
          accessCode: `[${this.assignee.indexingExpr.getRepeatCode()}]`
        };
      } else {
        throw this.error("Unexpected assignee type for early super transform.");
      }
    }
    return null;
  }
  removeUnnecessaryThenToken() {
    const thenIndex = this.indexOfSourceTokenBetweenPatchersMatching(this.assignee, this.expression, (token) => token.type === SourceType49.THEN);
    if (thenIndex) {
      const thenToken = notNull(this.sourceTokenAtIndex(thenIndex));
      if (this.slice(thenToken.start - 1, thenToken.start) === " ") {
        this.remove(thenToken.start - 1, thenToken.end);
      } else {
        this.remove(thenToken.start, thenToken.end);
      }
    }
  }
};

// src/stages/normalize/patchers/ConditionalPatcher.ts
import { SourceType as SourceType52 } from "coffee-lex";

// src/utils/postfixExpressionRequiresParens.ts
function postfixExpressionRequiresParens(exprCode) {
  return exprCode.includes("then") || exprCode.includes("\n");
}

// src/utils/postfixNodeNeedsOuterParens.ts
import { SourceType as SourceType51 } from "coffee-lex";
import { Spread as Spread4 } from "decaffeinate-parser";

// src/stages/normalize/patchers/ObjectInitialiserPatcher.ts
import { SourceType as SourceType50 } from "coffee-lex";
var ObjectInitialiserPatcher2 = class extends NodePatcher {
  constructor(patcherContext, members) {
    super(patcherContext);
    this.members = members;
  }
  patchAsExpression() {
    for (const [i, member] of this.members.entries()) {
      member.patch();
      normalizeListItem(this, member, this.members[i + 1]);
    }
  }
  isImplicitObjectInitializer() {
    return this.firstToken().type !== SourceType50.LBRACE;
  }
};

// src/utils/postfixNodeNeedsOuterParens.ts
function postfixNodeNeedsOuterParens(patcher) {
  const parent = patcher.parent;
  if (parent) {
    if (parent.node instanceof Spread4) {
      return true;
    }
    const grandparent = parent.parent;
    if (grandparent && grandparent instanceof ObjectInitialiserPatcher2 && grandparent.isImplicitObjectInitializer()) {
      return true;
    }
  }
  const nextToken = patcher.nextSemanticToken();
  if (nextToken) {
    return nextToken.type === SourceType51.COMMA || nextToken.type === SourceType51.SEMICOLON;
  }
  return false;
}

// src/stages/normalize/patchers/ConditionalPatcher.ts
var ConditionalPatcher2 = class extends NodePatcher {
  constructor(patcherContext, condition, consequent, alternate) {
    super(patcherContext);
    this.condition = condition;
    this.consequent = consequent;
    this.alternate = alternate;
  }
  patchAsExpression() {
    if (this.isPostIf()) {
      this.patchPostIf();
    } else {
      this.condition.patch();
      if (this.consequent !== null) {
        this.consequent.patch();
      }
      if (this.alternate !== null) {
        this.alternate.patch();
      }
    }
  }
  patchPostIf() {
    if (!this.consequent) {
      throw this.error("Expected non-null consequent for post-if.");
    }
    this.condition.patch();
    if (postfixExpressionRequiresParens(this.slice(this.condition.contentStart, this.condition.contentEnd)) && !this.condition.isSurroundedByParentheses()) {
      this.condition.surroundInParens();
    }
    const ifTokenIndex = this.getIfTokenIndex();
    const ifToken = this.sourceTokenAtIndex(ifTokenIndex);
    if (!ifToken) {
      throw this.error("Unable to find `if` token.");
    }
    const needsParens = postfixNodeNeedsOuterParens(this);
    const ifAndConditionCode = this.slice(ifToken.start, this.condition.outerEnd);
    if (needsParens) {
      this.insert(this.consequent.outerStart, "(");
    }
    this.insert(this.consequent.outerStart, `${ifAndConditionCode} then `);
    this.consequent.patch();
    if (needsParens) {
      this.insert(this.consequent.outerEnd, ")");
    }
    this.remove(this.consequent.outerEnd, this.contentEnd);
  }
  isPostIf() {
    return this.consequent !== null && this.condition.contentStart > this.consequent.contentStart;
  }
  getIfTokenIndex() {
    const start = this.contentStartTokenIndex;
    let index = this.condition.outerStartTokenIndex;
    while (index !== start) {
      const token = this.sourceTokenAtIndex(index);
      if (token && token.type === SourceType52.IF) {
        break;
      }
      index = notNull(index.previous());
    }
    if (!index) {
      throw this.error("unable to find `if` token in conditional");
    }
    return index;
  }
};

// src/stages/normalize/patchers/ForPatcher.ts
import { SourceType as SourceType53 } from "coffee-lex";
var ForPatcher2 = class extends NodePatcher {
  constructor(patcherContext, keyAssignee, valAssignee, target, filter, body) {
    super(patcherContext);
    this.keyAssignee = keyAssignee;
    this.valAssignee = valAssignee;
    this.target = target;
    this.filter = filter;
    this.body = body;
  }
  patchAsExpression() {
    let bodyPrefixLine = null;
    if (this.keyAssignee) {
      this.keyAssignee.patch();
    }
    if (this.valAssignee) {
      bodyPrefixLine = this.patchValAssignee();
    }
    this.target.patch();
    if (this.filter) {
      this.filter.patch();
    }
    if (this.isPostFor()) {
      this.surroundThenUsagesInParens();
      const forToken = this.getForToken();
      const forThroughEnd = this.slice(forToken.start, this.contentEnd);
      const needsParens = postfixNodeNeedsOuterParens(this);
      this.remove(this.body.outerEnd, this.contentEnd);
      if (needsParens) {
        this.insert(this.body.outerStart, "(");
      }
      this.insert(this.body.outerStart, `${forThroughEnd} then `);
      if (needsParens) {
        this.insert(this.contentEnd, ")");
      }
    }
    if (bodyPrefixLine !== null) {
      if (this.body) {
        this.body.insertLineBefore(bodyPrefixLine);
      } else {
        this.insert(this.contentEnd, ` ${bodyPrefixLine}`);
      }
    }
    if (this.body) {
      this.body.patch();
    }
  }
  patchAsStatement() {
    this.patchAsExpression();
  }
  patchValAssignee() {
    if (!this.valAssignee) {
      throw this.error("Expected to find a valAssignee.");
    }
    if (canPatchAssigneeToJavaScript(this.valAssignee.node, this.options)) {
      this.valAssignee.patch();
      return null;
    } else {
      const assigneeName = this.claimFreeBinding("value");
      const assigneeCode = this.valAssignee.patchAndGetCode();
      this.overwrite(this.valAssignee.contentStart, this.valAssignee.contentEnd, assigneeName);
      return `${assigneeCode} = ${assigneeName}`;
    }
  }
  isPostFor() {
    return this.body && this.body.contentStart < this.target.contentStart;
  }
  surroundThenUsagesInParens() {
    if (postfixExpressionRequiresParens(this.slice(this.target.contentStart, this.target.contentEnd))) {
      this.target.surroundInParens();
    }
    if (this.filter && postfixExpressionRequiresParens(this.slice(this.filter.contentStart, this.filter.contentEnd))) {
      this.filter.surroundInParens();
    }
  }
  getForToken() {
    if (this.isPostFor()) {
      const afterForToken = this.getFirstHeaderPatcher();
      const index = this.indexOfSourceTokenBetweenPatchersMatching(this.body, afterForToken, (token) => token.type === SourceType53.FOR);
      if (!index) {
        throw this.error(`cannot find 'for' token in loop`);
      }
      return notNull(this.sourceTokenAtIndex(index));
    } else {
      const token = this.sourceTokenAtIndex(this.contentStartTokenIndex);
      if (!token || token.type !== SourceType53.FOR) {
        throw this.error(`expected 'for' at start of loop`);
      }
      return token;
    }
  }
  getFirstHeaderPatcher() {
    const candidates = [this.keyAssignee, this.valAssignee, this.target];
    let result = null;
    candidates.forEach((candidate) => {
      if (!candidate) {
        return;
      }
      if (result === null || candidate.contentStart < result.contentStart) {
        result = candidate;
      }
    });
    if (result === null) {
      throw this.error(`cannot get first patcher of 'for' loop header`);
    }
    return result;
  }
};

// src/stages/normalize/patchers/ForFromPatcher.ts
var ForFromPatcher2 = class extends ForPatcher2 {
};

// src/stages/normalize/patchers/ForInPatcher.ts
var ForInPatcher2 = class extends ForPatcher2 {
  constructor(patcherContext, keyAssignee, valAssignee, target, step, filter, body) {
    super(patcherContext, keyAssignee, valAssignee, target, filter, body);
    this.step = step;
  }
  patchAsExpression() {
    if (this.step) {
      this.step.patch();
    }
    super.patchAsExpression();
  }
  surroundThenUsagesInParens() {
    if (this.step && postfixExpressionRequiresParens(this.slice(this.step.contentStart, this.step.contentEnd))) {
      this.step.surroundInParens();
    }
    super.surroundThenUsagesInParens();
  }
};

// src/stages/normalize/patchers/ForOfPatcher.ts
var ForOfPatcher2 = class extends ForPatcher2 {
};

// src/stages/normalize/patchers/FunctionApplicationPatcher.ts
import { SourceType as SourceType54 } from "coffee-lex";
import { CSXElement } from "decaffeinate-parser";
var FunctionApplicationPatcher2 = class extends NodePatcher {
  constructor(patcherContext, fn, args) {
    super(patcherContext);
    this.fn = fn;
    this.args = args;
  }
  patchAsExpression() {
    const implicitCall = this.isImplicitCall();
    const { args } = this;
    this.fn.patch();
    if (implicitCall) {
      const firstArg = args[0];
      const firstArgIsOnNextLine = !firstArg ? false : /\n/.test(this.context.source.slice(this.fn.outerEnd, firstArg.outerStart));
      const funcEnd = this.getFuncEnd();
      if (firstArgIsOnNextLine) {
        this.insert(funcEnd, "(");
      } else {
        this.overwrite(funcEnd, firstArg.outerStart, "(");
      }
    }
    for (const [i, arg] of args.entries()) {
      arg.patch();
      normalizeListItem(this, arg, args[i + 1]);
    }
    if (implicitCall) {
      this.insertImplicitCloseParen();
    }
  }
  insertImplicitCloseParen() {
    if (this.fn.node instanceof CSXElement && !this.fn.isSurroundedByParentheses()) {
      this.fn.surroundInParens();
    }
    const argListCode = this.slice(this.args[0].contentStart, this.args[this.args.length - 1].contentEnd);
    const isArgListMultiline = argListCode.indexOf("\n") !== -1;
    const lastTokenType = this.lastToken().type;
    if (!isArgListMultiline || lastTokenType === SourceType54.RBRACE || lastTokenType === SourceType54.RBRACKET) {
      this.insert(this.contentEnd, ")");
      return;
    }
    const followingCloseParen = this.getFollowingCloseParenIfExists();
    if (followingCloseParen) {
      this.insert(Math.min(followingCloseParen.start, this.getMaxCloseParenInsertPoint()), ")");
      return;
    }
    const { args } = this;
    const lastArg = args[args.length - 1];
    if (lastArg.isMultiline()) {
      const nextSemanticToken = this.getFirstSemanticToken(this.contentEnd);
      if (nextSemanticToken && nextSemanticToken.type === SourceType54.DOT) {
        this.overwrite(this.outerEnd, nextSemanticToken.start, ")");
      } else {
        this.insert(this.contentEnd, `
${this.getIndent()})`);
      }
      return;
    }
    this.insert(this.contentEnd, ")");
  }
  getFollowingCloseParenIfExists() {
    let tokenIndex = this.contentEndTokenIndex;
    let token;
    do {
      const nextTokenIndex = tokenIndex.next();
      if (nextTokenIndex === null) {
        return null;
      }
      tokenIndex = nextTokenIndex;
      token = this.sourceTokenAtIndex(tokenIndex);
      if (token === null) {
        return null;
      }
    } while (token.type === SourceType54.NEWLINE);
    if (token.type === SourceType54.CALL_END || token.type === SourceType54.RPAREN) {
      return token;
    }
    return null;
  }
  getMaxCloseParenInsertPoint() {
    const maxInsertionPoint = this.getEditingBounds()[1];
    let enclosingIndentedPatcher = this;
    while (!enclosingIndentedPatcher.isFirstNodeInLine(enclosingIndentedPatcher.contentStart) && enclosingIndentedPatcher.parent) {
      enclosingIndentedPatcher = enclosingIndentedPatcher.parent;
    }
    return Math.min(maxInsertionPoint, enclosingIndentedPatcher.contentEnd);
  }
  isImplicitCall() {
    if (this.args.length === 0) {
      return false;
    }
    const searchStart = this.fn.outerEnd;
    const searchEnd = this.args[0].outerStart;
    return this.indexOfSourceTokenBetweenSourceIndicesMatching(searchStart, searchEnd, (token) => token.type === SourceType54.CALL_START) === null;
  }
  getFuncEnd() {
    if (this.node.type === "SoakedFunctionApplication" || this.node.type === "SoakedNewOp") {
      const questionMarkTokenIndex = this.indexOfSourceTokenAfterSourceTokenIndex(this.fn.outerEndTokenIndex, SourceType54.EXISTENCE);
      if (!questionMarkTokenIndex) {
        throw this.error("Expected to find question mark token index.");
      }
      const questionMarkToken = this.sourceTokenAtIndex(questionMarkTokenIndex);
      if (!questionMarkToken) {
        throw this.error("Expected to find question mark token.");
      }
      return questionMarkToken.end;
    } else {
      return this.fn.outerEnd;
    }
  }
};

// src/stages/normalize/patchers/LoopPatcher.ts
import { SourceType as SourceType55 } from "coffee-lex";
var LoopPatcher2 = class extends NodePatcher {
  constructor(patcherContext, body) {
    super(patcherContext);
    this.body = body;
  }
  patchAsExpression() {
    const loop = this.firstToken();
    const next = this.sourceTokenAtIndex(notNull(this.contentStartTokenIndex.next()));
    if (!next) {
      throw this.error("Expected to find a next token.");
    }
    if (loop.type !== SourceType55.LOOP) {
      throw this.error(`expected first token of loop to be LOOP, but got: ${SourceType55[loop.type]}`);
    }
    if (next.type === SourceType55.THEN || !this.body.node.inline) {
      this.overwrite(loop.start, loop.end, "while true");
    } else {
      this.overwrite(loop.start, loop.end, "while true then");
    }
    if (this.body) {
      this.body.patch();
    }
  }
};

// src/stages/normalize/patchers/ObjectInitialiserMemberPatcher.ts
var ObjectInitialiserMemberPatcher2 = class extends PassthroughPatcher {
  constructor(patcherContext, key, expression) {
    super(patcherContext, key, expression);
    this.key = key;
    this.expression = expression;
  }
};

// src/stages/normalize/patchers/ProtoMemberAccessOpPatcher.ts
import { SourceType as SourceType56 } from "coffee-lex";
var ProtoMemberAccessOpPatcher = class extends NodePatcher {
  constructor(patcherContext, expression) {
    super(patcherContext);
    this.expression = expression;
  }
  patchAsExpression() {
    this.expression.patch();
    const protoToken = this.getProtoToken();
    if (this.parent instanceof MemberAccessOpPatcher2) {
      this.overwrite(protoToken.start, protoToken.end, ".prototype.");
    } else {
      this.overwrite(protoToken.start, protoToken.end, ".prototype");
    }
  }
  getProtoToken() {
    const protoIndex = this.indexOfSourceTokenBetweenSourceIndicesMatching(this.expression.outerEnd, this.contentEnd, (token) => token.type === SourceType56.PROTO);
    if (protoIndex) {
      const protoToken = this.sourceTokenAtIndex(protoIndex);
      if (protoToken) {
        return protoToken;
      }
    }
    throw this.error(`unable to find '::' token after proto member access`);
  }
};

// src/stages/normalize/patchers/SuperPatcher.ts
import { SourceType as SourceType57 } from "coffee-lex";
var SuperPatcher2 = class extends NodePatcher {
  patchAsExpression() {
    const earlyTransformInfo = this.getEarlyTransformInfo();
    if (earlyTransformInfo) {
      this.patchEarlySuperTransform(earlyTransformInfo);
    } else if (this.node.type === "BareSuperFunctionApplication") {
      this.insert(this.contentEnd, "(arguments...)");
    }
  }
  patchEarlySuperTransform({ classCode, accessCode }) {
    const replacement = `${classCode}.__proto__${accessCode}.call(this, `;
    if (this.node.type === "BareSuperFunctionApplication") {
      this.overwrite(this.contentStart, this.contentEnd, `${replacement}arguments...)`);
    } else {
      const followingOpenParen = this.getFollowingOpenParenToken();
      this.overwrite(this.contentStart, followingOpenParen.end, replacement);
    }
  }
  getEarlyTransformInfo() {
    let parent = this.parent;
    while (parent) {
      if (parent instanceof AssignOpPatcher2) {
        const earlyTransformInfo = parent.getEarlySuperTransformInfo();
        if (earlyTransformInfo) {
          return earlyTransformInfo;
        }
      } else if (parent instanceof ClassPatcher2) {
        return null;
      }
      parent = parent.parent;
    }
    return null;
  }
  getFollowingOpenParenToken() {
    const openParenTokenIndex = this.indexOfSourceTokenAfterSourceTokenIndex(this.contentEndTokenIndex, SourceType57.CALL_START);
    if (!openParenTokenIndex) {
      throw this.error("Expected open-paren after super.");
    }
    return notNull(this.sourceTokenAtIndex(openParenTokenIndex));
  }
};

// src/stages/normalize/patchers/ThisPatcher.ts
import { SourceType as SourceType58 } from "coffee-lex";
var ThisPatcher2 = class extends PassthroughPatcher {
  patchAsRepeatableExpression(repeatableOptions = {}, patchOptions = {}) {
    const ref = super.patchAsRepeatableExpression(repeatableOptions, patchOptions);
    const addedParens = (!this.isRepeatable() || repeatableOptions.forceRepeat) && repeatableOptions.parens;
    if (addedParens && this.parent instanceof MemberAccessOpPatcher2) {
      const nextToken = this.nextSemanticToken();
      if (!nextToken || nextToken.type !== SourceType58.DOT) {
        this.insert(this.innerEnd, ".");
      }
    }
    return ref;
  }
};

// src/stages/normalize/patchers/TryPatcher.ts
var TryPatcher2 = class extends NodePatcher {
  constructor(patcherContext, body, catchAssignee, catchBody, finallyBody) {
    super(patcherContext);
    this.body = body;
    this.catchAssignee = catchAssignee;
    this.catchBody = catchBody;
    this.finallyBody = finallyBody;
  }
  patchAsExpression() {
    if (this.body) {
      this.body.patch();
    }
    const bodyPrefixLine = this.patchCatchAssignee();
    if (bodyPrefixLine !== null && this.catchAssignee) {
      if (this.catchBody) {
        this.catchBody.insertLineBefore(bodyPrefixLine);
      } else {
        this.insert(this.catchAssignee.outerEnd, ` then ${bodyPrefixLine}`);
      }
    }
    if (this.catchBody) {
      this.catchBody.patch();
    }
    if (this.finallyBody) {
      this.finallyBody.patch();
    }
  }
  patchCatchAssignee() {
    if (!this.catchAssignee) {
      return null;
    }
    if (this.needsExpressionExtracted()) {
      const assigneeName = this.claimFreeBinding("error");
      const assigneeCode = this.catchAssignee.patchAndGetCode();
      this.overwrite(this.catchAssignee.contentStart, this.catchAssignee.contentEnd, assigneeName);
      return `${assigneeCode} = ${assigneeName}`;
    } else {
      this.catchAssignee.patch();
      return null;
    }
  }
  needsExpressionExtracted() {
    if (!this.catchAssignee) {
      return false;
    }
    if (!(this.catchAssignee instanceof IdentifierPatcher2)) {
      return true;
    }
    const varName = this.catchAssignee.node.data;
    const exceptionVarUsages = this.catchBody ? countVariableUsages(this.catchBody.node, varName) + 1 : 1;
    const totalVarUsages = countVariableUsages(this.getScope().containerNode, varName);
    return totalVarUsages > exceptionVarUsages;
  }
};

// src/stages/normalize/patchers/WhilePatcher.ts
var WhilePatcher2 = class extends NodePatcher {
  constructor(patcherContext, condition, guard, body) {
    super(patcherContext);
    this.condition = condition;
    this.guard = guard;
    this.body = body;
  }
  patchAsExpression() {
    this.condition.patch();
    if (this.guard) {
      this.guard.patch();
    }
    if (this.body) {
      this.body.patch();
    }
    if (this.isPostWhile()) {
      this.normalize();
    }
  }
  patchAsStatement() {
    this.patchAsExpression();
  }
  normalize() {
    if (this.body === null) {
      throw this.error("Expected non-null body.");
    }
    let patchedCondition = this.slice(this.condition.outerStart, this.condition.outerEnd);
    if (postfixExpressionRequiresParens(patchedCondition) && !this.condition.isSurroundedByParentheses()) {
      patchedCondition = `(${patchedCondition})`;
    }
    const patchedBody = this.slice(this.body.outerStart, this.body.outerEnd);
    let patchedGuard = null;
    if (this.guard) {
      patchedGuard = this.slice(this.guard.outerStart, this.guard.outerEnd);
      if (postfixExpressionRequiresParens(patchedGuard) && !this.guard.isSurroundedByParentheses()) {
        patchedGuard = `(${patchedGuard})`;
      }
    }
    const whileToken = this.node.isUntil ? "until" : "while";
    let newContent = `${whileToken} ${patchedCondition} ${patchedGuard ? `when ${patchedGuard} ` : ""}then ${patchedBody}`;
    if (postfixNodeNeedsOuterParens(this)) {
      newContent = `(${newContent})`;
    }
    this.overwrite(this.contentStart, this.contentEnd, newContent);
  }
  isPostWhile() {
    return this.body !== null && this.condition.contentStart > this.body.contentStart;
  }
};

// src/stages/normalize/index.ts
var NormalizeStage = class extends TransformCoffeeScriptStage {
  patcherConstructorForNode(node) {
    switch (node.type) {
      case "ArrayInitialiser":
        return ArrayInitialiserPatcher2;
      case "MemberAccessOp":
        return MemberAccessOpPatcher2;
      case "DynamicMemberAccessOp":
        return DynamicMemberAccessOpPatcher2;
      case "Block":
        return BlockPatcher2;
      case "BoundFunction":
      case "Function":
      case "BoundGeneratorFunction":
      case "GeneratorFunction":
        return FunctionPatcher2;
      case "Conditional":
        return ConditionalPatcher2;
      case "Constructor":
        return ConstructorPatcher2;
      case "DoOp":
        return DoOpPatcher2;
      case "Expansion":
        return ExpansionPatcher2;
      case "ForIn":
        return ForInPatcher2;
      case "ForOf":
        return ForOfPatcher2;
      case "ForFrom":
        return ForFromPatcher2;
      case "FunctionApplication":
      case "NewOp":
      case "SoakedFunctionApplication":
      case "SoakedNewOp":
        return FunctionApplicationPatcher2;
      case "Super":
      case "BareSuperFunctionApplication":
        return SuperPatcher2;
      case "Identifier":
        return IdentifierPatcher2;
      case "While":
        return WhilePatcher2;
      case "Loop":
        return LoopPatcher2;
      case "Class":
        return ClassPatcher2;
      case "AssignOp":
      case "ClassProtoAssignOp":
      case "CompoundAssignOp":
        return AssignOpPatcher2;
      case "Program":
        return ProgramPatcher3;
      case "DefaultParam":
        return DefaultParamPatcher2;
      case "Rest":
      case "Spread":
        return SpreadPatcher2;
      case "ObjectInitialiser":
        return ObjectInitialiserPatcher2;
      case "ObjectInitialiserMember":
        return ObjectInitialiserMemberPatcher2;
      case "ProtoMemberAccessOp":
      case "SoakedProtoMemberAccessOp":
        return ProtoMemberAccessOpPatcher;
      case "Try":
        return TryPatcher2;
      case "This":
        return ThisPatcher2;
      default:
        return PassthroughPatcher;
    }
  }
};

// src/stages/semicolons/index.ts
import { parse } from "@codemod/parser";
import * as asi from "automatic-semicolon-insertion";
import MagicString3 from "magic-string";
var SemicolonsStage = class {
  static run(content) {
    const log = logger(this.name);
    log(content);
    const editor = new MagicString3(content);
    const ast = parse(content, {
      tokens: true
    });
    const { insertions, removals } = asi.process(content, ast);
    insertions.forEach(({ index, content: content2 }) => editor.appendLeft(index, content2));
    removals.forEach(({ start, end }) => editor.remove(start, end));
    return {
      code: editor.toString(),
      suggestions: []
    };
  }
};

// src/utils/CodeContext.ts
import { LinesAndColumns as LinesAndColumns2 } from "lines-and-columns";
var CodeContext = class {
  constructor(source) {
    this.source = source;
    this.linesAndColumns = new LinesAndColumns2(source);
  }
  formatRange(startIndex, endIndex) {
    return `[${this.formatIndex(startIndex)}-${this.formatIndex(endIndex)}]`;
  }
  formatIndex(index) {
    if (index > this.source.length) {
      index = this.source.length;
    }
    const location = this.linesAndColumns.locationForIndex(index);
    if (!location) {
      return "INVALID POSITION";
    }
    const { line, column } = location;
    return `${line + 1}:${column + 1}(${index})`;
  }
};

// src/utils/convertNewlines.ts
function convertNewlines(source, newlineStr) {
  if (newlineStr === "\n") {
    return source.replace(/\r\n/g, "\n");
  } else if (newlineStr === "\r\n") {
    source = source.replace(/\r\n/g, "\n");
    return source.replace(/\n/g, "\r\n");
  } else {
    throw new Error(`Unexpected newline string to convert to: ${JSON.stringify(newlineStr)}`);
  }
}

// src/utils/detectNewlineStr.ts
function detectNewlineStr(source) {
  let numLFs = 0;
  let numCRLFs = 0;
  for (let i = 0; i < source.length; i++) {
    if (source[i] === "\n" && (i === 0 || source[i - 1] !== "\r")) {
      numLFs++;
    }
    if (source.slice(i, i + 2) === "\r\n") {
      numCRLFs++;
    }
  }
  return numCRLFs > numLFs ? "\r\n" : "\n";
}

// src/utils/formatCoffeeLexTokens.ts
import { SourceType as SourceType59 } from "coffee-lex";
function formatCoffeeLexTokens(tokens, context) {
  const resultLines = tokens.map((token) => `${context.formatRange(token.start, token.end)}: ${SourceType59[token.type]}`);
  return resultLines.map((line) => line + "\n").join("");
}

// src/utils/formatCoffeeScriptAst.ts
import { Base as CS1Base } from "decaffeinate-coffeescript/lib/coffee-script/nodes";
import { Base as CS2Base } from "decaffeinate-coffeescript2/lib/coffeescript/nodes";

// src/utils/formatCoffeeScriptLocationData.ts
function formatCoffeeScriptLocationData(locationData, context) {
  const firstIndex = context.linesAndColumns.indexForLocation({
    line: locationData.first_line,
    column: locationData.first_column
  });
  if (firstIndex === null) {
    return "INVALID RANGE";
  }
  const lastIndex = context.linesAndColumns.indexForLocation({
    line: locationData.last_line,
    column: locationData.last_column
  });
  if (lastIndex === null) {
    return "INVALID RANGE";
  }
  return context.formatRange(firstIndex, lastIndex + 1);
}

// src/utils/formatCoffeeScriptAst.ts
function formatCoffeeScriptAst(program, context) {
  const resultLines = formatAstNodeLines(program, context);
  return resultLines.map((line) => line + "\n").join("");
}
function formatAstNodeLines(node, context) {
  const propLines = [];
  const blacklistedProps = ["locationData"];
  for (const key of Object.keys(node)) {
    const value = node[key];
    if (shouldTraverse(value) || blacklistedProps.indexOf(key) !== -1) {
      continue;
    }
    let valueText;
    try {
      valueText = JSON.stringify(value);
    } catch (e) {
      valueText = "(error)";
    }
    propLines.push(`${key}: ${valueText}`);
  }
  for (const key of Object.keys(node)) {
    const value = node[key];
    if (!shouldTraverse(value)) {
      continue;
    }
    if (Array.isArray(value) && value.length === 0) {
      propLines.push(`${key}: []`);
    } else if (Array.isArray(value)) {
      propLines.push(`${key}: [`);
      for (const child of value) {
        if (Array.isArray(child)) {
          propLines.push(`  [`);
          for (const grandchild of child) {
            propLines.push(...formatAstNodeLines(grandchild, context).map((s) => "    " + s));
          }
          propLines.push(`  ]`);
        } else {
          propLines.push(...formatAstNodeLines(child, context).map((s) => "  " + s));
        }
      }
      propLines.push(`]`);
    } else {
      const childLines = formatAstNodeLines(value, context);
      childLines[0] = `${key}: ${childLines[0]}`;
      propLines.push(...childLines);
    }
  }
  return [
    `${node.constructor.name} ${formatCoffeeScriptLocationData(node.locationData, context)} {`,
    ...propLines.map((s) => "  " + s),
    "}"
  ];
}
function shouldTraverse(value) {
  if (Array.isArray(value)) {
    return value.length === 0 || shouldTraverse(value[0]);
  }
  return value instanceof CS1Base || value instanceof CS2Base;
}

// src/utils/formatCoffeeScriptLexerTokens.ts
function formatCoffeeScriptLexerTokens(tokens, context) {
  const resultLines = tokens.map(([tag, value, locationData]) => `${formatCoffeeScriptLocationData(locationData, context)}: ${tag}: ${JSON.stringify(value)}`);
  return resultLines.map((line) => line + "\n").join("");
}

// src/utils/formatDecaffeinateParserAst.ts
function formatDecaffeinateParserAst(program, context) {
  const resultLines = formatAstNodeLines2(program, context);
  return resultLines.map((line) => line + "\n").join("");
}
function formatAstNodeLines2(node, context) {
  const propLines = [];
  const childPropNames = node.getChildNames();
  const blacklistedProps = childPropNames.concat([
    "raw",
    "line",
    "column",
    "type",
    "parentNode",
    "context",
    "start",
    "end"
  ]);
  for (const key of Object.keys(node)) {
    if (blacklistedProps.indexOf(key) !== -1) {
      continue;
    }
    let valueText;
    try {
      valueText = JSON.stringify(node[key]);
    } catch (e) {
      valueText = "(error)";
    }
    propLines.push(`${key}: ${valueText}`);
  }
  for (const childProp of childPropNames) {
    const value = node[childProp];
    if (value === null) {
      propLines.push(`${childProp}: null`);
    } else if (Array.isArray(value) && value.length === 0) {
      propLines.push(`${childProp}: []`);
    } else if (Array.isArray(value)) {
      propLines.push(`${childProp}: [`);
      for (const child of value) {
        propLines.push(...formatAstNodeLines2(child, context).map((s) => "  " + s));
      }
      propLines.push(`]`);
    } else {
      const childLines = formatAstNodeLines2(value, context);
      childLines[0] = `${childProp}: ${childLines[0]}`;
      propLines.push(...childLines);
    }
  }
  const rangeStr = context.formatRange(node.start, node.end);
  return [`${node.type} ${rangeStr} {`, ...propLines.map((s) => "  " + s), "}"];
}

// src/utils/removeUnicodeBOMIfNecessary.ts
function removeUnicodeBOMIfNecessary(source) {
  if (source[0] === "\uFEFF") {
    return source.slice(1);
  } else {
    return source;
  }
}

// src/utils/resolveToPatchError.ts
import { LinesAndColumns as LinesAndColumns3 } from "lines-and-columns";
function resolveToPatchError(err, content, stageName) {
  const makePatchError = (start, end, source) => new PatchError(`${stageName} failed to parse: ${err.message}`, source, start, end);
  if ("pos" in err) {
    let { pos } = err;
    if (pos === content.length) {
      pos--;
    }
    return makePatchError(pos, pos + 1, content);
  } else if ("syntaxError" in err) {
    const {
      syntaxError: { location }
    } = err;
    const lineMap = new LinesAndColumns3(content);
    const firstIndex = lineMap.indexForLocation({ line: location.first_line, column: location.first_column });
    let lastIndex = lineMap.indexForLocation({ line: location.last_line, column: location.last_column });
    if (firstIndex !== null) {
      if (lastIndex === null) {
        lastIndex = firstIndex + 1;
      } else {
        lastIndex++;
      }
      return makePatchError(firstIndex, lastIndex, content);
    }
  }
  return null;
}

// src/cli.ts
import assert2 from "assert";
import { readdir, readFile, stat, writeFile } from "mz/fs";
import { basename, dirname, extname, join } from "path";

// src/options.ts
var DEFAULT_OPTIONS = {
  filename: "input.coffee",
  useCS2: false,
  runToStage: null,
  literate: false,
  disableSuggestionComment: false,
  noArrayIncludes: false,
  useJSModules: false,
  looseJSModules: false,
  safeImportFunctionIdentifiers: [],
  preferLet: false,
  loose: false,
  looseDefaultParams: false,
  looseForExpressions: false,
  looseForOf: false,
  looseIncludes: false,
  looseComparisonNegation: false,
  disallowInvalidConstructors: false,
  optionalChaining: false,
  logicalAssignment: false,
  nullishCoalescing: false,
  bare: true
};
function resolveOptions(options) {
  if (options.loose) {
    options = {
      ...options,
      looseDefaultParams: true,
      looseForExpressions: true,
      looseForOf: true,
      looseIncludes: true,
      looseComparisonNegation: true,
      looseJSModules: true
    };
  }
  if (options.runToStage === "EsnextStage") {
    options = { ...options, runToStage: "ResugarStage" };
  }
  return { ...DEFAULT_OPTIONS, ...options };
}

// src/cli.ts
async function run(args, io = { stdin: process.stdin, stdout: process.stdout, stderr: process.stderr }) {
  const parseResult = parseArguments(args, io);
  if (parseResult.kind === "error") {
    io.stderr.write(`${parseResult.message}
`);
    return 1;
  }
  const options = parseResult;
  if (options.help) {
    usage(args[0], io.stdout);
    return 0;
  }
  if (options.version) {
    version(io.stdout);
    return 0;
  }
  const success = options.paths.length ? await runWithPaths(options.paths, options, io) : await runWithStdio(options, io);
  return success ? 0 : 1;
}
function parseArguments(args, io) {
  const paths = [];
  const baseOptions = { ...DEFAULT_OPTIONS };
  let modernizeJS2 = false;
  let help = false;
  let version2 = false;
  for (let i = 2; i < args.length; i++) {
    const arg = args[i];
    switch (arg) {
      case "-h":
      case "--help":
        help = true;
        break;
      case "-v":
      case "--version":
        version2 = true;
        break;
      case "--use-cs2":
        baseOptions.useCS2 = true;
        break;
      case "--modernize-js":
        modernizeJS2 = true;
        break;
      case "--bare":
      case "--no-bare":
        baseOptions.bare = arg === "--bare";
        break;
      case "--literate":
        baseOptions.literate = true;
        break;
      case "--disable-suggestion-comment":
        baseOptions.disableSuggestionComment = true;
        break;
      case "--no-array-includes":
        baseOptions.noArrayIncludes = true;
        break;
      case "--use-js-modules":
        baseOptions.useJSModules = true;
        break;
      case "--loose-js-modules":
        baseOptions.looseJSModules = true;
        break;
      case "--safe-import-function-identifiers":
        i++;
        baseOptions.safeImportFunctionIdentifiers = args[i].split(",");
        break;
      case "--prefer-let":
        baseOptions.preferLet = true;
        break;
      case "--disable-babel-constructor-workaround":
        io.stderr.write(`${arg} no longer has any effect as it is the only supported behavior
`);
        break;
      case "--disallow-invalid-constructors":
        baseOptions.disallowInvalidConstructors = true;
        break;
      case "--loose":
        baseOptions.loose = true;
        break;
      case "--loose-default-params":
        baseOptions.looseDefaultParams = true;
        break;
      case "--loose-for-expressions":
        baseOptions.looseForExpressions = true;
        break;
      case "--loose-for-of":
        baseOptions.looseForOf = true;
        break;
      case "--loose-includes":
        baseOptions.looseIncludes = true;
        break;
      case "--loose-comparison-negation":
        baseOptions.looseComparisonNegation = true;
        break;
      case "--prefer-const":
      case "--keep-commonjs":
      case "--enable-babel-constructor-workaround":
        break;
      case "--force-default-export":
        baseOptions.useJSModules = true;
        break;
      case "--allow-invalid-constructors":
        baseOptions.disallowInvalidConstructors = false;
        break;
      case "--optional-chaining":
      case "--use-optional-chaining":
        baseOptions.optionalChaining = true;
        break;
      case "--logical-assignment":
        baseOptions.logicalAssignment = true;
        break;
      case "--nullish-coalescing":
        baseOptions.nullishCoalescing = true;
        break;
      default:
        if (arg.startsWith("-")) {
          return { kind: "error", message: `unrecognized option: ${arg}` };
        }
        paths.push(arg);
        break;
    }
  }
  if (!baseOptions.bare && baseOptions.useJSModules) {
    return { kind: "error", message: "cannot use --use-js-modules with --no-bare" };
  }
  return { kind: "success", paths, baseOptions, modernizeJS: modernizeJS2, version: version2, help };
}
async function runWithPaths(paths, options, io) {
  async function processPath(path) {
    const info = await stat(path);
    if (info.isDirectory()) {
      return await processDirectory(path);
    } else {
      return await processFile(path);
    }
  }
  async function processDirectory(path) {
    const children = await readdir(path);
    for (const child of children) {
      const childPath = join(path, child);
      const childStat = await stat(childPath);
      if (childStat.isDirectory()) {
        if (!await processDirectory(childPath)) {
          return false;
        }
      } else if (options.modernizeJS) {
        if (child.endsWith(".js")) {
          if (!await processPath(childPath)) {
            return false;
          }
        }
      } else if (child.endsWith(".coffee") || child.endsWith(".litcoffee") || child.endsWith(".coffee.md")) {
        if (!await processPath(childPath)) {
          return false;
        }
      }
    }
    return true;
  }
  async function processFile(path) {
    const extension = path.endsWith(".coffee.md") ? ".coffee.md" : extname(path);
    const outputPath = join(dirname(path), basename(path, extension)) + ".js";
    io.stdout.write(`${path} \u2192 ${outputPath}
`);
    const data = await readFile(path, "utf8");
    const resultCode = runWithCode(path, data, options, io);
    const success = typeof resultCode === "string";
    if (success) {
      await writeFile(outputPath, resultCode);
    }
    return success;
  }
  for (const path of paths) {
    if (!await processPath(path)) {
      return false;
    }
  }
  return true;
}
async function runWithStdio(options, io) {
  return new Promise((resolve) => {
    let data = "";
    io.stdin.on("data", (chunk) => data += chunk);
    io.stdin.on("end", () => {
      const resultCode = runWithCode("stdin", data, options, io);
      const success = typeof resultCode === "string";
      if (success) {
        io.stdout.write(resultCode);
      }
      resolve(success);
    });
  });
}
function runWithCode(name, code, options, io) {
  const baseOptions = Object.assign({ filename: name }, options.baseOptions);
  try {
    if (options.modernizeJS) {
      return modernizeJS(code, baseOptions).code;
    } else {
      return convert(code, baseOptions).code;
    }
  } catch (err) {
    assert2(err instanceof Error);
    if (PatchError.detect(err)) {
      io.stderr.write(`${name}: ${PatchError.prettyPrint(err)}
`);
      return void 0;
    }
    throw err;
  }
}
function version(out) {
  out.write(`${"decaffeinate"} v${"8.1.2"}`);
}
function usage(exe, out) {
  out.write(`${exe} [OPTIONS] PATH [PATH \u2026]
`);
  out.write(`${exe} [OPTIONS] < INPUT
`);
  out.write("\n");
  out.write("Move your CoffeeScript source to JavaScript using modern syntax.\n");
  out.write("\n");
  out.write("OPTIONS\n");
  out.write("\n");
  out.write("  -h, --help               Display this help message.\n");
  out.write("  --use-cs2                Treat the input as CoffeeScript 2 code. CoffeeScript 2 has\n");
  out.write("                           some small breaking changes and differences in behavior\n");
  out.write("                           compared with CS1, so decaffeinate assumes CS1 by default\n");
  out.write("                           and allows CS2 via this flag.\n");
  out.write("  --modernize-js           Treat the input as JavaScript and only run the\n");
  out.write("                           JavaScript-to-JavaScript transforms, modifying the file(s)\n");
  out.write("                           in-place.\n");
  out.write("  --literate               Treat the input file as Literate CoffeeScript.\n");
  out.write("  --disable-suggestion-comment\n");
  out.write("                           Do not include a comment with followup suggestions at the\n");
  out.write("                           top of the output file.\n");
  out.write("  --no-array-includes      Do not use Array.prototype.includes in generated code.\n");
  out.write("  --use-js-modules         Convert require and module.exports to import and export.\n");
  out.write("  --loose-js-modules       Allow named exports when converting to JS modules.\n");
  out.write("  --safe-import-function-identifiers\n");
  out.write("                           Comma-separated list of function names that may safely be in the \n");
  out.write("                           import/require section of the file. All other function calls \n");
  out.write("                           will disqualify later requires from being converted to imports.\n");
  out.write("  --prefer-let             Use let instead of const for most variables in output code.\n");
  out.write("  --loose                  Enable all --loose... options.\n");
  out.write("  --loose-default-params   Convert CS default params to JS default params.\n");
  out.write("  --loose-for-expressions  Do not wrap expression loop targets in Array.from.\n");
  out.write("  --loose-for-of           Do not wrap JS for...of loop targets in Array.from.\n");
  out.write("  --loose-includes         Do not wrap in Array.from when converting in to includes.\n");
  out.write("  --loose-comparison-negation\n");
  out.write("                           Allow unsafe simplifications like `!(a > b)` to `a <= b`.\n");
  out.write("  --disallow-invalid-constructors\n");
  out.write("                           Give an error when constructors use this before super or\n");
  out.write("                           omit the super call in a subclass.\n");
  out.write("  --optional-chaining      Target JavaScript optional chaining. Note the semantics may not\n");
  out.write("                           match exactly.\n");
  out.write("  --logical-assignment     Use the ES2021 logical assignment operators `&&=`, `||=`,\n");
  out.write("                           and `??=`.\n");
  out.write("\n");
  out.write("EXAMPLES\n");
  out.write("\n");
  out.write("  # Convert a .coffee file to a .js file.\n");
  out.write("  $ decaffeinate index.coffee\n");
  out.write("\n");
  out.write("  # Pipe an example from the command-line.\n");
  out.write('  $ echo "a = 1" | decaffeinate\n');
  out.write("\n");
  out.write("  # On macOS this may come in handy:\n");
  out.write("  $ pbpaste | decaffeinate | pbcopy\n");
  out.write("\n");
  out.write("  # Process everything in a directory.\n");
  out.write("  $ decaffeinate src/\n");
  out.write("\n");
  out.write("  # Redirect input from a file.\n");
  out.write("  $ decaffeinate < index.coffee\n");
}

// src/index.ts
import assert3 from "assert";
function convert(source, options = {}) {
  if (!options.bare && options.useJSModules) {
    throw new Error("useJSModules requires bare output");
  }
  source = removeUnicodeBOMIfNecessary(source);
  options = resolveOptions(options);
  const originalNewlineStr = detectNewlineStr(source);
  source = convertNewlines(source, "\n");
  const literate = options.literate || notNull(options.filename).endsWith(".litcoffee") || notNull(options.filename).endsWith(".coffee.md");
  let stages = [
    ...literate ? [LiterateStage] : [],
    NormalizeStage,
    MainStage,
    AddVariableDeclarationsStage,
    SemicolonsStage,
    ResugarStage
  ];
  const runToStage = options.runToStage;
  if (runToStage !== null && runToStage !== void 0) {
    const stageIndex = stages.findIndex((stage) => stage.name === runToStage);
    if (stageIndex !== -1) {
      stages = stages.slice(0, stageIndex + 1);
    } else {
      return convertCustomStage(source, runToStage, Boolean(options.useCS2));
    }
  }
  const result = runStages(source, options, stages);
  if (!options.disableSuggestionComment) {
    result.code = prependSuggestionComment(result.code, result.suggestions);
  }
  result.code = convertNewlines(result.code, originalNewlineStr);
  return {
    code: options.bare ? result.code : `(function() {
${result.code}
}).call(this);`
  };
}
function modernizeJS(source, options = {}) {
  source = removeUnicodeBOMIfNecessary(source);
  options = resolveOptions(options);
  const originalNewlineStr = detectNewlineStr(source);
  source = convertNewlines(source, "\n");
  const stages = [ResugarStage];
  const result = runStages(source, options, stages);
  result.code = convertNewlines(result.code, originalNewlineStr);
  return {
    code: options.bare ? result.code : `(function() {
${result.code}
}).call(this);`
  };
}
function runStages(initialContent, options, stages) {
  let content = initialContent;
  const suggestions = [];
  stages.forEach((stage) => {
    const { code, suggestions: stageSuggestions } = runStage(stage, content, options);
    content = code;
    suggestions.push(...stageSuggestions);
  });
  return { code: content, suggestions: mergeSuggestions(suggestions) };
}
function runStage(stage, content, options) {
  try {
    return stage.run(content, options);
  } catch (err) {
    assert3(err instanceof Error);
    const patchError = resolveToPatchError(err, content, stage.name);
    if (patchError !== null) {
      throw patchError;
    }
    throw err;
  }
}
function convertCustomStage(source, stageName, useCS2) {
  const context = new CodeContext(source);
  if (stageName === "coffeescript-lexer") {
    const tokens = useCS2 ? getCoffee2Tokens(source) : getCoffee1Tokens(source);
    return {
      code: formatCoffeeScriptLexerTokens(tokens, context)
    };
  } else if (stageName === "coffeescript-parser") {
    const nodes = useCS2 ? getCoffee2Nodes(source) : getCoffee1Nodes(source);
    return {
      code: formatCoffeeScriptAst(nodes, context)
    };
  } else if (stageName === "coffee-lex") {
    return {
      code: formatCoffeeLexTokens(lex(source, { useCS2 }), context)
    };
  } else if (stageName === "decaffeinate-parser") {
    return {
      code: formatDecaffeinateParserAst(decaffeinateParse2(source, { useCS2 }), context)
    };
  } else {
    throw new Error(`Unrecognized stage name: ${stageName}`);
  }
}
export {
  PatchError,
  convert,
  modernizeJS,
  run
};
